#define INCL_WIN
#define INCL_GPI
#define INCL_DOS
#define INCL_DOSPROCESS
#define INCL_DOSERRORS
#define INCL_MCIOS2
#define INCL_MACHDR

#include <os2.h>
#include <os2me.h>
#include <dive.h>
#include <stdio.h>
#include <malloc.h>
#include <process.h>
#include <string.h>
#include <stdlib.h>
#include "SessionManager.h"
#include "Resource.h"

#define IM_IDLE_TIMEOUT      300
// Timeout before reporting idle status to AOL server

#define BUBBLE_POPUP_TIMEOUT 500
// Time of mouse motionlessness before popping up bubble info/help
// Hard-coded for now

#define WID_SESSION_CONTAINER 100
#define WID_DELETE_SESSION    104

#define WID_FILTER_ONLINE     200
#define WID_FILTER_AWAY       201
#define WID_FILTER_OFFLINE    202

// Menu ID is also equal to WID_'s here, so only one set of WM_COMMAND
//  processing is needed.

#define SESSION_MGR_WIDTH  640
#define SESSION_MGR_HEIGHT 480
#define BUDDY_LIST_WIDTH   320
#define BUDDY_LIST_HEIGHT  480
#define IM_WINDOW_WIDTH    320
#define IM_WINDOW_HEIGHT   320
// Hard-coded defaults for now

const char *UI_EventTypeStrings[] =
{
  "Program Startup",
  "Session Startup",
  "Error Message",
  "First IM with Buddy",
  "IM Received from Buddy",
  "IM Sent to Buddy",
  "Buddy Goes Online",
  "Buddy Goes Offline",
  "Session End"
};

void sessionThread( sessionThreadInfo *info )
{
  debugf( "Session thread started.\n" );

  new OscarSession( info->userName, info->password,
   info->managerWin, info->record, &info->theSession, &info->settings );
  // Doing this in a separate thread allows us to service the UI separately
  //  from communicating with the AOL servers.  Definitely a good thing if
  //  the AOL servers are slow or non-responsive.

  info->theSession->serviceSessionUntilClosed();
  
  delete info->theSession;
  
  debugf( "Session object was deleted.\n" );
  
  info->theSession = NULL;
  info->threadID = 0;
  
  debugf( "Session thread exiting.\n" );
}

#define WID_SEND_BUTTON 101
#define WID_IM_RECEIVED 102
#define WID_IM_SEND     103


static void querySessionSettings( HINI iniFile,
 char *appName, sessionThreadSettings *storeHere )
{
  char buffer[40];
  int i;
  ULONG len;
  unsigned char oldInherit = storeHere->inheritSettings;

  len = 1;
  PrfQueryProfileData( iniFile, appName, "Inherit Settings",
   &(storeHere->inheritSettings), &len );
  if ( storeHere->inheritSettings ) return;
  
  for ( i=0; i<EVENT_MAXEVENTS; ++i )
  {
    sprintf( buffer, "Event %d Flags", i );
    len = 4;
    PrfQueryProfileData( iniFile, appName, (char *)buffer,
     &(storeHere->settingFlags[i]), &len );
    
    sprintf( buffer, "Event %d Sound", i );
    if ( PrfQueryProfileSize( iniFile, appName, (char *)buffer, &len ) )
    {
      if ( storeHere->sounds[i] && !oldInherit )
      {
        CHECKED_FREE( storeHere->sounds[i] );
        storeHere->sounds[i] = NULL;
      }
      CHECKED_MALLOC( len + 1, storeHere->sounds[i] );
      PrfQueryProfileData( iniFile, appName, (char *)buffer,
       storeHere->sounds[i], &len );
      storeHere->sounds[i][len] = 0;
    }
    sprintf( buffer, "Event %d REXX Script", i );
    if ( PrfQueryProfileSize( iniFile, appName, (char *)buffer, &len ) )
    {
      if ( storeHere->rexxScripts[i] && !oldInherit)
      {
        CHECKED_FREE( storeHere->rexxScripts[i] );
        storeHere->sounds[i] = NULL;
      }
      CHECKED_MALLOC( len + 1, storeHere->rexxScripts[i] );
      PrfQueryProfileData( iniFile, appName, (char *)buffer,
       storeHere->rexxScripts[i], &len );
      storeHere->rexxScripts[i][len] = 0;
    }
    sprintf( buffer, "Event %d Shell Script", i );
    if ( PrfQueryProfileSize( iniFile, appName, (char *)buffer, &len ) )
    {
      if ( storeHere->shellCmds[i] && !oldInherit )
      {
        CHECKED_FREE( storeHere->shellCmds[i] );
        storeHere->shellCmds[i] = NULL;
      }
      CHECKED_MALLOC( len + 1, storeHere->shellCmds[i] );
      PrfQueryProfileData( iniFile, appName, (char *)buffer,
       storeHere->shellCmds[i], &len );
      storeHere->shellCmds[i][len] = 0;
    }
  }
}

static void saveSessionSettings( HINI iniFile,
 char *appName, sessionThreadSettings *storeThis )
{
  char buffer[40];
  int i;
  
  PrfWriteProfileData( iniFile, appName, "Inherit Settings",
   &(storeThis->inheritSettings), 1 );
  if ( storeThis->inheritSettings ) return;

  for ( i=0; i<EVENT_MAXEVENTS; ++i )
  {
    sprintf( buffer, "Event %d Flags", i );
    PrfWriteProfileData( iniFile, appName, (char *)buffer,
     &(storeThis->settingFlags[i]), sizeof( unsigned long ) );
    
    sprintf( buffer, "Event %d Sound", i );
    if ( storeThis->sounds[i] )
      PrfWriteProfileData( iniFile, appName, (char *)buffer,
       storeThis->sounds[i], strlen( storeThis->sounds[i] ) );
    else PrfWriteProfileData( iniFile, appName, (char *)buffer, NULL, 0 );
    
    sprintf( buffer, "Event %d REXX Script", i );
    if ( storeThis->rexxScripts[i] )
      PrfWriteProfileData( iniFile, appName, (char *)buffer,
       storeThis->rexxScripts[i], strlen( storeThis->rexxScripts[i] ) );
    else PrfWriteProfileData( iniFile, appName, (char *)buffer, NULL, 0 );
    
    sprintf( buffer, "Event %d Shell Script", i );
    if ( storeThis->shellCmds[i] )
      PrfWriteProfileData( iniFile, appName, (char *)buffer,
       storeThis->shellCmds[i], strlen( storeThis->shellCmds[i] ) );
    else PrfWriteProfileData( iniFile, appName, (char *)buffer, NULL, 0 );
  }
}

MRESULT EXPENTRY mleEnterPerLine( HWND myWin, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  HWND win = WinQueryWindow( myWin, QW_PARENT );
  MRESULT EXPENTRY (*oldProc) ( HWND, ULONG, MPARAM, MPARAM );
  
  switch( msg )
  {
    case WM_CHAR:
    {
      if ( !(SHORT1FROMMP( mp1 ) & KC_KEYUP) &&
            ((SHORT1FROMMP( mp1 ) & KC_VIRTUALKEY) &&
             (SHORT2FROMMP( mp2 ) == VK_NEWLINE ||
              SHORT2FROMMP( mp2 ) == VK_ENTER)) )
      {
        // Pressed ENTER.  If there is something to send, simulate pressing
        //  the SEND button.
        
        ULONG textLen = LONGFROMMR( WinSendMsg( myWin, MLM_QUERYTEXTLENGTH,
         NULL, NULL ) );
        
        if ( textLen )
        {
          WinSendMsg( win, WM_COMMAND, MPFROMSHORT( WID_SEND_BUTTON ), NULL );
        } else {
          WinAlarm( HWND_DESKTOP, WA_WARNING );
        }
        return 0;
      }
    }
    break;
  }
  
  oldProc = (MRESULT EXPENTRY (*) ( HWND, ULONG, MPARAM, MPARAM ))
   WinQueryWindowPtr( win, 12 );
  
  return (*oldProc) ( myWin, msg, mp1, mp2 );
}

#define SPLASH_WIDTH 396
#define SPLASH_HEIGHT 300

MRESULT EXPENTRY pmSplashScreen( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch ( msg )
  {
    case WM_CREATE:
    {
      HBITMAP bmp;
      HMODULE dll = getModuleHandle();
      HPS hps = WinGetPS( win );
      RECTL desktop;
      sessionThreadSettings *mySettings = (sessionThreadSettings *)
       PVOIDFROMMP( mp1 );
      
      bmp = GpiLoadBitmap( hps, dll, 1, 0, 0 );
      
      WinReleasePS( hps );
      WinSetWindowULong( win, 0, bmp );
      
      WinQueryWindowRect( HWND_DESKTOP, &desktop );
      WinSetWindowPos( win, 0, ((desktop.xRight - desktop.xLeft) / 2) -
       (SPLASH_WIDTH / 2), ((desktop.yTop - desktop.yBottom) / 2) -
       (SPLASH_HEIGHT / 2), SPLASH_WIDTH, SPLASH_HEIGHT, SWP_SIZE | SWP_MOVE |
       SWP_SHOW );
      
      if ( mySettings->settingFlags[EVENT_SPLASHSCREEN] &
            SET_SOUND_ENABLED )
      {
        mciPlayFile( HWND_DESKTOP, mySettings->sounds[EVENT_SPLASHSCREEN],
         MCI_ASYNC, NULL, NULLHANDLE );
      }
      
      WinStartTimer( WinQueryAnchorBlock( win ), win, 1, 5000 );
    }
    break;
    case WM_PAINT:
    {
      HBITMAP bmp = WinQueryWindowULong( win, 0 );
      RECTL rectl;
      HPS hps = WinBeginPaint( win, 0, &rectl );
      POINTL ptl;
      
      ptl.x = rectl.xLeft;
      ptl.y = rectl.yBottom;
      
      WinDrawBitmap( hps, bmp, &rectl, &ptl, 0, 0, DBM_NORMAL );
      
      WinEndPaint( hps );
    }
    break;
    case WM_TIMER:
      WinPostMsg( win, WM_QUIT, 0, 0 );
    break;
    case WM_DESTROY:
      GpiDeleteBitmap( WinQueryWindowULong( win, 0 ) );
    break;
  }
  return WinDefWindowProc( win, msg, mp1, mp2 );
}

MRESULT EXPENTRY pmIMwindow( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch( msg )
  {
    case WM_BUDDYSTATUS:
      if ( LONGFROMMP( mp1 ) )
      {
        ULONG textLen = LONGFROMMR( WinSendDlgItemMsg( win, WID_IM_SEND,
         MLM_QUERYTEXTLENGTH, NULL, NULL ) );
        
        WinSetWindowULong( win, 8, 1 );
        WinEnableWindow( WinWindowFromID( win, WID_SEND_BUTTON ),
         textLen != 0 );
      } else {
        WinSetWindowULong( win, 8, 0 );
        WinEnableWindow( WinWindowFromID( win, WID_SEND_BUTTON ), FALSE );
      }
    break;
    case WM_CREATE:
    {
      HWND theWin;
      MLECTLDATA mleData;
      PFNWP oldFunc;
      
      // Create our controls, but wait until we know the window size to
      //  position them.  Wait for the initial WM_SIZE message.
      
      theWin = WinCreateWindow( win, WC_BUTTON, "Send",
       WS_DISABLED | WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_DEFAULT |
        BS_PUSHBUTTON, 0, 0, 0, 0, win, HWND_TOP, WID_SEND_BUTTON,
        NULL, NULL );
      
      theWin = WinCreateWindow( win, WC_BUTTON, "Close",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON, 0, 0, 0, 0, win,
       HWND_TOP, DID_CANCEL, NULL, NULL );
      
      mleData.cbCtlData = sizeof( MLECTLDATA );
      mleData.afIEFormat = MLFIE_CFTEXT;
      mleData.cchText = 65536;
      mleData.iptAnchor = 0;
      mleData.iptCursor = 0;
      mleData.cxFormat = 0;
      mleData.cyFormat = 0;
      mleData.afFormatFlags = MLFFMTRECT_MATCHWINDOW;

      theWin = WinCreateWindow( win, WC_MLE, NULL,
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | MLS_BORDER | MLS_READONLY |
       MLS_VSCROLL | MLS_LIMITVSCROLL | MLS_WORDWRAP, 0, 0, 0, 0, win,
       HWND_TOP, WID_IM_RECEIVED, &mleData, NULL );
      
      theWin = WinCreateWindow( win, WC_MLE, NULL,
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | MLS_BORDER | MLS_VSCROLL |
       MLS_WORDWRAP, 0, 0, 0, 0, win, HWND_TOP, WID_IM_SEND, &mleData,
       NULL );
      
      oldFunc = WinSubclassWindow( theWin, mleEnterPerLine );
      WinSetWindowPtr( win, 12, oldFunc );
      
      WinSetWindowULong( win, 8, 1 );
    }
    break;
    case WM_SIZE:
    {
      int newCx, newCy, buttonHeight, heightLeft;
      HPS hps;
      POINTL points[3];
      HWND theWin;
      
      newCx = SHORT1FROMMP( mp2 );
      newCy = SHORT2FROMMP( mp2 );
      
      theWin = WinWindowFromID( win, WID_SEND_BUTTON );
      hps = WinGetPS( theWin );
      // Make sure we pick up presentation params appropriate to the buttons
      //  when sizing the text.
      
      GpiQueryTextBox( hps, 10, "Send Close", 3, (PPOINTL) points );
      buttonHeight = points[0].y - points[1].y + 8;
      
      WinReleasePS( hps );
      
      WinSetWindowPos( theWin, 0, newCx / 50, newCy / 50, 47 * newCx / 100,
       buttonHeight, SWP_SIZE | SWP_MOVE );
      
      theWin = WinWindowFromID( win, DID_CANCEL );
      
      WinSetWindowPos( theWin, 0, 25 * newCx / 50, newCy / 50, 47 * newCx / 100,
       buttonHeight, SWP_SIZE | SWP_MOVE );
      
      heightLeft = newCy - buttonHeight - (newCy / 25);
      
      theWin = WinWindowFromID( win, WID_IM_RECEIVED );
      
      buttonHeight += newCx / 50;
      
      WinSetWindowPos( theWin, 0, newCx / 50,
       buttonHeight + (heightLeft * 51 / 100), newCx * 24 / 25,
       heightLeft * 49 / 100, SWP_SIZE | SWP_MOVE );
       
      theWin = WinWindowFromID( win, WID_IM_SEND );
      
      WinSetWindowPos( theWin, 0, newCx / 50,
       buttonHeight + (heightLeft / 100), newCx * 24 / 25,
       heightLeft * 49 / 100, SWP_SIZE | SWP_MOVE );
    }
    break;
    case WM_FOCUSCHANGE:
    {
      if ( SHORT1FROMMP( mp2 ) )
      {
        // Getting the focus.  Move the focus to the multi-line entry window.
        WinSetFocus( HWND_DESKTOP, WinWindowFromID( win, WID_IM_SEND ) );
      }
    }
    break;
    case WM_PAINT:
    {
      HPS hps;
      RECTL rect;
      POINTL pt;
      ULONG colorIdx, ret;
      
      ret = WinQueryPresParam( win, PP_BACKGROUNDCOLORINDEX, 0, NULL, 4,
       &colorIdx, QPF_ID1COLORINDEX | QPF_NOINHERIT );
      if ( !ret )
      {
        ret = WinQueryPresParam( win, PP_BACKGROUNDCOLOR, 0, NULL, 4,
         &colorIdx, QPF_NOINHERIT );
        if ( !ret )
        {
          ret = WinQueryPresParam( win, PP_BACKGROUNDCOLORINDEX, 0, NULL, 4,
           &colorIdx, QPF_ID1COLORINDEX );
        }
      }
      
      hps = WinBeginPaint( win, NULLHANDLE, &rect );
      if ( ret < 4 )
        GpiSetColor( hps, CLR_PALEGRAY );
      else {
        GpiCreateLogColorTable( hps, 0, LCOLF_RGB, 0, 0, NULL );
        GpiSetColor( hps, colorIdx );
      }
      pt.x = rect.xLeft;
      pt.y = rect.yBottom;
      GpiMove( hps, &pt );
      pt.x = rect.xRight;
      pt.y = rect.yTop;
      GpiBox( hps, DRO_FILL, &pt, 0, 0 );
      WinEndPaint( hps );
    }
    break;
    case WM_CLOSE:
    {
      HWND buddyListWin = WinQueryWindowULong( win, 4 );
      Buddy *theBuddy = (Buddy *) WinQueryWindowPtr( win, 0 );
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr(
       buddyListWin, 24 );
      HWND mgrWin = WinQueryWindowULong( buddyListWin, 16 );
      HWND frameWin = WinQueryWindow( win, QW_PARENT );
      HINI iniFile = WinQueryWindowULong( mgrWin, 20 );
      char *appName;
      SWP swp;
      int len;
      
      WinDestroyWindow( WinWindowFromID( win, WID_SEND_BUTTON ) );
      WinDestroyWindow( WinWindowFromID( win, DID_CANCEL ) );
      
      WinSendMsg( WinWindowFromID( buddyListWin, WID_SESSION_CONTAINER ),
       CM_SETRECORDEMPHASIS, MPFROMP( theBuddy->record ),
       MPFROM2SHORT( FALSE, CRA_INUSE ) );
      
      theBuddy->imChatWin = 0;
      theBuddy->imChatClientWin = 0;
      
      len = strlen( theSession->getUserName() ) +
       strlen( theBuddy->record->pszTree ) + 30;
      
      CHECKED_MALLOC( len, appName );
      
      sprintf( appName, "[%s <-> %s] Instant Message Window",
       theSession->getUserName(), theBuddy->record->pszTree );
       
      WinQueryWindowPos( frameWin, &swp );
      if ( (swp.fl & SWP_MINIMIZE) || (swp.fl & SWP_MAXIMIZE) )
      {
        swp.x = WinQueryWindowUShort( frameWin, QWS_XRESTORE );
        swp.y = WinQueryWindowUShort( frameWin, QWS_YRESTORE );
        swp.cx = WinQueryWindowUShort( frameWin, QWS_CXRESTORE );
        swp.cy = WinQueryWindowUShort( frameWin, QWS_CYRESTORE );
      }
      
      PrfWriteProfileData( iniFile, appName, "Window position data",
       &swp, sizeof( SWP ) );
      
      CHECKED_FREE( appName );
      
      WinDestroyWindow( frameWin );
      
      return 0;
      // Prevent the default window procedure from sending a WM_QUIT to the
      //  message queue and shutting down the whole app.
    }
    case WM_COMMAND:
      switch ( SHORT1FROMMP( mp1 ) )
      {
        case DID_CANCEL:
          WinSendMsg( win, WM_CLOSE, 0, 0 );
        break;
        case WID_SEND_BUTTON:
        {
          HWND buddyListWin = WinQueryWindowULong( win, 4 );
          Buddy *theBuddy = (Buddy *) WinQueryWindowPtr( win, 0 );
          OscarSession *theSession = (OscarSession *) WinQueryWindowPtr(
           buddyListWin, 24 );
          OscarData *theIM;
          char *theText, *annotatedMessage;
          ULONG origLen = LONGFROMMR( WinSendDlgItemMsg( win, WID_IM_SEND,
           MLM_QUERYTEXTLENGTH, NULL, NULL ) );
          ULONG textLen;
          
          if ( !theBuddy->userData.isOnline )
          {
            WinAlarm( HWND_DESKTOP, WA_WARNING );
            break;
          }
          
          if ( theSession->settings->settingFlags[ EVENT_SENT ] &
                SET_SOUND_ENABLED )
          {
            mciPlayFile( HWND_DESKTOP,
             theSession->settings->sounds[ EVENT_SENT ], MCI_ASYNC, NULL,
             NULLHANDLE );
          }
          
          theIM = new OscarData( theSession->getSeqNumPointer() );
          
          textLen = origLen + 1;
          
          CHECKED_MALLOC( textLen, theText );
          WinQueryDlgItemText( win, WID_IM_SEND, textLen, theText );
          theText[ textLen - 1 ] = 0;
          theIM->sendInstantMessage( theBuddy->record->pszTree, theText );
          
          CHECKED_MALLOC( strlen( theText ) +
           strlen( theSession->getUserName() ) + 9, annotatedMessage );
          
          sprintf( annotatedMessage, "[%s]: %s\n\n", theSession->getUserName(),
           theText );
          
          CHECKED_FREE( theText );
          
          textLen = LONGFROMMR( WinSendDlgItemMsg( win, WID_IM_RECEIVED,
           MLM_QUERYTEXTLENGTH, NULL, NULL ) );
          
          WinSendDlgItemMsg( win, WID_IM_RECEIVED, MLM_SETSEL,
           MPFROMLONG( textLen ), MPFROMLONG( textLen ) );
          // Move the insertion point to the end of the text, in case the user
          //  clicked somewhere.
          
          WinSendDlgItemMsg( win, WID_IM_RECEIVED, MLM_INSERT,
           MPFROMP( annotatedMessage ), NULL );
          
          CHECKED_FREE( annotatedMessage );
          
          theSession->queueForSend( theIM );
          
          WinSendDlgItemMsg( win, WID_IM_SEND, MLM_DELETE, MPFROMLONG( 0 ),
           MPFROMLONG( origLen ) );
        }
        break;
      }
    break;
    case WM_CONTROL:
      switch ( SHORT1FROMMP( mp1 ) )
      {
        case WID_IM_SEND:
          if ( SHORT2FROMMP( mp1 ) == MLN_CHANGE )
          {
            HWND buddyListWin = WinQueryWindowULong( win, 4 );
            // Text changed in entry field.
            ULONG textLen = LONGFROMMR( WinSendDlgItemMsg( win, WID_IM_SEND,
             MLM_QUERYTEXTLENGTH, NULL, NULL ) );
            ULONG allowSend = WinQueryWindowULong( win, 8 );
            
            if ( textLen ) WinSendMsg( buddyListWin, WM_NOTIDLE, 0, 0 );
            WinEnableWindow( WinWindowFromID( win, WID_SEND_BUTTON ),
             textLen != 0 && allowSend );
          }
        break;
      }
    break;
    case WM_MOUSEMOVE:
    case WM_CHAR:
    {
      HWND buddyListWin = WinQueryWindowULong( win, 4 );
      WinSendMsg( buddyListWin, WM_NOTIDLE, 0, 0 );
    }
    break;
  }
  
  return WinDefWindowProc( win, msg, mp1, mp2 );
}


typedef struct
{
  char *windowText;
} bubbleWindowInitStruct;

MRESULT EXPENTRY pmBubbleWindow( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch ( msg )
  {
    case WM_CREATE:
    {
      bubbleWindowInitStruct *initParms = (bubbleWindowInitStruct *)
       PVOIDFROMMP( mp1 );
      char *tmpString, *strPos, *oldPos;
      int len, cx, cy, i, numLines, maxSize;
      HPS hps, hpsMem;
      HDC windowDC, memDC;
      HBITMAP backgroundBmp;
      SIZEL psSize;
      POINTL textBox[3];
      SWP winPos;
      DEVOPENSTRUC dop = {0L, "DISPLAY", NULL, 0L, 0L, 0L, 0L, 0L, 0L};
      BITMAPINFOHEADER2 bmpInfo = { 0 };
      SIZEL sizl = { 0, 0 };
      FATTRS fat = { 0 };
      PFONTMETRICS pfm;
      char *theFont = "WarpSans";
      LONG tmpLong, numFonts, hRes, vRes;

      len = strlen( initParms->windowText );
      CHECKED_MALLOC( len + 1, tmpString );
      strcpy( tmpString, initParms->windowText );
      WinSetWindowPtr( win, 0, tmpString );
      
      psSize.cx = 0;
      psSize.cy = 0;
      windowDC = WinOpenWindowDC( win );
      hps = GpiCreatePS( WinQueryAnchorBlock( win ), windowDC, &psSize,
       PU_PELS | GPIF_DEFAULT | GPIT_NORMAL | GPIA_ASSOC );
      WinSetWindowULong( win, 4, hps );
      
      DevQueryCaps( windowDC, CAPS_HORIZONTAL_FONT_RES, 1, &hRes );
      DevQueryCaps( windowDC, CAPS_VERTICAL_FONT_RES,   1, &vRes );
      
      tmpLong = 0;
      numFonts = GpiQueryFonts( hps, QF_PUBLIC, theFont, &tmpLong, 0, NULL );
      
      if ( numFonts <= 0 )
      {
        // Fall back to Helv
        theFont = "Helvetica";
        tmpLong = 0;
        numFonts = GpiQueryFonts( hps, QF_PUBLIC, theFont, &tmpLong,
         0, NULL );
      }
      
      if ( numFonts > 0 )
      {
        CHECKED_MALLOC( numFonts * sizeof( FONTMETRICS ), pfm );
        
        GpiQueryFonts( hps, QF_PUBLIC, theFont, &numFonts,
         sizeof( FONTMETRICS ), pfm );
         
        for ( i=0; i<numFonts; ++i )
        {
          if ( (pfm[i].fsDefn & FM_DEFN_OUTLINE) ||
                (pfm[i].sNominalPointSize < 100 &&
                 pfm[i].sXDeviceRes == hRes &&
                 pfm[i].sYDeviceRes == vRes &&
                 (pfm[i].fsDefn & FM_DEFN_GENERIC)) )
          {
            fat.usRecordLength = sizeof( FATTRS );
            fat.fsSelection = FATTR_SEL_BOLD;
            fat.lMatch = pfm[i].lMatch;
            strcpy( fat.szFacename, theFont );
            
            GpiCreateLogFont( hps, NULL, 1, &fat );
            GpiSetCharSet( hps, 1 );
            
            if ( pfm[i].fsDefn & FM_DEFN_OUTLINE )
            {
              SIZEF sizef;
              fat.fsSelection = 0;
              sizef.cx = (65536 * hRes * 8) / 72;
              sizef.cy = (65536 * vRes * 8) / 72;
              GpiSetCharBox( hps, &sizef );
            }
            
            break;
          }
        }
        CHECKED_FREE( pfm );
      }
      
      len = strlen( tmpString );
      maxSize = 0;
      for ( i=0; i<len; ++i )
      {
        if ( tmpString[i] != '\n' ) maxSize++;
          else maxSize = 0;
        
        if ( maxSize > 39 && tmpString[i] == ' ' )
        {
          tmpString[i] = '\n';
          maxSize = 0;
        }
      }
      // Preprocess by trying to split lines longer than 40 characters
      
      numLines = 0;
      strPos = strchr( tmpString, '\n' );
      maxSize = 0;
      oldPos = tmpString;
      
      while ( strPos )
      {
        *strPos = 0;
        if ( *oldPos )
        {
          GpiQueryTextBox( hps, strlen( oldPos ), oldPos, 3, (POINTL *)textBox );
          if ( textBox[2].x - textBox[0].x + 1 > maxSize )
            maxSize = textBox[2].x - textBox[0].x + 1;
          numLines++;
        }
        *strPos = '\n';
        oldPos = strPos + 1;
        strPos = strchr( oldPos, '\n' );
      }
      
      if ( *oldPos )
      {
        GpiQueryTextBox( hps, strlen( oldPos ), oldPos, 3, (POINTL *)textBox );
        if ( textBox[2].x - textBox[0].x + 1 > maxSize )
          maxSize = textBox[2].x - textBox[0].x + 1;
        numLines++;
      }
      
      WinSetWindowULong( win, 20, textBox[0].y - textBox[1].y + 1 );
      
      cx = maxSize + 16;
      cy = ((textBox[0].y - textBox[1].y + 1) * numLines) + 16;
      // Add 8 pixel breathing room on each side of the window
      
      memDC = DevOpenDC( WinQueryAnchorBlock( win ), OD_MEMORY, "*", 5L,
       (PDEVOPENDATA) &dop, 0 );
      hpsMem = GpiCreatePS( WinQueryAnchorBlock( win ), memDC, &sizl,
       PU_PELS | GPIA_ASSOC );
      
      bmpInfo.cbFix = sizeof( BITMAPINFOHEADER2 );
      bmpInfo.cx = cx;
      bmpInfo.cy = cy;
      bmpInfo.cPlanes = 1;
      bmpInfo.cBitCount = 24;
      backgroundBmp = GpiCreateBitmap( hps, &bmpInfo, 0, NULL, NULL );
      
      GpiSetBitmap( hpsMem, backgroundBmp );

      WinSetWindowULong( win, 8, memDC );
      WinSetWindowULong( win, 12, hpsMem );
      WinSetWindowULong( win, 16, backgroundBmp );
      
      WinQueryWindowPos( win, &winPos );
      WinSetWindowPos( win, 0, winPos.x, winPos.y - cy, cx - 1, cy - 1,
       SWP_MOVE | SWP_SIZE | SWP_SHOW );
    }
    break;
    case WM_MOVE:
    {
      // Grab the bitmap from our new location if something forced the bubble
      //  to move (like SNAP's multihead window centering code)
      HPS hps = WinGetPS( win );
      HPS hpsMem = WinQueryWindowULong( win, 12 );
      POINTL bmpPts[3];
      unsigned char *bmpData;
      int cx, cy, i;
      RECTL rect;
      BITMAPINFO2 bmpInf2 = { 0 };
      
      WinQueryWindowRect( win, &rect );
      cx = rect.xRight - rect.xLeft + 1;
      cy = rect.yTop - rect.yBottom + 1;
      
      bmpPts[0].x = 0;      bmpPts[0].y = 0;
      bmpPts[1].x = cx - 1; bmpPts[1].y = cy - 1;
      bmpPts[2].x = 0;      bmpPts[2].y = 0;
      GpiBitBlt( hpsMem, hps, 3, (PPOINTL) bmpPts, ROP_SRCCOPY, BBO_IGNORE );
      
      CHECKED_MALLOC( cx * (cy+1) * 3, bmpData );

      bmpInf2.cbFix = 16;
      bmpInf2.cx = 0;
      bmpInf2.cy = 0;
      bmpInf2.cPlanes = 1;
      bmpInf2.cBitCount = 24;
      if ( GpiQueryBitmapBits( hpsMem, 0, cy - 1, (char *)bmpData, &bmpInf2 ) != cy - 1 )
      {
        debugf( "Error querying bitmap bits: 0x%lx\n",
         WinGetLastError( WinQueryAnchorBlock( win ) ) );
      }
      
      for ( i=0; i<cx*(cy+1)*3; ++i ) bmpData[i] = 0x7f + (bmpData[i] / 4);
      
      bmpInf2.cbFix = 16;
      
      if ( GpiSetBitmapBits( hpsMem, 0, cy - 1, (char *)bmpData, &bmpInf2 ) != cy - 1 )
      {
        debugf( "Error setting bitmap bits: 0x%lx\n",
         WinGetLastError( WinQueryAnchorBlock( win ) ) );
      }
      WinReleasePS( hps );
      CHECKED_FREE( bmpData );
      WinInvalidateRect( win, NULL, TRUE );
    }
    break;
    case WM_DESTROY:
    {
      HPS hps = WinQueryWindowULong( win, 4 );
      HPS hpsMem = WinQueryWindowULong( win, 12 );
      HBITMAP hbm = WinQueryWindowULong( win, 16 );
      HDC memDC = WinQueryWindowULong( win, 8 );
      char *windowText = (char *) WinQueryWindowPtr( win, 0 );
      if ( windowText )
      {
        CHECKED_FREE( windowText );
        WinSetWindowPtr( win, 16, NULL );
      }
      GpiSetBitmap( hpsMem, 0 );
      GpiDeleteBitmap( hbm );
      GpiDestroyPS( hpsMem );
      DevCloseDC( memDC );
      GpiDeleteSetId( hps, 1 );
      GpiDestroyPS( hps );
    }
    break;
    case WM_PAINT:
    {
      HPS hps = WinQueryWindowULong( win, 4 );
      HPS hpsMem = WinQueryWindowULong( win, 12 );
      HBITMAP hbm = WinQueryWindowULong( win, 16 );
      char *windowText = (char *) WinQueryWindowPtr( win, 0 );
      char *strPos, *oldPos;
      POINTL bmpPts[3];
      POINTL pt = { 8, 8 };
      RECTL updateRect;

      hps = WinBeginPaint( win, hps, &updateRect );
      bmpPts[0].x = updateRect.xLeft;      bmpPts[0].y = updateRect.yBottom;
      bmpPts[1].x = updateRect.xRight;     bmpPts[1].y = updateRect.yTop;
      bmpPts[2].x = updateRect.xLeft;      bmpPts[2].y = updateRect.yBottom;
      GpiBitBlt( hps, hpsMem, 3, (PPOINTL) bmpPts, ROP_SRCCOPY, BBO_IGNORE );
      
      oldPos = NULL;
      strPos = strrchr( windowText, '\n' );
      if ( strPos )
      {
        while ( strPos )
        {
          if ( *(strPos + 1) )
          {
            GpiCharStringPosAt( hps, &pt, &updateRect, CHS_CLIP,
             strlen( strPos + 1 ), strPos + 1, NULL );
            pt.y += WinQueryWindowULong( win, 20 );
          }
          if ( oldPos ) *oldPos = '\n';
          *strPos = 0;
          oldPos = strPos;
          strPos = strrchr( windowText, '\n' );
        }
      }
      
      if ( *windowText )
        GpiCharStringPosAt( hps, &pt, &updateRect, CHS_CLIP,
         strlen( windowText ), windowText, NULL );
      
      if ( oldPos ) *oldPos = '\n';
      WinEndPaint( hps );
    }
    break;
  }
  return WinDefWindowProc( win, msg, mp1, mp2 );
}

MRESULT EXPENTRY pmBubbleHelpCapableWindow( HWND win, ULONG msg, MPARAM mp1,
 MPARAM mp2 )
{
  MRESULT EXPENTRY (*oldProc) ( HWND, ULONG, MPARAM, MPARAM );
  HWND parentWin = WinQueryWindow( win, QW_PARENT );
  
  switch ( msg )
  {
    case WM_TIMER:
    {
      // Time to pop up some bubble information if appropriate
      OscarSession *theSession = (OscarSession *)
       WinQueryWindowPtr( parentWin, 24 );
      Buddy **buddyList;
      RECORDCORE *record;
      RECTL recordRect;
      QUERYRECORDRECT rectQuery;
      POINTL pointerPos, pointerPosAbs;
      int i, numBuddies;
      
      if ( SHORT1FROMMP( mp1 ) != 1 )
        break;
      
      WinStopTimer( WinQueryAnchorBlock( win ), win, 1 );
      buddyList = theSession->getBuddyList( &numBuddies );
      rectQuery.cb = sizeof( QUERYRECORDRECT );
      rectQuery.fRightSplitWindow = 0;
      rectQuery.fsExtent = CMA_TREEICON | CMA_ICON | CMA_TEXT;
      WinQueryPointerPos( HWND_DESKTOP, &pointerPosAbs );
      pointerPos.x = pointerPosAbs.x;
      pointerPos.y = pointerPosAbs.y;
      WinMapWindowPoints( HWND_DESKTOP, win, &pointerPos, 1 );
      for ( i=0; i<numBuddies; ++i )
      {
        record = buddyList[i]->record;
        if ( !record ) continue;
        rectQuery.pRecord = record;
        
        WinSendMsg( win, CM_QUERYRECORDRECT, MPFROMP( &recordRect ),
         MPFROMP( &rectQuery ) );
        
        if ( pointerPos.x >= recordRect.xLeft &&
              pointerPos.x <= recordRect.xRight &&
              pointerPos.y >= recordRect.yBottom &&
              pointerPos.y <= recordRect.yTop )
        {
          // We have a hit!
          HWND bubbleWin = WinQueryWindowULong( parentWin, 36 );
          HPOINTER folderIcon = WinQueryWindowULong( parentWin, 12 );
          bubbleWindowInitStruct bubbleInit;
          
          if ( bubbleWin ) WinDestroyWindow( bubbleWin );
          
          debugf( "Pointer is hovering over record: %s.\n", record->pszTree );
          
          WinSetWindowPtr( parentWin, 44, record );
          
          if ( record->hptrIcon == folderIcon )
          {
            CHECKED_MALLOC( 19 + strlen( record->pszTree ),
             bubbleInit.windowText );
            sprintf( bubbleInit.windowText, "Buddy Group Name: %s",
             record->pszTree );
          } else {
            bubbleInit.windowText = buddyList[i]->userData.getUserBlurbString();
          }
          
          WinSetCapture( HWND_DESKTOP, win );
          
          bubbleWin = WinCreateWindow( HWND_DESKTOP, "MrMessage Bubble Window",
           NULL, 0, pointerPosAbs.x, pointerPosAbs.y - 10, 0, 0, win,
           HWND_TOP, 9000, &bubbleInit, NULL );
          
          CHECKED_FREE( bubbleInit.windowText );
          
          WinSetWindowULong( parentWin, 36, bubbleWin );
          
          break;
        }
      }
    }
    break;
    case WM_MOUSEMOVE:
    {
      static int oldX = -1000000;
      static int oldY = -1000000;
      int newX, newY;
      
      newX = SHORT1FROMMP( mp1 );
      newY = SHORT2FROMMP( mp1 );
      
      if ( oldX != newX || oldY != newY )
      {
        HAB ab = WinQueryAnchorBlock( win );
        HWND bubbleWin;
        RECTL recordRect;
        QUERYRECORDRECT rectQuery;
        RECORDCORE *record = (RECORDCORE *) WinQueryWindowPtr( parentWin, 44 );
        
        oldX = newX;
        oldY = newY;
        
        WinStopTimer( ab, win, 1 );
        
        if ( !record )
        {
          WinStartTimer( ab, win, 1, BUBBLE_POPUP_TIMEOUT );
          break;
        }
        
        rectQuery.cb = sizeof( QUERYRECORDRECT );
        rectQuery.fRightSplitWindow = 0;
        rectQuery.fsExtent = CMA_TREEICON | CMA_ICON | CMA_TEXT;
        rectQuery.pRecord = record;
        WinSendMsg( win, CM_QUERYRECORDRECT, MPFROMP( &recordRect ),
         MPFROMP( &rectQuery ) );
         
        if ( newX < recordRect.xLeft || newX > recordRect.xRight ||
              newY < recordRect.yBottom || newY > recordRect.yTop )
        {
          WinSetCapture( HWND_DESKTOP, NULLHANDLE );
          
          bubbleWin = WinQueryWindowULong( parentWin, 36 );
          if ( bubbleWin )
          {
            WinDestroyWindow( bubbleWin );
            WinSetWindowULong( parentWin, 36, 0 );
            WinSetWindowPtr( parentWin, 44, NULL );
          }
          WinStartTimer( ab, win, 1, BUBBLE_POPUP_TIMEOUT );
        }
      }
      WinSendMsg( parentWin, WM_NOTIDLE, 0, 0 );
    }
    break;
    case WM_CHAR:
      WinSendMsg( parentWin, WM_NOTIDLE, 0, 0 );
    break;
    case WM_DESTROY:
    {
      HWND bubbleWin = WinQueryWindowULong( parentWin, 36 );
      
      if ( bubbleWin ) WinDestroyWindow( bubbleWin );
    }
    break;
  }

  oldProc = (MRESULT EXPENTRY (*) ( HWND, ULONG, MPARAM, MPARAM ))
   WinQueryWindowPtr( parentWin, 32 );
  
  if ( oldProc )
    return oldProc( win, msg, mp1, mp2 );
  
  return NULL;
}

typedef struct
{
  int excludeTypes;
  Buddy **buddyList;
  int numBuddies;
} FilterInfo;

int APIENTRY buddyListFilter( RECORDCORE *record, FilterInfo *filterInf )
{
  int i, found;
  
  found = 0;
  for ( i=0; i<filterInf->numBuddies; ++i )
  {
    if ( !filterInf->buddyList[i] ) continue;
    
    if ( record && record == filterInf->buddyList[i]->record )
    {
      found = 1;
      
      if ( filterInf->buddyList[i]->isGroup ) return 1;
      // Never filter an entire group
      
      if ( (filterInf->excludeTypes & (2|8) ) &&
            filterInf->buddyList[i]->userData.isOnline )
      {
        // Exclude all online
        if ( !filterInf->buddyList[i]->animStep ||
              filterInf->buddyList[i]->oldFlashState != BUDDY_STATE_OFFLINE )
        {
          return 0;
        }
      }
      if ( (filterInf->excludeTypes & 1) &&
            !(filterInf->buddyList[i]->userData.isOnline) )
      {
        // Exclude all offline
        if ( !filterInf->buddyList[i]->animStep )
        {
          return 0;
        }
      }
      if ( (filterInf->excludeTypes & 4 ) &&
            ((filterInf->buddyList[i]->userData.userClass & 32) || 
             (filterInf->buddyList[i]->userData.userStatus & 1)) )
      {
        // Exclude all away
        if ( !filterInf->buddyList[i]->animStep )
        {
          return 0;
        }
      }
      break;
    }
  }
  
  return found;
}

MRESULT EXPENTRY pmBuddyList( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch ( msg )
  {
    case WM_NOTIDLE:
      WinStartTimer( WinQueryAnchorBlock( win ), win, 3,
       IM_IDLE_TIMEOUT * 1000 );
      if ( WinQueryWindowULong( win, 60 ) )
      {
        OscarSession *theSession = (OscarSession *)
         WinQueryWindowPtr( win, 24 );
        OscarData *idleData = new OscarData( theSession->getSeqNumPointer() );
        
        idleData->prepareSetIdle( 0 );
        theSession->queueForSend( idleData );
        WinSetWindowULong( win, 60, 0 );
        debugf( "Re-activated after being idle.\n" );
      }
      // If we have reported idle-ness, report that we are active again.
    break;
    case WM_TIMER:
    {
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr( win, 24 );
      Buddy **buddyList;
      Buddy *theBuddy;
      int i, numBuddies;
      
      // Timer ID 1 is reserved for the bubble help and is handled by the
      //  subclassed window procedure.
      
      if ( SHORT1FROMMP( mp1 ) == 2 )
      {
        // Initial timeout for initialization of entry status.
        
        WinStopTimer( WinQueryAnchorBlock( win ), win, 2 );
        buddyList = theSession->getBuddyList( &numBuddies );
        
        for ( i=0; i<numBuddies; ++i )
        {
          // Assume all records have been initialized after timeout so
          //  we can start playing audible notifications for users coming
          //  and going.
          theBuddy = buddyList[i];
          theBuddy->userData.statusUninit = 0;
        }
      } else if ( SHORT1FROMMP( mp1 ) == 3 )
      {
        // Idle timeout.  Only have to report idle-ness once.  The server
        //  will keep posting updates to listening clients.
        OscarData *idleData = new OscarData( theSession->getSeqNumPointer() );
        idleData->prepareSetIdle( IM_IDLE_TIMEOUT );
        theSession->queueForSend( idleData );
        WinStopTimer( WinQueryAnchorBlock( win ), win, 3 );
        WinSetWindowULong( win, 60, 1 );
        debugf( "Reporting idle status (idle for %d seconds).\n",
         IM_IDLE_TIMEOUT );
      } else if ( SHORT1FROMMP( mp1 ) == 4 )
      {
        // Flashing changes for icon status
        
        HPOINTER offlineIco = WinQueryWindowULong( win, 28 );
        HPOINTER onlineIco = WinQueryWindowULong( win, 20 );
        HPOINTER hotOnlineIco = WinQueryWindowULong( win, 40 );
        HPOINTER awayIco = WinQueryWindowULong( win, 8 );
        unsigned char theState;

        buddyList = theSession->getBuddyList( &numBuddies );
        
        for ( i=0; i<numBuddies; ++i )
        {
          // Assume all records have been initialized after timeout so
          //  we can start playing audible notifications for users coming
          //  and going.
          theBuddy = buddyList[i];
        
          if ( !theBuddy->record ) continue;
          // The PLEASEUPGRADE000 entry gets added to the buddy list but
          // doesn't have a record created for it in the container.
          
          if ( theBuddy->animStep )
          {
            if ( theBuddy->animStep % 2 )
            {
              theState = theBuddy->flashState;
            } else {
              theState = theBuddy->oldFlashState;
            }
            
            switch ( theState )
            {
              case BUDDY_STATE_ONLINE:
                theBuddy->record->hptrIcon = onlineIco;
              break;
              case BUDDY_STATE_AWAY:
                theBuddy->record->hptrIcon = awayIco;
              break;
              case BUDDY_STATE_ACTIVE:
                theBuddy->record->hptrIcon = hotOnlineIco;
              break;
              case BUDDY_STATE_OFFLINE:
                theBuddy->record->hptrIcon = offlineIco;
              break;
            }
            theBuddy->animStep--;
            
            WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INVALIDATERECORD,
             MPFROMP( &(theBuddy->record) ),
             MPFROM2SHORT( 1, CMA_NOREPOSITION | CMA_ERASE ) );
            
            if ( !theBuddy->animStep )
            {
              FilterInfo theFilter;
              
              theSession->numFlashing--;
              if ( !theSession->numFlashing )
              {
                WinStopTimer( WinQueryAnchorBlock( win ), win, 4 );
              }
              
              theFilter.excludeTypes = WinQueryWindowULong( win, 48 );
              theFilter.buddyList =
               theSession->getBuddyList( &theFilter.numBuddies );
              
              if ( theFilter.excludeTypes )
              {
                WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER,
                 MPFROMP( buddyListFilter ), MPFROMP( &theFilter ) );
              }
            }
          }
        }
      }
    }
    break;
    case WM_IMRECEIVED:
    {
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr( win, 24 );
      char *receivedFrom = (char *) PVOIDFROMMP( mp1 );
      char *theMessage = (char *) PVOIDFROMMP( mp2 );
      char *strippedMessage, *annotatedMessage;
      Buddy **buddyList;
      Buddy *theBuddy;
      int i, numBuddies, found, newWin;
      ULONG textLen;
      
      buddyList = theSession->getBuddyList( &numBuddies );
      
      for ( i=0; i<numBuddies; ++i )
      {
        theBuddy = buddyList[i];
        
        if ( !theBuddy->record ) continue;
        // The PLEASEUPGRADE000 entry gets added to the buddy list but doesn't
        //  have a record created for it in the container.
        
        if ( theBuddy->record->pszTree &&
              stricmp( theBuddy->record->pszTree, receivedFrom ) == 0 )
        {
          // Found a match in the buddy list.  So far so good.
          found = 1;
          newWin = 0;
          if ( !theBuddy->imChatWin )
          {
            // Chat window is not already up.
            NOTIFYRECORDENTER recordEnter;
            
            recordEnter.hwndCnr = WinWindowFromID( win, WID_SESSION_CONTAINER );
            recordEnter.fKey = TRUE;
            recordEnter.pRecord = theBuddy->record;
            
            WinSendMsg( win, WM_CONTROL, MPFROM2SHORT( WID_SESSION_CONTAINER,
             CN_ENTER ), MPFROMP( &recordEnter ) );
            
            newWin = 1;
          }
          
          // Chat window is up, so populate it with the received message.
          
          strippedMessage = stripTags( theMessage );
          
          debugf( "Stripped message: %s\n", strippedMessage );
          
          CHECKED_MALLOC( strlen( strippedMessage ) +
           strlen( theBuddy->record->pszTree ) + 9, annotatedMessage );
          
          sprintf( annotatedMessage, "[%s]: %s\n\n", theBuddy->record->pszTree,
           strippedMessage );
           
          CHECKED_FREE( strippedMessage );
          
          textLen = LONGFROMMR( WinSendDlgItemMsg( theBuddy->imChatClientWin,
           WID_IM_RECEIVED, MLM_QUERYTEXTLENGTH, NULL, NULL ) );
           
          WinSendDlgItemMsg( theBuddy->imChatClientWin, WID_IM_RECEIVED,
           MLM_SETSEL, MPFROMLONG( textLen ), MPFROMLONG( textLen ) );
          // Move the insertion point to the end of the text, in case the user
          //  clicked somewhere.
          
          WinSendDlgItemMsg( theBuddy->imChatClientWin, WID_IM_RECEIVED,
           MLM_INSERT, MPFROMP( annotatedMessage ), NULL );
          
          if ( !newWin )
          {
            if ( theSession->settings->settingFlags[ EVENT_RECEIVED ] &
                  SET_SOUND_ENABLED )
            {
              mciPlayFile( HWND_DESKTOP,
               theSession->settings->sounds[ EVENT_RECEIVED ], MCI_ASYNC, NULL,
               NULLHANDLE );
            }
          }
          
          CHECKED_FREE( annotatedMessage );
          
          break;
        }
      }
      
      if ( !found )
      {
        debugf( "IM received, but couldn't find a matching buddy entry for it.\n" );
        debugf( "Orphaned IM received from %s.  I'll fix this later.\n", receivedFrom );
      }
      
      CHECKED_FREE( receivedFrom );
      CHECKED_FREE( theMessage );
    }
    break;
    case WM_REGISTERWAKEUPSEM:
      WinSetWindowULong( win, 4, LONGFROMMP( mp1 ) );
      debugf( "Session wakeup semaphore was registered.\n" );
    break;
    case WM_ADDBUDDY:
    {
      RECORDINSERT recordInsert;
      buddyListEntry *buddyInfo = (buddyListEntry *) PVOIDFROMMP( mp1 );
      HEV wakeupSem = LONGFROMMP( mp2 );
      HPOINTER offlineIco = WinQueryWindowULong( win, 28 );
      HPOINTER folder = WinQueryWindowULong( win, 12 );
      HPOINTER onlineIco = WinQueryWindowULong( win, 20 );
      HPOINTER hotOnlineIco = WinQueryWindowULong( win, 40 );
      HPOINTER awayIco = WinQueryWindowULong( win, 8 );
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr( win, 24 );
      FilterInfo theFilter;
      
      debugf( "Requested to add buddy: %s\n", buddyInfo->entryName );
      
      buddyInfo->myRecord = (RECORDCORE *) PVOIDFROMMR( WinSendDlgItemMsg( win,
       WID_SESSION_CONTAINER, CM_ALLOCRECORD, MPFROMLONG( 0 ),
       MPFROMSHORT( 1 ) ) );
       
      buddyInfo->myRecord->flRecordAttr = CRA_EXPANDED;
      buddyInfo->myRecord->ptlIcon.x = 0;
      buddyInfo->myRecord->ptlIcon.y = 0;
      buddyInfo->myRecord->pszTree = buddyInfo->entryName;
      buddyInfo->myRecord->hptrMiniIcon = 0;
      buddyInfo->myRecord->hbmMiniBitmap = 0;
      buddyInfo->myRecord->hptrIcon = 0;
      buddyInfo->myRecord->hbmBitmap = 0;
      
      if ( buddyInfo->numMembers )
      {
        // This is a group, not a buddy.
        buddyInfo->myRecord->hptrIcon = folder;
      } else {
        buddyInfo->myRecord->hptrIcon = offlineIco;
      }
      
      buddyInfo->myRecord->pszTree = strdup( buddyInfo->entryName );
      buddyInfo->myRecord->pTreeItemDesc = NULL;
      
      recordInsert.cb = sizeof( RECORDINSERT );
      recordInsert.pRecordOrder = (RECORDCORE *)CMA_END;
      recordInsert.pRecordParent = buddyInfo->parentRecord;
      recordInsert.fInvalidateRecord = TRUE;
      recordInsert.zOrder = CMA_TOP;
      recordInsert.cRecordsInsert = 1;
      
      WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INSERTRECORD,
       MPFROMP( buddyInfo->myRecord ), MPFROMP( &recordInsert ) );
      
      theFilter.excludeTypes = WinQueryWindowULong( win, 48 );
      theFilter.buddyList = theSession->getBuddyList( &theFilter.numBuddies );
      
      if ( theFilter.excludeTypes )
      {
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER,
         MPFROMP( buddyListFilter ), MPFROMP( &theFilter ) );
      } else {
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER, NULL, NULL );
      }
      
      DosPostEventSem( wakeupSem );
    }
    break;
    case WM_BUDDYSTATUS:
    {
      UserInformation *userInfo = (UserInformation *) PVOIDFROMMP( mp1 );
      RECORDCORE *record;
      HPOINTER offlineIco = WinQueryWindowULong( win, 28 );
      HPOINTER onlineIco = WinQueryWindowULong( win, 20 );
      HPOINTER hotOnlineIco = WinQueryWindowULong( win, 40 );
      HPOINTER awayIco = WinQueryWindowULong( win, 8 );
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr( win, 24 );
      Buddy **buddyList;
      char *formattedName = NULL;
      int numBuddies, i, found, retry;
      unsigned char statusChange = 0, idleChange = 0, statusUninit = 0;
      FilterInfo theFilter;
      
      theFilter.excludeTypes = WinQueryWindowULong( win, 48 );
      
      buddyList = theSession->getBuddyList( &numBuddies );
      theFilter.buddyList = buddyList;
      theFilter.numBuddies = numBuddies;
      
      found = 0;
      retry = 0;
      
      lookForReformattedSN:
      
      for ( i=0; i<numBuddies; ++i )
      {
        record = buddyList[i]->record;
        if ( !record ) continue;
        if ( stricmp( userInfo->screenName, record->pszTree ) == 0 )
        {
          if ( strcmp( userInfo->screenName, record->pszTree ) != 0 )
          {
            // Capitalization changes.  Update the buddy list window.
            strcpy( record->pszTree, userInfo->screenName );
            WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INVALIDATERECORD,
             MPFROMP( &record ), MPFROM2SHORT( 1, CMA_TEXTCHANGED | CMA_ERASE ) );
          }
          
          if ( userInfo->isOnline != buddyList[i]->userData.isOnline ||
                userInfo->userStatus != buddyList[i]->userData.userStatus ||
                userInfo->userClass != buddyList[i]->userData.userClass )
            statusChange = 1;
          if ( userInfo->idleMinutes != buddyList[i]->userData.idleMinutes )
            idleChange = 1;
          if ( buddyList[i]->userData.statusUninit )
            statusUninit = 1;
          
          if ( statusChange || idleChange )
          {
            buddyList[i]->userData.incorporateData( *userInfo );
            if ( userInfo->isOnline )
            {
              if ( (userInfo->userClass & 32) || 
                    (userInfo->userStatus & 1) )
              {
                debugf( "%s is away.\n", record->pszTree );
                record->hptrIcon = awayIco;
                
                buddyList[i]->oldFlashState = buddyList[i]->flashState;
                buddyList[i]->flashState = BUDDY_STATE_AWAY;
                
                if ( statusChange )
                {
                  OscarData *awayReq =
                   new OscarData( theSession->getSeqNumPointer() );
                  awayReq->prepareRequestUserInfo( record->pszTree, 3 );
                  theSession->queueForSend( awayReq );
                }
              } else {
                debugf( "%s is on line.\n", record->pszTree );
                record->hptrIcon = onlineIco;
                
                buddyList[i]->oldFlashState = buddyList[i]->flashState;
                buddyList[i]->flashState = BUDDY_STATE_ONLINE;
                
                if ( buddyList[i]->userData.idleMinutes < 10 )
                {
                  debugf( "%s has been active in the last 10 minutes.\n",
                   record->pszTree );
                  record->hptrIcon = hotOnlineIco;
                  buddyList[i]->flashState = BUDDY_STATE_ACTIVE;
                }
                if ( statusChange )
                {
                  OscarData *profileReq =
                   new OscarData( theSession->getSeqNumPointer() );
                  profileReq->prepareRequestUserInfo( record->pszTree, 1 );
                  theSession->queueForSend( profileReq );
                  
                  if ( !statusUninit )
                  {
                    if ( theSession->settings->settingFlags[ EVENT_ARRIVE ] &
                          SET_SOUND_ENABLED )
                    {
                      mciPlayFile( HWND_DESKTOP,
                       theSession->settings->sounds[ EVENT_ARRIVE ],
                       MCI_ASYNC, NULL, NULLHANDLE );
                    }
                  }
                }
              }
              if ( buddyList[i]->imChatClientWin )
              {
                WinSendMsg( buddyList[i]->imChatClientWin, WM_BUDDYSTATUS,
                 MPFROMLONG( 1 ), NULL );
                // Same window message used elsewhere, but different params.
                //  1 indicates that the user is at least connected to AOL.
              }
            } else {
              debugf( "%s is off line.\n", record->pszTree );
              record->hptrIcon = offlineIco;
              
              buddyList[i]->oldFlashState = buddyList[i]->flashState;
              buddyList[i]->flashState = BUDDY_STATE_OFFLINE;
              
              if ( !statusUninit )
              {
                if ( theSession->settings->settingFlags[ EVENT_LEAVE ] &
                      SET_SOUND_ENABLED )
                {
                  mciPlayFile( HWND_DESKTOP,
                   theSession->settings->sounds[ EVENT_LEAVE ], MCI_ASYNC,
                   NULL, NULLHANDLE );
                }
              }
              
              if ( buddyList[i]->imChatClientWin )
              {
                WinSendMsg( buddyList[i]->imChatClientWin, WM_BUDDYSTATUS,
                 MPFROMLONG( 0 ), NULL );
                // Same window message used elsewhere, but different params.
                //  0 indicates that the user is offline.
              }
            }
            
            if ( buddyList[i]->flashState != buddyList[i]->oldFlashState )
            {
              if ( !buddyList[i]->animStep )
              {
                theSession->numFlashing++;
              }
              buddyList[i]->animStep = 8;
              
              WinStartTimer( WinQueryAnchorBlock( win ), win, 4, 500 );
            }
          } else {
            buddyList[i]->userData.incorporateData( *userInfo );
            debugf( "%s user status has not changed.\n", record->pszTree );
          }
          
          WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INVALIDATERECORD,
           MPFROMP( &record ), MPFROM2SHORT( 1, CMA_NOREPOSITION | CMA_ERASE ) );
          
          if ( buddyList[i]->imChatWin )
          {
            WinSendMsg( buddyList[i]->imChatWin, WM_SETICON,
             MPFROMLONG( record->hptrIcon ), NULL );
          }
          found = 1;
        }
      }  
      
      if ( !found && !retry )
      {
        int len;
        int cpos = 0;
        
        formattedName = userInfo->screenName;
        len = strlen( formattedName );
        
        CHECKED_MALLOC( len + 1, userInfo->screenName );
        
        for ( i=0; i<len; ++i )
        {
          if ( formattedName[i] != ' ' )
          {
            userInfo->screenName[cpos] = formattedName[i];
            cpos++;
          }
        }
        userInfo->screenName[cpos] = 0;
        retry = 1;
        
        goto lookForReformattedSN;
      } else if ( retry )
      {
        if ( found )
        {
          for ( i=0; i<numBuddies; ++i )
          {
            record = buddyList[i]->record;
            if ( !record ) continue;
            if ( stricmp( userInfo->screenName, record->pszTree ) == 0 )
            {
              record->pszTree = formattedName;
              
              WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INVALIDATERECORD,
               MPFROMP( &record ), MPFROM2SHORT( 1, CMA_TEXTCHANGED | CMA_ERASE ) );
              buddyList[i]->userData.screenName = strdup( formattedName );
            }
          }
          debugf( "Screen name '%s' was updated to be '%s'.\n",
           userInfo->screenName, formattedName );
        } else {
          CHECKED_FREE( userInfo->screenName );
          userInfo->screenName = formattedName;
          debugf( "Received user status for %s (not in buddy list).\n",
           formattedName );
        }
      }
      if ( found )
      {
        if ( theFilter.excludeTypes )
        {
          WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER,
           MPFROMP( buddyListFilter ), MPFROMP( &theFilter ) );
        } else {
          WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER,
           NULL, NULL );
        }
      }
      delete userInfo;
    }
    break;
    case WM_CREATE:
    {
      HWND theWin;
      CNRINFO cnrInfo;
      HMODULE dll = getModuleHandle();
      HPOINTER onlineIco = WinLoadPointer( HWND_DESKTOP, dll, 3 );
      HPOINTER awayIco = WinLoadPointer( HWND_DESKTOP, dll, 4 );
      HPOINTER offlineIco = WinLoadPointer( HWND_DESKTOP, dll, 5 );
      HPOINTER folder = WinLoadPointer( HWND_DESKTOP, dll, 2 );
      HPOINTER hotOnlineIco = WinLoadPointer( HWND_DESKTOP, dll, 6 );
      HPOINTER bigX = WinLoadPointer( HWND_DESKTOP, dll, 7 );
      BTNCDATA ctrlData;
      
      WinStartTimer( WinQueryAnchorBlock( win ), win, 2, 20000 );
      // Timeout to assume that all initial user updates have arrived to
      //  start audible notifications.
      
      WinSetWindowULong( win, 28, offlineIco );
      WinSetWindowULong( win, 12, folder );
      WinSetWindowULong( win, 20, onlineIco );
      WinSetWindowULong( win, 8, awayIco );
      WinSetWindowULong( win, 32, 0 );
      WinSetWindowULong( win, 36, 0 );
      WinSetWindowULong( win, 40, hotOnlineIco );
      WinSetWindowULong( win, 48, 0 );
      WinSetWindowULong( win, 52, bigX );
      
      theWin = WinCreateWindow( win, WC_CONTAINER, "Mr. Message Buddy List",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | CCS_AUTOPOSITION | CCS_EXTENDSEL |
       CCS_MINIICONS, 0, 0, 0, 0, win, HWND_TOP, WID_SESSION_CONTAINER, NULL,
       NULL );
       
      cnrInfo.cb = sizeof( CNRINFO );
      cnrInfo.pSortRecord = NULL;
      cnrInfo.pFieldInfoLast = NULL;
      cnrInfo.pFieldInfoObject = NULL;
      cnrInfo.pszCnrTitle = NULL;
      cnrInfo.flWindowAttr = CV_TREE | CV_ICON | CV_MINI | CA_DRAWICON |
       CA_TREELINE;
      cnrInfo.ptlOrigin.x = 0;
      cnrInfo.ptlOrigin.y = 0;
      cnrInfo.cDelta = 0;
      cnrInfo.cRecords = 0;
      cnrInfo.slBitmapOrIcon.cx = 0;
      cnrInfo.slBitmapOrIcon.cy = 0;
      cnrInfo.slTreeBitmapOrIcon.cx = 0;
      cnrInfo.slTreeBitmapOrIcon.cy = 0;
      cnrInfo.hbmExpanded = 0;
      cnrInfo.hbmCollapsed = 0;
      cnrInfo.hptrExpanded = 0;
      cnrInfo.hptrCollapsed = 0;
      cnrInfo.cyLineSpacing = -1;
      cnrInfo.cxTreeIndent = -1;
      cnrInfo.cxTreeLine = -1;
      cnrInfo.cFields = 0;
      cnrInfo.xVertSplitbar = -1;
      WinSendMsg( theWin, CM_SETCNRINFO, MPFROMP( &cnrInfo ),
       MPFROMLONG( CMA_FLWINDOWATTR ) );
      
      WinSetWindowPtr( win, 32, WinSubclassWindow( theWin,
       pmBubbleHelpCapableWindow ) );
      
      ctrlData.cb = sizeof( BTNCDATA );
      ctrlData.fsCheckState = 0;
      ctrlData.fsHiliteState = 0;
      ctrlData.hImage = onlineIco;
      
      theWin = WinCreateWindow( win, WC_BUTTON, "",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON | BS_MINIICON |
        BS_AUTOSIZE, 0, 0, 0, 0, win, HWND_TOP, WID_FILTER_ONLINE,
        &ctrlData, NULL );
      
      ctrlData.hImage = awayIco;
      theWin = WinCreateWindow( win, WC_BUTTON, "",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON | BS_MINIICON |
        BS_AUTOSIZE, 0, 0, 0, 0, win, HWND_TOP, WID_FILTER_AWAY,
        &ctrlData, NULL );
      
      ctrlData.hImage = offlineIco;
      theWin = WinCreateWindow( win, WC_BUTTON, "",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON | BS_MINIICON |
        BS_AUTOSIZE, 0, 0, 0, 0, win, HWND_TOP, WID_FILTER_OFFLINE,
        &ctrlData, NULL );
    }
    break;
    case WM_COMMAND:
    {
      HMODULE dll = getModuleHandle();
      int filterTypes = WinQueryWindowULong( win, 48 );
      HPOINTER bigX = WinLoadPointer( HWND_DESKTOP, dll, 7 );
      HPOINTER onlineIco = WinLoadPointer( HWND_DESKTOP, dll, 3 );
      HPOINTER awayIco = WinLoadPointer( HWND_DESKTOP, dll, 4 );
      HPOINTER offlineIco = WinLoadPointer( HWND_DESKTOP, dll, 5 );
      HPOINTER hotOnlineIco = WinLoadPointer( HWND_DESKTOP, dll, 6 );
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr( win, 24 );
      BTNCDATA ctrlData;
      WNDPARAMS wndParms;
      
      ctrlData.cb = sizeof( BTNCDATA );
      ctrlData.fsCheckState = 0;
      ctrlData.fsHiliteState = 0;
      wndParms.fsStatus = WPM_CBCTLDATA | WPM_CTLDATA;
      wndParms.cbCtlData = sizeof( BTNCDATA );
      wndParms.pCtlData = &ctrlData;
      
      switch ( SHORT1FROMMP( mp1 ) )
      {
        case WID_FILTER_ONLINE:
          filterTypes = filterTypes ^ (2|8);
          if ( filterTypes & (2|8) )
          {
            ctrlData.hImage = bigX;
          } else {
            ctrlData.hImage = onlineIco;
          }
          WinSendDlgItemMsg( win, WID_FILTER_ONLINE, WM_SETWINDOWPARAMS,
           MPFROMP( &wndParms ), NULL );
          WinSetWindowULong( win, 48, filterTypes );
        break;
        case WID_FILTER_AWAY:
          filterTypes = filterTypes ^ 4;
          if ( filterTypes & 4 )
          {
            ctrlData.hImage = bigX;
          } else {
            ctrlData.hImage = awayIco;
          }
          WinSendDlgItemMsg( win, WID_FILTER_AWAY, WM_SETWINDOWPARAMS,
           MPFROMP( &wndParms ), NULL );
          WinSetWindowULong( win, 48, filterTypes );
        break;
        case WID_FILTER_OFFLINE:
          filterTypes = filterTypes ^ 1;
          if ( filterTypes & 1 )
          {
            ctrlData.hImage = bigX;
          } else {
            ctrlData.hImage = offlineIco;
          }
          WinSendDlgItemMsg( win, WID_FILTER_OFFLINE, WM_SETWINDOWPARAMS,
           MPFROMP( &wndParms ), NULL );
          WinSetWindowULong( win, 48, filterTypes );
        break;
      }
      
      if ( filterTypes )
      {
        FilterInfo theFilter;
        
        theFilter.excludeTypes = filterTypes;
        theFilter.buddyList = theSession->getBuddyList( &theFilter.numBuddies );
        
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER,
         MPFROMP( buddyListFilter ), MPFROMP( &theFilter ) );
      } else {
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_FILTER, NULL, NULL );
      }
    }
    break;
    case WM_SIZE:
    {
      USHORT newCx = SHORT1FROMMP( mp2 );
      USHORT newCy = SHORT2FROMMP( mp2 );
      HWND theWin;
      RECTL bRec;
      int bHeight;
      
      theWin = WinWindowFromID( win, WID_FILTER_ONLINE );
      WinQueryWindowRect( theWin, &bRec );
      bHeight = bRec.yTop - bRec.yBottom;
      
      WinSetWindowPos( theWin, 0, 0, 2, (newCx - 8) / 3, bHeight,
       SWP_SIZE | SWP_MOVE );
      theWin = WinWindowFromID( win, WID_FILTER_AWAY );
      WinSetWindowPos( theWin, 0, 4 + ((newCx - 8) / 3), 2,
       (newCx - 8) / 3, bHeight, SWP_SIZE | SWP_MOVE );
      theWin = WinWindowFromID( win, WID_FILTER_OFFLINE );
      WinSetWindowPos( theWin, 0, 8 + (2 * ((newCx - 8) / 3)), 2,
       (newCx - 8) / 3, bHeight, SWP_SIZE | SWP_MOVE );
      
      theWin = WinWindowFromID( win, WID_SESSION_CONTAINER );
      WinSetWindowPos( theWin, 0, 0, bHeight + 4, newCx, newCy - (bHeight + 4),
       SWP_SIZE | SWP_MOVE );
    }
    break;
    case WM_PAINT:
    {
      HPS hps;
      RECTL rect;
      POINTL pt;
      ULONG colorIdx, ret;
      
      ret = WinQueryPresParam( win, PP_BACKGROUNDCOLORINDEX, 0, NULL, 4,
       &colorIdx, QPF_ID1COLORINDEX | QPF_NOINHERIT );
      if ( !ret )
      {
        ret = WinQueryPresParam( win, PP_BACKGROUNDCOLOR, 0, NULL, 4,
         &colorIdx, QPF_NOINHERIT );
        if ( !ret )
        {
          ret = WinQueryPresParam( win, PP_BACKGROUNDCOLORINDEX, 0, NULL, 4,
           &colorIdx, QPF_ID1COLORINDEX );
        }
      }
      
      hps = WinBeginPaint( win, NULLHANDLE, &rect );
      if ( ret < 4 )
        GpiSetColor( hps, CLR_PALEGRAY );
      else {
        GpiCreateLogColorTable( hps, 0, LCOLF_RGB, 0, 0, NULL );
        GpiSetColor( hps, colorIdx );
      }
      pt.x = rect.xLeft;
      pt.y = rect.yBottom;
      GpiMove( hps, &pt );
      pt.x = rect.xRight;
      pt.y = rect.yTop;
      GpiBox( hps, DRO_FILL, &pt, 0, 0 );
      WinEndPaint( hps );
    }
    break;
    case WM_INITULONGS:
    {
      MINIRECORDCORE *myRecord = (MINIRECORDCORE *) PVOIDFROMMP( mp1 );
      HWND mgrWin = LONGFROMMP( mp2 );
      HINI iniFile = WinQueryWindowULong( mgrWin, 20 );
      char *appName;
      ULONG tmpLong, tmpLong2;
      
      CHECKED_MALLOC( 14 + strlen( myRecord->pszIcon ), appName );
      sprintf( appName, "Buddy List [%s]", myRecord->pszIcon );

      tmpLong = 0;
      tmpLong2 = sizeof( ULONG );
      PrfQueryProfileData( iniFile, appName, "Filtration", &tmpLong,
       &tmpLong2 );
      // The initial filter settings
      
      CHECKED_FREE( appName );
      
      if ( tmpLong & (2|8) )
        WinSendMsg( win, WM_COMMAND, MPFROMSHORT( WID_FILTER_ONLINE ), NULL );
      
      if ( tmpLong & 1 )
        WinSendMsg( win, WM_COMMAND, MPFROMSHORT( WID_FILTER_OFFLINE ), NULL );

      if ( tmpLong & 4 )
        WinSendMsg( win, WM_COMMAND, MPFROMSHORT( WID_FILTER_AWAY ), NULL );

      WinShowWindow( WinQueryWindow( win, QW_PARENT ), TRUE );
      WinSetFocus( HWND_DESKTOP,
       WinWindowFromID( win, WID_SESSION_CONTAINER ) );
      WinSetWindowPtr( win, 0, myRecord );
      // My record in the session manager window
      WinSetWindowULong( win, 16, mgrWin );
      // Window handle to session manager
    }
    break;
    case WM_CLOSE:
    {
      RECORDCORE *record;
      MINIRECORDCORE *myRecord = (MINIRECORDCORE *) WinQueryWindowPtr( win, 0 );
      HEV wakeupSem = WinQueryWindowULong( win, 4 );
      HPOINTER offlineIco = WinQueryWindowULong( win, 28 );
      HPOINTER folder = WinQueryWindowULong( win, 12 );
      HPOINTER onlineIco = WinQueryWindowULong( win, 20 );
      HPOINTER awayIco = WinQueryWindowULong( win, 8 );
      HPOINTER hotOnlineIco = WinQueryWindowULong( win, 40 );
      HPOINTER bigX = WinQueryWindowULong( win, 52 );
      HWND mgrWin = WinQueryWindowULong( win, 16 );
      HINI iniFile = WinQueryWindowULong( mgrWin, 20 );
      HWND frameWin = WinQueryWindow( win, QW_PARENT );
      OscarSession *theSession = (OscarSession *) WinQueryWindowPtr( win, 24 );
      Buddy **buddyList;
      int numBuddies, i;
      SWP swp;
      char *appName;
      ULONG tmpLong;
      
      if ( theSession->settings->settingFlags[ EVENT_ENDSESSION ] &
            SET_SOUND_ENABLED )
      {
        mciPlayFile( HWND_DESKTOP,
         theSession->settings->sounds[ EVENT_ENDSESSION ], MCI_ASYNC, NULL,
         NULLHANDLE );
      }
      
      buddyList = theSession->getBuddyList( &numBuddies );
      for ( i=0; i<numBuddies; ++i )
      {
        if ( buddyList[i]->imChatWin )
        {
          WinSendMsg( buddyList[i]->imChatWin, WM_CLOSE, 0, 0 );
        }
      }
      
      CHECKED_MALLOC( 14 + strlen( myRecord->pszIcon ), appName );
      sprintf( appName, "Buddy List [%s]", myRecord->pszIcon );
      
      WinQueryWindowPos( frameWin, &swp );
      if ( (swp.fl & SWP_MINIMIZE) || (swp.fl & SWP_MAXIMIZE) )
      {
        swp.x = WinQueryWindowUShort( frameWin, QWS_XRESTORE );
        swp.y = WinQueryWindowUShort( frameWin, QWS_YRESTORE );
        swp.cx = WinQueryWindowUShort( frameWin, QWS_CXRESTORE );
        swp.cy = WinQueryWindowUShort( frameWin, QWS_CYRESTORE );
      }
      PrfWriteProfileData( iniFile, appName, "Window position data",
       &swp, sizeof( SWP ) );
      
      tmpLong = WinQueryWindowULong( win, 48 );
      PrfWriteProfileData( iniFile, appName, "Filtration", &tmpLong,
       sizeof( ULONG ) );
      
      CHECKED_FREE( appName );
      
      debugf( "Deleting buddies from buddy list container.\n" );
      do {
        record = (RECORDCORE *) WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
         CM_QUERYRECORD, NULL, MPFROM2SHORT( CMA_LAST, CMA_ITEMORDER ) );
        if ( !record || (LONG)record == -1 ) break;
        
        CHECKED_FREE( record->pszTree );
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_REMOVERECORD,
         MPFROMP( &record ), MPFROM2SHORT( 1, CMA_FREE ) );
      } while ( 1 );
      
      WinDestroyPointer( offlineIco );
      WinDestroyPointer( onlineIco );
      WinDestroyPointer( hotOnlineIco );
      WinDestroyPointer( awayIco );
      WinDestroyPointer( folder );
      WinDestroyPointer( bigX );
      
      debugf( "Closed buddy list window.\n" );
      WinDestroyWindow( WinWindowFromID( win, WID_SESSION_CONTAINER ) );
      WinDestroyWindow( WinWindowFromID( win, WID_FILTER_ONLINE ) );
      WinDestroyWindow( WinWindowFromID( win, WID_FILTER_OFFLINE ) );
      WinDestroyWindow( WinWindowFromID( win, WID_FILTER_AWAY ) );
      WinDestroyWindow( WinQueryWindow( win, QW_PARENT ) );
      // Kill this frame window
      
      debugf( "Waking up session thread.\n" );
      DosPostEventSem( wakeupSem );
      // Tell the session object to pack it in
      
      WinSendMsg( mgrWin, WM_SESSIONCLOSED, MPFROMP( myRecord ), 0 );
      
      return 0;
      // Prevent the default window procedure from sending a WM_QUIT to the
      //  message queue and shutting down the whole app.
    }
    case WM_CONTROL:
    {
      if ( SHORT1FROMMP( mp1 ) == WID_SESSION_CONTAINER )
      {
        switch ( SHORT2FROMMP( mp1 ) )
        {
          case CN_ENTER:
          {
            // A record was double-clicked or enter was pressed on one.
            // Open an IM window.
            
            NOTIFYRECORDENTER *notify = (NOTIFYRECORDENTER *)
             PVOIDFROMMP( mp2 );
            ULONG fcf = FCF_TITLEBAR | FCF_MINMAX | FCF_SYSMENU | FCF_TASKLIST |
                        FCF_SIZEBORDER | FCF_ICON | FCF_NOMOVEWITHOWNER;
            HMODULE dll = getModuleHandle();
            HWND mgrWin = WinQueryWindowULong( win, 16 );
            HINI iniFile = WinQueryWindowULong( mgrWin, 20 );
            HPOINTER folder = WinQueryWindowULong( win, 12 );
            OscarSession *theSession = (OscarSession *)
             WinQueryWindowPtr( win, 24 );
            char *text;
            HWND imFrame, imClient;
            RECTL desktop;
            SWP swp;
            ULONG tmpLong;
            Buddy **buddyList;
            Buddy *theBuddy;
            int i, numBuddies;
            
            if ( !notify->pRecord ) break;
            
            theBuddy = NULL;
            buddyList = theSession->getBuddyList( &numBuddies );
            for ( i=0; i<numBuddies; ++i )
            {
              if ( buddyList[i]->record &&
                    notify->pRecord == buddyList[i]->record )
              {
                theBuddy = buddyList[i];
                break;
              }
            }
            
            if ( !theBuddy )
            {
              debugf( "Couldn't find the buddy we just picked!\n" );
              WinAlarm( HWND_DESKTOP, WA_WARNING );
              break;
            }
            
            if ( theBuddy->isGroup ) break;
            // Don't do anything if the user double-clicked the background or
            //  a buddy group.
            
            if ( theBuddy->imChatWin )
            {
              debugf( "IM chat window for this buddy already exists.\n" );
              WinSetFocus( HWND_DESKTOP, theBuddy->imChatWin );
              break;
            }
            
            if ( !theBuddy->userData.isOnline )
            {
              if ( theSession->settings->settingFlags[ EVENT_ERRORBOX ] &
                    SET_SOUND_ENABLED )
              {
                mciPlayFile( HWND_DESKTOP,
                 theSession->settings->sounds[ EVENT_ERRORBOX ], MCI_ASYNC,
                 NULL, NULLHANDLE );
              }
              WinMessageBox( HWND_DESKTOP, win,
               "This user is currently not on-line.  You cannot send them an instant message.",
               "Cannot send an instant message", 999,
               MB_CANCEL | MB_APPLMODAL | MB_MOVEABLE );
              break;
            }
            
            WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
             CM_SETRECORDEMPHASIS, MPFROMP( notify->pRecord ),
             MPFROM2SHORT( TRUE, CRA_INUSE ) );
            
            tmpLong = strlen( theSession->getUserName() ) +
             strlen( notify->pRecord->pszTree ) + 30;
             
            CHECKED_MALLOC( tmpLong, text );
            
            sprintf( text, "[%s <-> %s] Instant Message Window",
             theSession->getUserName(), notify->pRecord->pszTree );
            
            imFrame = WinCreateStdWindow( HWND_DESKTOP, WS_ANIMATE, &fcf,
             "MrMessage Instant Message Window", text, 0, dll, 1,
             &imClient );
            
            WinSetOwner( imFrame, win );
            
            theBuddy->imChatWin = imFrame;
            theBuddy->imChatClientWin = imClient;
            
            WinSetWindowPtr( imClient, 0, buddyList[i] );
            WinSetWindowULong( imClient, 4, win );
            
            WinQueryWindowRect( HWND_DESKTOP, &desktop );
            
            swp.x = desktop.xLeft +
             ((desktop.xRight - desktop.xLeft - IM_WINDOW_WIDTH) / 2);
            swp.y = desktop.yBottom + ((desktop.yTop - desktop.yBottom -
             IM_WINDOW_HEIGHT) / 2);
            swp.cx = IM_WINDOW_WIDTH;
            swp.cy = IM_WINDOW_HEIGHT;
            // Establish defaults
            
            tmpLong = sizeof( SWP );
            PrfQueryProfileData( iniFile, text, "Window position data",
             &swp, &tmpLong );
            // Read INI settings if there are any
            
            WinSendMsg( imFrame, WM_SETICON, MPFROMLONG(
             theBuddy->record->hptrIcon ), NULL );
            
            WinSetWindowPos( imFrame, 0, swp.x, swp.y, swp.cx, swp.cy,
             SWP_SIZE | SWP_MOVE | SWP_SHOW );
            
            WinSetFocus( HWND_DESKTOP, imFrame );
            
            CHECKED_FREE( text );
            
            if ( theSession->settings->settingFlags[ EVENT_FIRSTCOMM ] &
                  SET_SOUND_ENABLED )
            {
              mciPlayFile( HWND_DESKTOP,
               theSession->settings->sounds[ EVENT_FIRSTCOMM ],
               MCI_ASYNCRENDEZVOUS, NULL, NULLHANDLE );
            }
          }
          break;
        }
      }
    }
    break;
    case WM_MOUSEMOVE:
    case WM_CHAR:
      WinSendMsg( win, WM_NOTIDLE, 0, 0 );
    break;
  }
  return WinDefWindowProc( win, msg, mp1, mp2 );
}

MRESULT EXPENTRY pmSessionEditor( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch ( msg )
  {
    case WM_INITDLG:
    {
      char buffer[ 1024 ];
      ULONG attrLen;
      HWND ownerWin = WinWindowFromID( 
       WinQueryWindow( win, QW_OWNER ), FID_CLIENT );
      HWND buttonWin = WinWindowFromID( ownerWin, MRM_CreateSession );
      
      sessionEditorInit *initData = (sessionEditorInit *) PVOIDFROMMP( mp2 );
      if ( initData )
      {
        WinSetDlgItemText( win, MRM_EN_ScreenName, initData->userName );
        WinSetDlgItemText( win, MRM_EN_Password, initData->password );
        WinSetDlgItemText( win, DID_OK, "~Modify Session" );
      }
      attrLen = WinQueryPresParam( ownerWin, PP_FONTNAMESIZE, 0, NULL, 1024,
       buffer, 0 );
      if ( attrLen )
      {
        WinSetPresParam( win, PP_FONTNAMESIZE, attrLen, buffer );
        // Copy the default font over to this window
      }
      attrLen = WinQueryPresParam( buttonWin, PP_FONTNAMESIZE, 0, NULL, 1024,
       buffer, 0 );
      if ( attrLen )
      {
        WinSetPresParam( WinWindowFromID( win, DID_OK ), PP_FONTNAMESIZE,
         attrLen, buffer );
        WinSetPresParam( WinWindowFromID( win, DID_CANCEL ), PP_FONTNAMESIZE,
         attrLen, buffer );
      }
    }
    break;
    case WM_COMMAND:
    {
      if ( SHORT1FROMMP( mp1 ) == DID_OK )
      {
        // Make sure we have a username and password before we can OK this.
        LONG theLen;
        
        theLen = WinQueryDlgItemTextLength( win, MRM_EN_ScreenName );
        if ( !theLen )
        {
          // Didn't have a username
          WinAlarm( HWND_DESKTOP, WA_WARNING );
          WinSetFocus( HWND_DESKTOP, WinWindowFromID( win, MRM_EN_ScreenName ) );
          return 0;
        }
        
        theLen = WinQueryDlgItemTextLength( win, MRM_EN_Password );
        if ( !theLen )
        {
          // Didn't have a password
          WinAlarm( HWND_DESKTOP, WA_WARNING );
          WinSetFocus( HWND_DESKTOP, WinWindowFromID( win, MRM_EN_Password ) );
          return 0;
        }
      }
    }
    break;
  }
  
  return WinDefDlgProc( win, msg, mp1, mp2 );
}

MRESULT EXPENTRY pmPresParamNotifier( HWND win, ULONG msg, MPARAM mp1,
 MPARAM mp2 )
{
  MRESULT EXPENTRY (*oldProc) ( HWND, ULONG, MPARAM, MPARAM );
  switch ( msg )
  {
    case WM_PRESPARAMCHANGED:
      WinSendMsg( WinQueryWindow( win, QW_PARENT ), WM_WINPPCHANGED,
       MPFROMSHORT( WinQueryWindowUShort( win, QWS_ID ) ), mp1 );
    break;
    case WM_CHAR:
      if ( !(SHORT1FROMMP( mp1 ) & KC_KEYUP) &&
            SHORT1FROMMP( mp1 ) & KC_VIRTUALKEY )
      {
        USHORT theVK = SHORT2FROMMP( mp2 );
        if ( theVK == VK_TAB )
        {
          if ( WinQueryWindowUShort( win, QWS_ID ) == 104 )
          {
            WinSetFocus( HWND_DESKTOP, 
             WinWindowFromID( WinQueryWindow( win, QW_PARENT ), 100 ) );
          } else {
            WinSetFocus( HWND_DESKTOP, 
             WinWindowFromID( WinQueryWindow( win, QW_PARENT ),
             WinQueryWindowUShort( win, QWS_ID ) + 1 ) );
          }
        }
        if ( theVK == VK_BACKTAB )
        {
          if ( WinQueryWindowUShort( win, QWS_ID ) == 100 )
          {
            WinSetFocus( HWND_DESKTOP,
             WinWindowFromID( WinQueryWindow( win, QW_PARENT ), 104 ) );
          } else {
            WinSetFocus( HWND_DESKTOP,
             WinWindowFromID( WinQueryWindow( win, QW_PARENT ),
             WinQueryWindowUShort( win, QWS_ID ) - 1 ) );
          }
        }
      }
    break;
  }
  
  oldProc = (MRESULT EXPENTRY (*) ( HWND, ULONG, MPARAM, MPARAM ))
   WinQueryWindowPtr( WinQueryWindow( win, QW_PARENT ), 32 +
    ((WinQueryWindowUShort( win, QWS_ID ) - 100) * 4) );
  
  // Window procedures are stored in the parent window from 32 on in order of
  //  window IDs.  Window ID - 100 is the index into this array of window
  //  procedure addresses.

  return oldProc( win, msg, mp1, mp2 );
}

static void uniquifySettingsStrings( sessionThreadSettings *theSettings )
{
  int i;
  
  for ( i=0; i<EVENT_MAXEVENTS; ++i )
  {
    if ( theSettings->sounds[i] )
      theSettings->sounds[i] = strdup( theSettings->sounds[i] );
    if ( theSettings->rexxScripts[i] )
      theSettings->rexxScripts[i] = strdup( theSettings->rexxScripts[i] );
    if ( theSettings->shellCmds[i] )
      theSettings->shellCmds[i] = strdup( theSettings->shellCmds[i] );
  }
  theSettings->inheritSettings = 0;
}

MRESULT EXPENTRY pmSettings( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch ( msg )
  {
    case WM_CONTROL:
    {
      SHORT idx = -1;
      sessionThreadSettings *theSettings = NULL;
      int dupStrings = 0;
      
      if ( ((SHORT1FROMMP( mp1 ) == MRM_LB_SettingsScope ||
             SHORT1FROMMP( mp1 ) == MRM_LB_EventType) &&
            SHORT2FROMMP( mp1 ) == LN_SELECT) ||
           ((SHORT1FROMMP( mp1 ) == MRM_EN_EventPlaySoundFile ||
             SHORT1FROMMP( mp1 ) == MRM_EN_EventRunREXXScript ||
             SHORT1FROMMP( mp1 ) == MRM_EN_EventRunCommand) &&
            SHORT2FROMMP( mp1 ) == EN_KILLFOCUS) ||
           ((SHORT1FROMMP( mp1 ) == MRM_CB_EventPlaySound ||
             SHORT1FROMMP( mp1 ) == MRM_CB_EventRunREXX ||
             SHORT1FROMMP( mp1 ) == MRM_CB_EventRun) &&
            (SHORT2FROMMP( mp1 ) == BN_CLICKED ||
             SHORT2FROMMP( mp1 ) == BN_DBLCLICKED)) )
      {
        SHORT maxIdx;
        
        maxIdx = SHORT1FROMMR( WinSendDlgItemMsg( win,
         MRM_LB_SettingsScope, LM_QUERYITEMCOUNT, NULL, NULL ) );
        maxIdx--;
        
        idx = SHORT1FROMMR( WinSendDlgItemMsg( win,
         MRM_LB_SettingsScope, LM_QUERYSELECTION, MPFROMSHORT( LIT_FIRST ),
         NULL ) );
        
        if ( idx == LIT_NONE )
        {
          WinAlarm( HWND_DESKTOP, WA_ERROR );
          break;
        }
        
        theSettings = (sessionThreadSettings *) PVOIDFROMMR(
         WinSendDlgItemMsg( win, MRM_LB_SettingsScope, LM_QUERYITEMHANDLE,
         MPFROMSHORT( idx ), NULL ) );
        
        WinEnableWindow( WinWindowFromID( win, MRM_PB_UseGlobal ),
         idx != maxIdx && !theSettings->inheritSettings );
        
        if ( theSettings->inheritSettings )
        {
          // We should inherit our settings from the global settings.
          // Copy in the current global settings.
          sessionThreadSettings *globalSettings;
          
          globalSettings = (sessionThreadSettings *) PVOIDFROMMR(
           WinSendDlgItemMsg( win, MRM_LB_SettingsScope, LM_QUERYITEMHANDLE,
           MPFROMSHORT( maxIdx ), NULL ) );
          memcpy( theSettings, globalSettings,
           sizeof( sessionThreadSettings ) );
          theSettings->inheritSettings = 1;
        }
        
        dupStrings = theSettings->inheritSettings;
        
        idx = SHORT1FROMMR( WinSendDlgItemMsg( win,
         MRM_LB_EventType, LM_QUERYSELECTION, MPFROMSHORT( LIT_FIRST ),
         NULL ) );
      }
      
      if ( SHORT2FROMMP( mp1 ) == BN_CLICKED ||
            SHORT2FROMMP( mp1 ) == BN_DBLCLICKED )
      {
        if ( SHORT1FROMMP( mp1 ) == MRM_CB_EventPlaySound )
        {
          if ( WinQueryButtonCheckstate( win, MRM_CB_EventPlaySound ) )
          {
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventPlaySoundFile ),
             TRUE );
            theSettings->settingFlags[idx] |= SET_SOUND_ENABLED;
            dupStrings++;
          } else {
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventPlaySoundFile ),
             FALSE );
            theSettings->settingFlags[idx] &= ~SET_SOUND_ENABLED;
            dupStrings++;
          }
        } else if ( SHORT1FROMMP( mp1 ) == MRM_CB_EventRunREXX )
        {
          if ( WinQueryButtonCheckstate( win, MRM_CB_EventRunREXX ) )
          {
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunREXXScript ),
             TRUE );
            theSettings->settingFlags[idx] |= SET_REXX_ENABLED;
            dupStrings++;
          } else {
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunREXXScript ),
             FALSE );
            theSettings->settingFlags[idx] &= ~SET_REXX_ENABLED;
            dupStrings++;
          }
        } else if ( SHORT1FROMMP( mp1 ) == MRM_CB_EventRun )
        {
          if ( WinQueryButtonCheckstate( win, MRM_CB_EventRun ) )
          {
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunCommand ),
             TRUE );
            theSettings->settingFlags[idx] |= SET_SHELL_ENABLED;
            dupStrings++;
          } else {
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunCommand ),
             FALSE );
            theSettings->settingFlags[idx] &= ~SET_SHELL_ENABLED;
            dupStrings++;
          }
        }
        if ( dupStrings > 1 )
        {
          // Have to have inherit flag set and changed a checkbox
          uniquifySettingsStrings( theSettings );
          WinEnableWindow( WinWindowFromID( win, MRM_PB_UseGlobal ), TRUE );
        }
      }
      
      if ( SHORT1FROMMP( mp1 ) == MRM_LB_SettingsScope ||
            SHORT1FROMMP( mp1 ) == MRM_LB_EventType )
      {
        if ( SHORT2FROMMP( mp1 ) == LN_SELECT )
        {
          if ( idx == LIT_NONE )
          {
            WinEnableWindow( WinWindowFromID( win, MRM_CB_EventPlaySound ),
             FALSE );
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventPlaySoundFile ),
             FALSE );
            WinEnableWindow( WinWindowFromID( win, MRM_CB_EventRunREXX ),
             FALSE );
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunREXXScript ),
             FALSE );
            WinEnableWindow( WinWindowFromID( win, MRM_CB_EventRun ),
             FALSE );
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunCommand ),
             FALSE );
          } else {
            WinCheckButton( win, MRM_CB_EventPlaySound,
             theSettings->settingFlags[idx] & SET_SOUND_ENABLED );
            WinSetDlgItemText( win, MRM_EN_EventPlaySoundFile,
             theSettings->sounds[idx] );
            WinSendDlgItemMsg( win, MRM_EN_EventPlaySoundFile,
             EM_QUERYCHANGED, NULL, NULL );
            // Reset the changed status
            
            WinCheckButton( win, MRM_CB_EventRunREXX,
             theSettings->settingFlags[idx] & SET_REXX_ENABLED );
            WinSetDlgItemText( win, MRM_EN_EventRunREXXScript,
             theSettings->rexxScripts[idx] );
            WinSendDlgItemMsg( win, MRM_EN_EventPlaySoundFile,
             EM_QUERYCHANGED, NULL, NULL );
            // Reset the changed status
            
            WinCheckButton( win, MRM_CB_EventRun,
             theSettings->settingFlags[idx] & SET_SHELL_ENABLED );
            WinSetDlgItemText( win, MRM_EN_EventRunCommand,
             theSettings->shellCmds[idx] );
            WinSendDlgItemMsg( win, MRM_EN_EventPlaySoundFile,
             EM_QUERYCHANGED, NULL, NULL );
            // Reset the changed status
            
            WinEnableWindow( WinWindowFromID( win, MRM_CB_EventPlaySound ),
             TRUE );
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventPlaySoundFile ),
             theSettings->settingFlags[idx] & SET_SOUND_ENABLED );
/*             
            WinEnableWindow( WinWindowFromID( win, MRM_CB_EventRunREXX ),
             TRUE );
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunREXXScript ),
             theSettings->settingFlags[idx] & SET_REXX_ENABLED );
            WinEnableWindow( WinWindowFromID( win, MRM_CB_EventRun ),
             TRUE );
            WinEnableWindow( WinWindowFromID( win, MRM_EN_EventRunCommand ),
             theSettings->settingFlags[idx] & SET_SHELL_ENABLED );
*/          
            // Don't enable because there's no functionality yet
          }
        }
      }
      
      if ( SHORT1FROMMP( mp1 ) == MRM_EN_EventPlaySoundFile &&
            SHORT2FROMMP( mp1 ) == EN_KILLFOCUS )
      {
        if ( WinSendDlgItemMsg( win, MRM_EN_EventPlaySoundFile,
              EM_QUERYCHANGED, NULL, NULL ) )
        {
          int textLen = WinQueryDlgItemTextLength( win,
                         MRM_EN_EventPlaySoundFile );
          
          if ( theSettings->inheritSettings )
            uniquifySettingsStrings( theSettings );
          
          if ( theSettings->sounds[idx] )
            CHECKED_FREE( theSettings->sounds[idx] );
          
          theSettings->sounds[idx] = NULL;
          theSettings->inheritSettings = 0;
          WinEnableWindow( WinWindowFromID( win, MRM_PB_UseGlobal ), TRUE );
              
          if ( textLen )
          {
            CHECKED_MALLOC( textLen + 1, theSettings->sounds[idx] );
            WinQueryDlgItemText( win, MRM_EN_EventPlaySoundFile, textLen + 1,
             theSettings->sounds[idx] );
          }
        }
      }
      if ( SHORT1FROMMP( mp1 ) == MRM_EN_EventRunREXXScript &&
            SHORT2FROMMP( mp1 ) == EN_KILLFOCUS )
      {
        if ( WinSendDlgItemMsg( win, MRM_EN_EventRunREXXScript,
              EM_QUERYCHANGED, NULL, NULL ) )
        {
          int textLen = WinQueryDlgItemTextLength( win,
           MRM_EN_EventRunREXXScript );
          
          if ( theSettings->inheritSettings )
            uniquifySettingsStrings( theSettings );
          
          if ( theSettings->rexxScripts[idx] )
            CHECKED_FREE( theSettings->rexxScripts[idx] );
          
          theSettings->rexxScripts[idx] = NULL;
          theSettings->inheritSettings = 0;
          WinEnableWindow( WinWindowFromID( win, MRM_PB_UseGlobal ), TRUE );
          
          if ( textLen )
          {
            CHECKED_MALLOC( textLen + 1, theSettings->rexxScripts[idx] );
            WinQueryDlgItemText( win, MRM_EN_EventRunREXXScript, textLen + 1,
             theSettings->rexxScripts[idx] );
          }
        }
      }
      if ( SHORT1FROMMP( mp1 ) == MRM_EN_EventRunREXXScript &&
            SHORT2FROMMP( mp1 ) == EN_KILLFOCUS )
      {
        if ( WinSendDlgItemMsg( win, MRM_EN_EventRunCommand,
              EM_QUERYCHANGED, NULL, NULL ) )
        {
          int textLen = WinQueryDlgItemTextLength( win,
           MRM_EN_EventRunCommand );
           
          if ( theSettings->inheritSettings )
            uniquifySettingsStrings( theSettings );
          
          if ( theSettings->shellCmds[idx] )
            CHECKED_FREE( theSettings->shellCmds[idx] );
          
          theSettings->shellCmds[idx] = NULL;
          theSettings->inheritSettings = 0;
          WinEnableWindow( WinWindowFromID( win, MRM_PB_UseGlobal ), TRUE );
          
          if ( textLen )
          {
            CHECKED_MALLOC( textLen + 1, theSettings->shellCmds[idx] );
            WinQueryDlgItemText( win, MRM_EN_EventRunCommand, textLen + 1,
             theSettings->shellCmds[idx] );
          }
        }
      }
    }
    break;
    case WM_COMMAND:
      switch ( SHORT1FROMMP( mp1 ) )
      {
        case MRM_PB_UseGlobal:
        {
          int i;
          SHORT idx = -1;
          sessionThreadSettings *theSettings = NULL;
          
          idx = SHORT1FROMMR( WinSendDlgItemMsg( win,
           MRM_LB_SettingsScope, LM_QUERYSELECTION, MPFROMSHORT( LIT_FIRST ),
           NULL ) );
          
          if ( idx == LIT_NONE )
          {
            WinAlarm( HWND_DESKTOP, WA_ERROR );
            break;
          }
          
          theSettings = (sessionThreadSettings *) PVOIDFROMMR(
           WinSendDlgItemMsg( win, MRM_LB_SettingsScope, LM_QUERYITEMHANDLE,
           MPFROMSHORT( idx ), NULL ) );
          
          if ( theSettings->inheritSettings ) return 0;
          
          for ( i=0; i<EVENT_MAXEVENTS; ++i )
          {
            if ( theSettings->sounds[i] )
            {
              CHECKED_FREE( theSettings->sounds[i] );
              theSettings->sounds[i] = NULL;
            }
            if ( theSettings->rexxScripts[i] )
            {
              CHECKED_FREE( theSettings->rexxScripts[i] );
              theSettings->rexxScripts[i] = NULL;
            }
            if ( theSettings->shellCmds[i] )
            {
              CHECKED_FREE( theSettings->shellCmds[i] );
              theSettings->shellCmds[i] = NULL;
            }
          }
          theSettings->inheritSettings = 1;
          WinSendMsg( win, WM_CONTROL,
           MPFROM2SHORT( MRM_LB_SettingsScope, LN_SELECT ), NULL );
          // Force contents update and button disabling
        }
        return 0;
        // Don't want to dismiss dialog
      }
    break;
  }
  return WinDefDlgProc( win, msg, mp1, mp2 );
}

MRESULT EXPENTRY pmSessionManager( HWND win, ULONG msg, MPARAM mp1, MPARAM mp2 )
{
  switch ( msg )
  {
    case WM_USERBOOTED:
    {
      ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
      sessionThreadInfo **info =
       (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
      int i;
      
      for ( i=0; i<(*numItems); ++i )
      {
        if ( (*info)[i].theSession == PVOIDFROMMP( mp1 ) )
        {
          WinSendMsg( (*info)[i].theSession->getBuddyListWindow(),
           WM_CLOSE, NULL, NULL );
          // I'll do something classier later on, but for now, this will do.
          break;
        }
      }
    }
    break;
    case WM_CREATEBUDDYLIST:
    {
      ULONG fcf = FCF_TITLEBAR | FCF_MINMAX | FCF_SYSMENU | FCF_TASKLIST |
                   FCF_SIZEBORDER | FCF_ICON | FCF_NOMOVEWITHOWNER;
      HMODULE dll = getModuleHandle();
      HWND buddyFrame, buddyClient;
      RECTL desktop;
      buddyListCreateData *buddyCreateData = (buddyListCreateData *)
       PVOIDFROMMP( mp2 );
      UserInformation *userInfo = buddyCreateData->userInfo;
      MINIRECORDCORE *myRecord = (MINIRECORDCORE *) PVOIDFROMMP( mp1 );
      char *buffer;
      HINI iniFile = WinQueryWindowULong( win, 20 );
      SWP swp;
      ULONG tmpLong;
      
      CHECKED_MALLOC( 14 + strlen( userInfo->screenName ), buffer );
      
      sprintf( buffer, "Buddy List [%s]", userInfo->screenName );
      buddyFrame = WinCreateStdWindow( HWND_DESKTOP, WS_ANIMATE, &fcf,
       "MrMessage Buddy List", buffer, 0, dll, 1, &buddyClient );
       
      WinSetOwner( buddyFrame, win );
      
      buddyCreateData->buddyListWin = buddyFrame;
      
      WinQueryWindowRect( HWND_DESKTOP, &desktop );
      
      swp.x = desktop.xLeft +
       ((desktop.xRight - desktop.xLeft - BUDDY_LIST_WIDTH) / 2);
      swp.y = desktop.yBottom + ((desktop.yTop - desktop.yBottom -
        BUDDY_LIST_HEIGHT) / 2);
      swp.cx = BUDDY_LIST_WIDTH;
      swp.cy = BUDDY_LIST_HEIGHT;
      // Establish defaults
      
      tmpLong = sizeof( SWP );
      PrfQueryProfileData( iniFile, buffer, "Window position data", &swp,
       &tmpLong );
      // Read INI settings if there are any
      
      CHECKED_FREE( buffer );
      
      WinSetWindowPos( buddyFrame, 0, swp.x, swp.y, swp.cx, swp.cy,
       SWP_SIZE | SWP_MOVE );
       
      WinPostMsg( buddyClient, WM_INITULONGS, mp1, MPFROMLONG( win ) );
      WinSetFocus( HWND_DESKTOP, buddyFrame );
      
      WinSetWindowPtr( buddyClient, 24, buddyCreateData->theSession );
      
      DosPostEventSem( buddyCreateData->wakeupSem );
    }
    break;
    case WM_ERRORMESSAGE:
    {
      char titleString[ 1024 ];
      MINIRECORDCORE *theRecord = (MINIRECORDCORE *)PVOIDFROMMP( mp2 );
      char *loginName = theRecord->pszIcon;
      ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
      sessionThreadInfo **info =
       (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
      OscarSession *theSession = NULL;
      int i;
      
      for ( i=0; i<(*numItems); ++i )
      {
        if ( (*info)[i].record == theRecord )
        {
          theSession = (*info)[i].theSession;
          break;
        }
      }
      
      if ( strlen( loginName ) > 998 )
      {
        strcpy( titleString, "Error with AIM session (" );
        strncat( titleString, loginName, 998 );
        strcat( titleString, ")" );
      } else {
        sprintf( titleString, "Error with AIM session (%s)", loginName );
      }
      
      if ( theSession && theSession->settings->
            settingFlags[ EVENT_ERRORBOX ] & SET_SOUND_ENABLED )
      {
        mciPlayFile( HWND_DESKTOP,
         theSession->settings->sounds[ EVENT_ERRORBOX ], MCI_ASYNC, NULL,
         NULLHANDLE );
      }
      
      WinMessageBox( HWND_DESKTOP, win,
       errorMessageStrings[ SHORT1FROMMP(mp1) ], titleString, 999,
       MB_OK | MB_APPLMODAL | MB_MOVEABLE );
      
      if ( SHORT1FROMMP( mp1 ) == MRM_AIM_BAD_PASSWORD ||
            SHORT1FROMMP( mp1 ) == MRM_AIM_BAD_IDPASS )
      {
        ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
        sessionThreadInfo **info =
         (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
        ULONG i, myEntry;
        MINIRECORDCORE *record;
        
        myEntry = *numItems;
        for ( i=0; i<*numItems; ++i )
        {
          if ( (*info)[i].userName && (*info)[i].record == theRecord )
          {
            myEntry = i;
            break;
          }
        }
        if ( myEntry == *numItems )
        {
          debugf( "A session reported a password error, but I couldn't find it.\n" );
          WinAlarm( HWND_DESKTOP, WA_ERROR );
        } else {
          record = (MINIRECORDCORE *)CMA_FIRST;
          do {
            record = (MINIRECORDCORE *) WinSendMsg(
             WinWindowFromID( win, WID_SESSION_CONTAINER ),
             CM_QUERYRECORDEMPHASIS, MPFROMP( record ),
             MPFROMSHORT( CRA_SELECTED ) );
            if ( !record || (LONG)record == -1 ) break;
          
            WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
             CM_SETRECORDEMPHASIS, MPFROMP( record ),
             MPFROM2SHORT( FALSE, CRA_SELECTED ) );
          } while ( 1 );
          // De-select everything
          
          WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
           CM_SETRECORDEMPHASIS, mp2, MPFROM2SHORT( TRUE, CRA_SELECTED ) );
          // Select the record with the problem
          
          WinSetWindowULong( win, 56, 1 );
          WinSendMsg( win, WM_COMMAND, MPFROMSHORT( MRM_EditSession ), NULL );
          WinSetWindowULong( win, 56, 0 );
          // And edit it
        }
      }
      
      WinPostMsg( win, WM_SESSIONCLOSED, mp2, NULL );
    }
    break;
    case WM_SESSIONCLOSED:
    {
      ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
      sessionThreadInfo **info =
       (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
      ULONG i;
      
      WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
       CM_SETRECORDEMPHASIS, mp1, MPFROM2SHORT( FALSE, CRA_INUSE ) );
      for ( i=0; i<(*numItems); ++i )
      {
        if ( (*info)[i].userName && (*info)[i].theSession &&
              (*info)[i].record == (MINIRECORDCORE *) PVOIDFROMMP( mp1 ) )
        {
          (*info)[i].theSession->windowClosed();
          break;
        }
      }
    }
    break;
    case WM_INITULONGS:
    {
      sessionManagerULONGs *initData =
       (sessionManagerULONGs *) PVOIDFROMMP( mp1 );
      ULONG numSessions, i, j, dataLen, myEntry, cursoredItem;
      USHORT txtLen;
      char keyString[80];
      char *text, *decodedPass, *ppKeys, *ppCurKey, *ppData;
      unsigned char *sessionData;
      MINIRECORDCORE *record, *cursoredRecord;
      RECORDINSERT insert;
      sessionThreadInfo **info;
      ULONG *numItems;
      HPOINTER icon;
      HMTX infoMux;
      const char *ppApp;
      sessionThreadSettings *mySettings;
      char *appName;

      WinSetWindowPtr( win, 4, initData->numItemsPtr );
      WinSetWindowPtr( win, 8, initData->threadInfoPtr );
      WinSetWindowULong( win, 12, initData->mutexSem );
      WinSetWindowULong( win, 20, initData->iniFile );
      WinSetWindowPtr( win, 64, initData->globalSettings );

      info = initData->threadInfoPtr;
      numItems = initData->numItemsPtr;
      infoMux = initData->mutexSem;
      
      // Now that we have the INI file handle, check if we can restore any
      //  AIM session data.
      
      for ( i=0; i<3; ++i )
      {
        switch ( i )
        {
          case 2:  ppApp = "Session Manager Presentation Parameters"; break;
          case 1:  ppApp = "Session Manager Presentation Parameters (Buttons)"; break;
          default: ppApp = "Session Manager Presentation Parameters (Container)";
        }
        if ( PrfQueryProfileSize( initData->iniFile, ppApp, NULL, &dataLen ) )
        {
          CHECKED_MALLOC( dataLen, ppKeys );
          
          if ( PrfQueryProfileData( initData->iniFile, ppApp, NULL, ppKeys,
                &dataLen ) )
          {
            ppCurKey = ppKeys;
            while ( *ppCurKey )
            {
              if ( PrfQueryProfileSize( initData->iniFile, ppApp, ppCurKey,
                    &dataLen ) )
              {
                CHECKED_MALLOC( dataLen, ppData );
                
                if ( PrfQueryProfileData( initData->iniFile, ppApp, ppCurKey,
                      ppData, &dataLen ) )
                {
                  switch ( i )
                  {
                    case 2:
                      WinSetPresParam( win, atol( ppCurKey ), dataLen, ppData );
                    break;
                    case 1:
                      WinSetPresParam( WinWindowFromID( win, MRM_StartSession ),
                       atol( ppCurKey ), dataLen, ppData );
                      WinSetPresParam( WinWindowFromID( win, MRM_CreateSession ),
                       atol( ppCurKey ), dataLen, ppData );
                      WinSetPresParam( WinWindowFromID( win, MRM_EditSession ),
                       atol( ppCurKey ), dataLen, ppData );
                      WinSetPresParam( WinWindowFromID( win, MRM_DeleteSession ),
                       atol( ppCurKey ), dataLen, ppData );
                    break;
                    default:
                      WinSetPresParam( WinWindowFromID( win, WID_SESSION_CONTAINER ),
                       atol( ppCurKey ), dataLen, ppData );
                  }
                }
                CHECKED_FREE( ppData );
              }
              ppCurKey += strlen( ppCurKey ) + 1;
            }
          }
          CHECKED_FREE( ppKeys );
        }
      }
        
      dataLen = 4;
      if ( !PrfQueryProfileData( initData->iniFile, "Session Manager",
            "Number of Sessions", &numSessions, &dataLen ) )
      {
        debugf( "Missing or corrupted session manager data.\n" );
        WinShowWindow( WinQueryWindow( win, QW_PARENT ), TRUE );
        WinSetFocus( HWND_DESKTOP, WinWindowFromID( win, WID_SESSION_CONTAINER ) );
        return NULL;
      }
      
      cursoredRecord = NULL;
      cursoredItem = 0;
      dataLen = 4;
      PrfQueryProfileData( initData->iniFile, "Session Manager",
       "Cursored Record", &cursoredItem, &dataLen );
      // If it's there, great.  If not... don't care.
      
      debugf( "Profile contains %d session entries.\n", numSessions );
      
      for ( i=1; i<=numSessions; ++i )
      {
        sprintf( keyString, "Session %d data", i );
        if ( PrfQueryProfileSize( initData->iniFile, "Session Manager",
              keyString, &dataLen ) )
        {
          CHECKED_MALLOC( dataLen, sessionData );
          
          PrfQueryProfileData( initData->iniFile, "Session Manager",
           keyString, sessionData, &dataLen );
          
          j = 0;
          while ( j < dataLen )
          {
            if ( j + 4 > dataLen )
            {
              debugf( "Profile entry for session %d is corrupt or incomplete.\n",
               i );
            }
            txtLen = *((USHORT *)(sessionData + j));
            if ( j + 4 + txtLen + *((USHORT *)(sessionData + j + 2 + txtLen)) >
                  dataLen )
            {
              debugf( "Profile entry for session %d is corrupt or incomplete.\n",
               i );
            }
            j += 2;
            
            CHECKED_MALLOC( txtLen + 1, text );
            
            strncpy( text, (char *)(sessionData + j), txtLen );
            text[ txtLen ] = 0;

            record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
             WID_SESSION_CONTAINER, CM_ALLOCRECORD, MPFROMLONG( 0 ),
             MPFROMSHORT( 1 ) );
            if ( !record )
            {
              debugf( "Error allocating record in container.\n" );
              CHECKED_FREE( text );
              CHECKED_FREE( sessionData );
              return 0;
            }
            
            icon = WinQueryWindowULong( win, 16 );
            
            if ( i == cursoredItem )
            {
              cursoredRecord = record;
            }
            
            record->flRecordAttr = 0;
            record->ptlIcon.x = 0;
            record->ptlIcon.y = 0;
            record->pszIcon = text;
            record->hptrIcon = icon;
            insert.cb = sizeof( RECORDINSERT );
            insert.pRecordOrder = (RECORDCORE *)CMA_END;
            insert.pRecordParent = NULL;
            insert.fInvalidateRecord = TRUE;
            insert.zOrder = CMA_TOP;
            insert.cRecordsInsert = 1;
            
            WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INSERTRECORD,
             MPFROMP( record ), MPFROMP( &insert ) );
            
            DosRequestMutexSem( infoMux, SEM_INDEFINITE_WAIT );
            // Paranoia, but so what...
            if ( !(*numItems) )
            {
              CHECKED_MALLOC( sizeof( sessionThreadInfo ), *info );
              myEntry = 0;
              (*numItems)++;
            } else {
              ULONG k;
              myEntry = *numItems;
              for ( k=0; k<*numItems; ++k )
              {
                if ( !(*info)[k].userName ) myEntry = k;
              }
              if ( *numItems == myEntry )
              {
                CHECKED_REALLOC( *info, ((*numItems) + 1) * sizeof( sessionThreadInfo ), *info );
                (*numItems)++;
              }
            }
            
            (*info)[ myEntry ].userName = text;
            mySettings = &((*info)[ myEntry ].settings);
            memcpy( mySettings, initData->globalSettings,
             sizeof( sessionThreadSettings ) );
            mySettings->inheritSettings = 1;
            
            CHECKED_MALLOC( strlen( (*info)[myEntry].userName ) + 10,
             appName );
            sprintf( appName, "%s Settings", (*info)[myEntry].userName );
            querySessionSettings( initData->iniFile, appName, mySettings );
            CHECKED_FREE( appName );
            appName = NULL;
            
            DosReleaseMutexSem( infoMux );
            
            j += txtLen;
            txtLen = *((USHORT *) (sessionData + j));
            j += 2;
            CHECKED_MALLOC( txtLen + 1, text );
            strncpy( text, (char *)(sessionData + j), txtLen );
            j += txtLen;
            decodedPass = (char *)aim_encode_password( text );
            strncpy( text, decodedPass, txtLen );
            text[ txtLen ] = 0;
            
            (*info)[ myEntry ].password = text;
            (*info)[ myEntry ].managerWin = win;
            (*info)[ myEntry ].theSession = NULL;
            (*info)[ myEntry ].record = record;
            (*info)[ myEntry ].threadID = 0;
          }
          
          CHECKED_FREE( sessionData );
        }
      }
      if ( cursoredRecord )
      {
        record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
         WID_SESSION_CONTAINER, CM_QUERYRECORD, NULL,
         MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER ) );
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_SETRECORDEMPHASIS,
         MPFROMP( record ), MPFROM2SHORT( FALSE,
          CRA_SELECTED | CRA_CURSORED ) );
        WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_SETRECORDEMPHASIS,
         MPFROMP( cursoredRecord ), MPFROM2SHORT( TRUE,
          CRA_SELECTED | CRA_CURSORED ) );
      }
      WinShowWindow( WinQueryWindow( win, QW_PARENT ), TRUE );
      WinSetFocus( HWND_DESKTOP, WinWindowFromID( win, WID_SESSION_CONTAINER ) );
    }
    break;
    case WM_CREATE:
    {
      ULONG tmpUl;
      HWND theWin;
      HMODULE dll = getModuleHandle();
      HPOINTER icon;
      HWND popupMenu;

      WinSetWindowULong( win, 0, 0 );
      WinSetWindowPtr( win, 4, NULL );
      WinSetWindowPtr( win, 8, NULL );
      WinSetWindowULong( win, 12, 0 );
      WinSetWindowULong( win, 52, 1 );
      WinSetWindowULong( win, 56, 0 );
      
      icon = WinLoadPointer( HWND_DESKTOP, dll, 1 );
      WinSetWindowULong( win, 16, icon );
      popupMenu = WinLoadMenu( HWND_DESKTOP, dll, MRM_SessionManagerPopup );
      WinSetWindowULong( win, 24, popupMenu );
      popupMenu = WinLoadMenu( HWND_DESKTOP, dll, MRM_SessionManagerBackPopup );
      WinSetWindowULong( win, 28, popupMenu );
      
      WinSendMsg( WinQueryWindow( win, QW_PARENT ), WM_SETICON,
       MPFROMLONG( icon ), NULL );
      
      // Since we're not taking a shell position, the window dimensions will
      //  be bogus, so don't worry about sizing the child windows yet.
      
      WinSetPresParam( win, PP_FONTNAMESIZE, 13, "12.Helv Bold" );
      // Default font for this window
      
      tmpUl = CLR_PALEGRAY;
      WinSetPresParam( win, PP_BACKGROUNDCOLORINDEX, 4, &tmpUl );
      // Default background color
      
      theWin = WinCreateWindow( win, WC_CONTAINER, "Mr. Message Sessions",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | CCS_AUTOPOSITION |
        CCS_MINIRECORDCORE | CCS_EXTENDSEL,
       0, 0, 0, 0, win, HWND_TOP, WID_SESSION_CONTAINER, NULL, NULL );
      
      WinSetWindowPtr( win, 32, WinSubclassWindow( theWin,
       pmPresParamNotifier ) );
       
      theWin = WinCreateWindow( win, WC_BUTTON, "Start",
       WS_DISABLED | WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_DEFAULT |
        BS_PUSHBUTTON,
       0, 0, 0, 0, win, HWND_TOP, MRM_StartSession, NULL, NULL );
      
      WinSetWindowPtr( win, 36, WinSubclassWindow( theWin,
       pmPresParamNotifier ) );
       
      theWin = WinCreateWindow( win, WC_BUTTON, "Create...",
       WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON,
       0, 0, 0, 0, win, HWND_TOP, MRM_CreateSession, NULL, NULL );
      
      WinSetWindowPtr( win, 40, WinSubclassWindow( theWin,
       pmPresParamNotifier ) );
       
      theWin = WinCreateWindow( win, WC_BUTTON, "Edit...",
       WS_DISABLED | WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON,
       0, 0, 0, 0, win, HWND_TOP, MRM_EditSession, NULL, NULL );
      
      WinSetWindowPtr( win, 44, WinSubclassWindow( theWin,
       pmPresParamNotifier ) );
       
      theWin = WinCreateWindow( win, WC_BUTTON, "Delete",
       WS_DISABLED | WS_VISIBLE | WS_TABSTOP | WS_GROUP | BS_PUSHBUTTON,
       0, 0, 0, 0, win, HWND_TOP, MRM_DeleteSession, NULL, NULL );
      
      WinSetWindowPtr( win, 48, WinSubclassWindow( theWin,
       pmPresParamNotifier ) );
       
      theWin = WinQueryWindow( win, QW_PARENT );
      theWin = WinWindowFromID( theWin, FID_MENU );
      WinEnableMenuItem( theWin, MRM_StartSession, FALSE );
      WinEnableMenuItem( theWin, MRM_EditSession, FALSE );
      WinEnableMenuItem( theWin, MRM_DeleteSession, FALSE );
    }
    break;
    case WM_SIZE:
    {
      USHORT newCx, newCy, widgetX, widgetY, widgetCx, widgetCy, buttonHeight;
      HWND theWin;
      HPS hps;
      POINTL points[3];
      
      newCx = SHORT1FROMMP( mp2 );
      newCy = SHORT2FROMMP( mp2 );
      
      hps = WinGetPS( WinWindowFromID( win, MRM_StartSession ) );
      // Make sure we pick up presentation params appropriate to the buttons
      //  when sizing the text.
      GpiQueryTextBox( hps, 24, "Create Start Edit Delete", 3,
       (PPOINTL) points );
      buttonHeight = (USHORT) (points[0].y - points[1].y + 8);
      
      WinReleasePS( hps );
      
      theWin = WinWindowFromID( win, WID_SESSION_CONTAINER );
      widgetX = 8;
      widgetCx = newCx - 16;
      widgetY = buttonHeight + 8;
      widgetCy = newCy - widgetY - 8;
      WinSetWindowPos( theWin, 0, widgetX, widgetY, widgetCx, widgetCy,
       SWP_SIZE | SWP_MOVE );
      
      theWin = WinWindowFromID( win, MRM_StartSession );
      widgetX = 8;
      widgetCx = (newCx / 4) - 16;
      widgetY = 0;
      widgetCy = buttonHeight;
      WinSetWindowPos( theWin, 0, widgetX, widgetY, widgetCx, widgetCy,
       SWP_SIZE | SWP_MOVE );
      
      theWin = WinWindowFromID( win, MRM_CreateSession );
      widgetX += widgetCx + 16;
      WinSetWindowPos( theWin, 0, widgetX, widgetY, widgetCx, widgetCy,
       SWP_SIZE | SWP_MOVE );
      
      theWin = WinWindowFromID( win, MRM_EditSession );
      widgetX += widgetCx + 16;
      WinSetWindowPos( theWin, 0, widgetX, widgetY, widgetCx, widgetCy,
       SWP_SIZE | SWP_MOVE );
      
      theWin = WinWindowFromID( win, MRM_DeleteSession );
      widgetX += widgetCx + 16;
      widgetCx = newCx - (widgetX + 8);
      WinSetWindowPos( theWin, 0, widgetX, widgetY, widgetCx, widgetCy,
       SWP_SIZE | SWP_MOVE );
    }
    break;
    case WM_PAINT:
    {
      HPS hps;
      RECTL rect;
      POINTL pt;
      ULONG colorIdx, ret;
      
      ret = WinQueryPresParam( win, PP_BACKGROUNDCOLORINDEX, 0, NULL, 4,
       &colorIdx, QPF_ID1COLORINDEX | QPF_NOINHERIT );
      if ( !ret )
      {
        ret = WinQueryPresParam( win, PP_BACKGROUNDCOLOR, 0, NULL, 4,
         &colorIdx, QPF_NOINHERIT );
        if ( !ret )
        {
          ret = WinQueryPresParam( win, PP_BACKGROUNDCOLORINDEX, 0, NULL, 4,
           &colorIdx, QPF_ID1COLORINDEX );
        }
      }
      
      hps = WinBeginPaint( win, NULLHANDLE, &rect );
      if ( ret < 4 )
        GpiSetColor( hps, CLR_PALEGRAY );
      else {
        GpiCreateLogColorTable( hps, 0, LCOLF_RGB, 0, 0, NULL );
        GpiSetColor( hps, colorIdx );
      }
      pt.x = rect.xLeft;
      pt.y = rect.yBottom;
      GpiMove( hps, &pt );
      pt.x = rect.xRight;
      pt.y = rect.yTop;
      GpiBox( hps, DRO_FILL, &pt, 0, 0 );
      WinEndPaint( hps );
    }
    break;
    case WM_COMMAND:
      switch ( SHORT1FROMMP( mp1 ) )
      {
        case MRM_SelectAll:
        case MRM_DeselectAll:
        {
          MINIRECORDCORE *record;
          HWND theWin = WinWindowFromID( win, WID_SESSION_CONTAINER );
          record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD,
           MPFROMSHORT( CMA_FIRST ), MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER ) );
          while ( record && (LONG) record != -1 )
          {
            if ( (!(record->flRecordAttr & CRA_SELECTED) &&
                  SHORT1FROMMP( mp1 ) == MRM_SelectAll) ||
                 ((record->flRecordAttr & CRA_SELECTED) &&
                  SHORT1FROMMP( mp1 ) == MRM_DeselectAll) )
            {
              WinSendMsg( theWin, CM_SETRECORDEMPHASIS, MPFROMP( record ),
               MPFROM2SHORT( (SHORT1FROMMP( mp1 ) == MRM_SelectAll),
               CRA_SELECTED ) );
            }
            record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD,
             MPFROMP( record ), MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER ) );
          }
        }
        break;
        case MRM_ExitAll:
          WinPostMsg( win, WM_CLOSE, NULL, NULL );
        break;
        case MRM_SessionSettings:
        case MRM_GlobalSettings:
        {
          ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
          sessionThreadInfo **info =
           (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
          HINI iniFile = WinQueryWindowULong( win, 20 );
          HMODULE dll = getModuleHandle();
          HWND dlgWin;
          RECTL desktop, dlgSize;
          SWP swp;
          ULONG tmpLong, i;
          SHORT idx, listLen, j;
          sessionThreadSettings *globalSettings = (sessionThreadSettings *)
           WinQueryWindowPtr( win, 64 );
          sessionThreadSettings *mySettings;

          dlgWin = WinLoadDlg( HWND_DESKTOP, WinQueryWindow( win, QW_PARENT ),
           pmSettings, dll, MRM_SettingsWindow, NULL );
          
          for ( i=0; i<EVENT_MAXEVENTS; ++i )
          {
            WinSendDlgItemMsg( dlgWin, MRM_LB_EventType, LM_INSERTITEM,
             MPFROMSHORT( LIT_END ), MPFROMP( UI_EventTypeStrings[i] ) );
          }
          
          for ( i=0; i<*numItems; ++i )
          {
            idx = SHORT1FROMMR( WinSendDlgItemMsg( dlgWin,
             MRM_LB_SettingsScope, LM_INSERTITEM,
             MPFROMSHORT( LIT_SORTASCENDING ),
             MPFROMP( (*info)[i].userName ) ) );
            WinSendDlgItemMsg( dlgWin, MRM_LB_SettingsScope, LM_SETITEMHANDLE,
             MPFROMSHORT( idx ), MPFROMP( &((*info)[i].settings) ) );
          }
          idx = SHORT1FROMMR( WinSendDlgItemMsg( dlgWin, MRM_LB_SettingsScope,
           LM_INSERTITEM, MPFROMSHORT( LIT_END ), MPFROMP( "All Sessions" ) ) );
          WinSendDlgItemMsg( dlgWin, MRM_LB_SettingsScope, LM_SETITEMHANDLE,
           MPFROMSHORT( idx ), MPFROMP( globalSettings ) );

          if ( SHORT1FROMMP( mp1 ) == MRM_GlobalSettings )
          {
            WinSendDlgItemMsg( dlgWin, MRM_LB_SettingsScope, LM_SELECTITEM,
             MPFROMSHORT( idx ), MPFROMLONG( TRUE ) );
            mySettings = globalSettings;
          } else {
            MINIRECORDCORE *record;
            
            record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
             WID_SESSION_CONTAINER, CM_QUERYRECORDEMPHASIS,
             MPFROMLONG( CMA_FIRST ), MPFROMSHORT( CRA_SOURCE ) );
            
            if ( !record || (LONG)record == -1 )
            {
              record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
               WID_SESSION_CONTAINER, CM_QUERYRECORDEMPHASIS,
               MPFROMLONG( CMA_FIRST ), MPFROMSHORT( CRA_SELECTED ) );
              
              if ( !record || (LONG)record == -1 )
              {
                WinAlarm( HWND_DESKTOP, WA_ERROR );
                debugf( "Attempted to pull up specific settings with nothing selected.\n" );
                WinDestroyWindow( dlgWin );
                return NULL;
              }
            } else {
              WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
               CM_SETRECORDEMPHASIS, MPFROMP( record ),
               MPFROM2SHORT( FALSE, CRA_SOURCE ) );
            }
            
            for ( i=0; i<(*numItems); ++i )
            {
              if ( (*info)[i].userName && (*info)[i].record == record )
              {
                listLen = idx;
                for ( j=0; j<listLen; ++j )
                {
                  mySettings = (sessionThreadSettings *) WinSendDlgItemMsg(
                   dlgWin, MRM_LB_SettingsScope, LM_QUERYITEMHANDLE,
                   MPFROMSHORT( j ), NULL );
                  if ( mySettings == &((*info)[i].settings) )
                  {
                    WinSendDlgItemMsg( dlgWin, MRM_LB_SettingsScope,
                     LM_SELECTITEM, MPFROMSHORT( j ), MPFROMLONG( TRUE ) );
                    break;
                  }
                }
                break;
              }
            }
          }
          
          WinQueryWindowRect( HWND_DESKTOP, &desktop );
          WinQueryWindowRect( dlgWin, &dlgSize );
          
          swp.x = desktop.xLeft + ((desktop.xRight - desktop.xLeft) / 2) -
           ((dlgSize.xRight - dlgSize.xLeft) / 2);
          swp.y = desktop.yBottom + ((desktop.yTop - desktop.yBottom) / 2) -
           ((dlgSize.yTop - dlgSize.yBottom) / 2);
          swp.cx = dlgSize.xRight - dlgSize.xLeft;
          swp.cy = dlgSize.yTop - dlgSize.yBottom;
          // Establish defaults
            
          tmpLong = sizeof( SWP );
          PrfQueryProfileData( iniFile, "Settings Dialog",
           "Window position data", &swp, &tmpLong );
          // Read INI settings if there are any
          
          WinSetWindowPos( dlgWin, 0, swp.x, swp.y, swp.cx, swp.cy,
           SWP_SIZE | SWP_MOVE | SWP_SHOW );
          
          WinProcessDlg( dlgWin );
          // Settings values are updated in the dialog window procedure
          
          WinQueryWindowPos( dlgWin, &swp );
          if ( (swp.fl & SWP_MINIMIZE) || (swp.fl & SWP_MAXIMIZE) )
          {
            swp.x = WinQueryWindowUShort( dlgWin, QWS_XRESTORE );
            swp.y = WinQueryWindowUShort( dlgWin, QWS_YRESTORE );
            swp.cx = WinQueryWindowUShort( dlgWin, QWS_CXRESTORE );
            swp.cy = WinQueryWindowUShort( dlgWin, QWS_CYRESTORE );
          }
          
          PrfWriteProfileData( iniFile, "Settings Dialog",
           "Window position data", &swp, sizeof( SWP ) );
          
          saveSessionSettings( iniFile, "Global Settings", mySettings );
          for ( i=0; i<(*numItems); ++i )
          {
            if ( (*info)[i].userName )
            {
              char *appName;
              CHECKED_MALLOC( strlen( (*info)[i].userName ) + 10, appName );
              sprintf( appName, "%s Settings", (*info)[i].userName );
              saveSessionSettings( iniFile, appName, &((*info)[i].settings) );
              CHECKED_FREE( appName );
            }
          }
          
          WinDestroyWindow( dlgWin );
        }
        break;
        case MRM_CreateSession:
        {
          HWND dlgWin;
          RECTL desktopSize, dlgSize;
          HPOINTER icon;
          HMODULE dll = getModuleHandle();
          ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
          sessionThreadInfo **info =
           (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
          HMTX infoMux = WinQueryWindowULong( win, 12 );
          
          dlgWin = WinLoadDlg( HWND_DESKTOP, win, pmSessionEditor,
           dll, 1, NULL );
          WinQueryWindowRect( HWND_DESKTOP, &desktopSize );
          WinQueryWindowRect( dlgWin, &dlgSize );
          
          WinSetWindowPos( dlgWin, HWND_TOP,
           desktopSize.xLeft + (
            ((desktopSize.xRight - desktopSize.xLeft) -
             (dlgSize.xRight - dlgSize.xLeft)) / 2),
           desktopSize.yBottom + (
            ((desktopSize.yTop - desktopSize.yBottom) -
             (dlgSize.yTop - dlgSize.yBottom)) / 2),
           0, 0, SWP_MOVE );
          // Center the dialog window
          
          if ( WinProcessDlg( dlgWin ) == DID_OK )
          {
            char *text;
            ULONG txtLen, myEntry;
            MINIRECORDCORE *record;
            RECORDINSERT insert;
            
            debugf( "Adding a session record.\n" );
            
            txtLen = WinQueryDlgItemTextLength( dlgWin, MRM_EN_ScreenName ) + 1;
            CHECKED_MALLOC( txtLen, text );
            WinQueryDlgItemText( dlgWin, MRM_EN_ScreenName, txtLen, text );
            record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
             WID_SESSION_CONTAINER, CM_ALLOCRECORD, MPFROMLONG( 0 ),
             MPFROMSHORT( 1 ) );
            if ( !record )
            {
              debugf( "Error allocating record in container.\n" );
              CHECKED_FREE( text );
              WinDestroyWindow( dlgWin );
              return 0;
            }
            
            icon = WinQueryWindowULong( win, 16 );
            
            record->flRecordAttr = CRA_CURSORED;
            record->ptlIcon.x = 0;
            record->ptlIcon.y = 0;
            record->pszIcon = text;
            record->hptrIcon = icon;
            insert.cb = sizeof( RECORDINSERT );
            insert.pRecordOrder = (RECORDCORE *)CMA_END;
            insert.pRecordParent = NULL;
            insert.fInvalidateRecord = TRUE;
            insert.zOrder = CMA_TOP;
            insert.cRecordsInsert = 1;
            
            WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INSERTRECORD,
             MPFROMP( record ), MPFROMP( &insert ) );
            
            DosRequestMutexSem( infoMux, SEM_INDEFINITE_WAIT );
            if ( !(*numItems) )
            {
              CHECKED_MALLOC( sizeof( sessionThreadInfo ), *info );
              myEntry = 0;
              (*numItems)++;
            } else {
              ULONG i;
              myEntry = *numItems;
              for ( i=0; i<*numItems; ++i )
              {
                if ( !(*info)[i].userName ) myEntry = i;
              }
              if ( *numItems == myEntry )
              {
                CHECKED_REALLOC( *info, ((*numItems) + 1) * sizeof( sessionThreadInfo ), *info );
                (*numItems)++;
              }
            }
            
            (*info)[ myEntry ].userName = text;
            DosReleaseMutexSem( infoMux );
            
            txtLen = WinQueryDlgItemTextLength( dlgWin, MRM_EN_Password ) + 1;
            CHECKED_MALLOC( txtLen, text );
            WinQueryDlgItemText( dlgWin, MRM_EN_Password, txtLen, text );
            
            (*info)[ myEntry ].password = text;
            (*info)[ myEntry ].managerWin = win;
            (*info)[ myEntry ].theSession = NULL;
            (*info)[ myEntry ].record = record;
            (*info)[ myEntry ].threadID = 0;
          }
          
          WinDestroyWindow( dlgWin );
        }
        break;
        case MRM_DeleteSession:
        {
          HWND theWin;
          MINIRECORDCORE *record, *prevRecord;
          ULONG i, onlyOneRecord;
          ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
          sessionThreadInfo **info =
           (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
          
          theWin = WinWindowFromID( win, WID_SESSION_CONTAINER );
          debugf( "User requested to delete record(s).\n" );
          
          record = (MINIRECORDCORE *)CMA_FIRST;
          
          record = (MINIRECORDCORE *) WinSendMsg( theWin,
           CM_QUERYRECORDEMPHASIS, MPFROMP( record ),
           MPFROMSHORT( CRA_SOURCE ) );
          
          if ( record && (LONG)record != -1 &&
                !(record->flRecordAttr & CRA_SELECTED) )
          {
            // Source emphasis, but not selected.  Act only on this object.
            onlyOneRecord = 1;
            prevRecord = (MINIRECORDCORE *)CMA_FIRST;
          } else {
            onlyOneRecord = 0;
            record = (MINIRECORDCORE *)CMA_FIRST;
          }
          
          do {
            if ( !onlyOneRecord )
            {
              prevRecord = record;
              record = (MINIRECORDCORE *) WinSendMsg( theWin,
               CM_QUERYRECORDEMPHASIS, MPFROMP( record ),
               MPFROMSHORT( CRA_SELECTED ) );
              if ( !record || (LONG)record == -1 ) break;
              
              debugf( "Found a selected record.\n" );
            }
            
            for ( i=0; i<(*numItems); ++i )
            {
              if ( (*info)[i].userName && (*info)[i].record == record )
              {
                if ( (*info)[i].theSession || (*info)[i].threadID )
                {
                  // Session is running... why are we deleting it?!
                  char *titleString;
                  
                  CHECKED_MALLOC( 24 + strlen( (*info)[i].userName ), titleString );
                  
                  sprintf( titleString, "Cannot remove session: %s",
                   (*info)[i].userName );
                  
                  if ( (*info)[i].settings.settingFlags[ EVENT_ERRORBOX ] &
                        SET_SOUND_ENABLED )
                  {
                    mciPlayFile( HWND_DESKTOP, (*info)[i].settings.
                     sounds[ EVENT_ERRORBOX ], MCI_ASYNC, NULL, NULLHANDLE );
                  }
                  
                  WinMessageBox( HWND_DESKTOP, win,
                   "Session is currently opened and running.  You cannot delete it.  Please close the session if you wish to delete it.",
                   titleString, 502,
                   MB_CANCEL | MB_APPLMODAL | MB_MOVEABLE );
                  CHECKED_FREE( titleString );
                } else {
                  HMTX infoMux = WinQueryWindowULong( win, 12 );
                  
                  DosRequestMutexSem( infoMux, SEM_INDEFINITE_WAIT );
                  (*info)[i].record = NULL;
                  if ( (*info)[i].userName ) CHECKED_FREE( (*info)[i].userName );
                  if ( (*info)[i].password ) CHECKED_FREE( (*info)[i].password );
                  (*info)[i].userName = NULL;
                  DosReleaseMutexSem( infoMux );
                  
                  WinSendMsg( theWin, CM_SETRECORDEMPHASIS, MPFROMP( record ),
                   MPFROM2SHORT( FALSE, CRA_SELECTED ) );
                  WinSendMsg( theWin, CM_REMOVERECORD, MPFROMP( &record ),
                   MPFROM2SHORT( 1, CMA_FREE | CMA_INVALIDATE ) );
                  
                  record = prevRecord;
                  // Need to back up one step so we don't miss something
                  debugf( "Record was removed.\n" );
                }
              }
            }
            
            if ( onlyOneRecord ) break;
          } while ( 1 );
        }
        break;
        case MRM_StartSession:
        {
          HWND theWin;
          MINIRECORDCORE *record;
          ULONG i, onlyOneRecord;
          ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
          sessionThreadInfo **info =
           (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
          
          theWin = WinWindowFromID( win, WID_SESSION_CONTAINER );
          debugf( "User requested to start session(s).\n" );
          
          record = (MINIRECORDCORE *) WinSendMsg( theWin,
           CM_QUERYRECORDEMPHASIS, MPFROMP( CMA_FIRST ),
           MPFROMSHORT( CRA_SOURCE ) );
          
          if ( record && (LONG)record != -1 &&
                !(record->flRecordAttr & CRA_SELECTED) )
          {
            // Source emphasis, but not selected, act only on this object
            onlyOneRecord = 1;
          } else {
            onlyOneRecord = 0;
            record = (MINIRECORDCORE *)CMA_FIRST;
          }
          
          do {
            if ( !onlyOneRecord )
            {
              record = (MINIRECORDCORE *) WinSendMsg( theWin,
               CM_QUERYRECORDEMPHASIS, MPFROMP( record ),
               MPFROMSHORT( CRA_SELECTED ) );
              if ( !record || (LONG)record == -1 ) break;
            
              debugf( "Found a selected record.\n" );
            }
            
            for ( i=0; i<(*numItems); ++i )
            {
              if ( (*info)[i].userName && (*info)[i].record == record )
              {
                if ( !(record->flRecordAttr & CRA_INUSE) )
                {
                  WinSendMsg( theWin, CM_SETRECORDEMPHASIS, MPFROMP( record ),
                   MPFROM2SHORT( TRUE, CRA_INUSE ) );
                  (*info)[i].threadID = _beginthread(
                   (void (*) (void *))sessionThread, NULL, 65536, (*info) + i );
                  debugf( "Session started in thread %d.\n",
                   (*info)[i].threadID );
                  if ( (*info)[i].settings.settingFlags[ EVENT_STARTSESSION ] &
                        SET_SOUND_ENABLED )
                  {
                    mciPlayFile( HWND_DESKTOP, (*info)[i].settings.
                     sounds[ SET_SOUND_ENABLED ], MCI_ASYNC, NULL, NULLHANDLE );
                  }
                } else {
                  debugf( "Attempted to launch an already active session for %s.\n",
                   (*info)[i].userName );
                  if ( (*info)[i].theSession )
                  {
                    WinSetFocus( HWND_DESKTOP,
                     (*info)[i].theSession->getBuddyListWindow() );
                  }
                  // Bring the existing session to the front
                }
                break;
              }
            }
            
            if ( onlyOneRecord ) break;
          } while ( 1 );
        }
        break;
        case MRM_EditSession:
        {
          HMODULE dll = getModuleHandle();
          HWND dlgWin;
          RECTL desktopSize, dlgSize;
          sessionEditorInit initData;
          ULONG i, myEntry;
          MINIRECORDCORE *record;
          ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
          sessionThreadInfo **info =
           (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
          
          record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
           WID_SESSION_CONTAINER, CM_QUERYRECORDEMPHASIS,
           MPFROMLONG( CMA_FIRST ), MPFROMSHORT( CRA_SOURCE ) );
          
          if ( !record || (LONG)record == -1 )
          {
            record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
             WID_SESSION_CONTAINER, CM_QUERYRECORDEMPHASIS,
             MPFROMLONG( CMA_FIRST ), MPFROMSHORT( CRA_SELECTED ) );
            
            if ( !record || (LONG)record == -1 )
            {
              debugf( "Couldn't find the selected record!\n" );
              WinAlarm( HWND_DESKTOP, WA_ERROR );
              return NULL;
            }
          } else {
            WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
             CM_SETRECORDEMPHASIS, MPFROMP( record ),
             MPFROM2SHORT( FALSE, CRA_SOURCE ) );
          }
          
          myEntry = *numItems;
          for ( i=0; i<(*numItems); ++i )
          {
            if ( (*info)[i].userName && (*info)[i].record == record )
            {
              myEntry = i;
              break;
            }
          }
          
          if ( myEntry == *numItems )
          {
            debugf( "Couldn't figure out the selected record!\n" );
            WinAlarm( HWND_DESKTOP, WA_ERROR );
            return NULL;
          }
          
          if ( ((*info)[myEntry].theSession || (*info)[myEntry].threadID) &&
                !WinQueryWindowULong( win, 56 ) )
          {
            // Session is running... why are we editting it?!
            char *titleString;
            
            CHECKED_MALLOC( 24 + strlen( (*info)[myEntry].userName ), titleString );
            
            sprintf( titleString, "Cannot modify session: %s",
             (*info)[myEntry].userName );
            
            if ( (*info)[myEntry].settings.settingFlags[ EVENT_ERRORBOX ] &
                  SET_SOUND_ENABLED )
            {
              mciPlayFile( HWND_DESKTOP, (*info)[myEntry].settings.
               sounds[ EVENT_ERRORBOX ], MCI_ASYNC, NULL, NULLHANDLE );
            }
            WinMessageBox( HWND_DESKTOP, win,
             "Session is currently opened and running.  You cannot modify it.  Please close the session and then proceed with your edits.",
             titleString, 503,
             MB_CANCEL | MB_APPLMODAL | MB_MOVEABLE );
            CHECKED_FREE( titleString );
            return NULL;
          }
                  
          initData.userName = (*info)[myEntry].userName;
          initData.password = (*info)[myEntry].password;
          
          dlgWin = WinLoadDlg( HWND_DESKTOP, win, pmSessionEditor,
           dll, 1, &initData );
          WinQueryWindowRect( HWND_DESKTOP, &desktopSize );
          WinQueryWindowRect( dlgWin, &dlgSize );
          
          WinSetWindowPos( dlgWin, HWND_TOP,
           desktopSize.xLeft + (
            ((desktopSize.xRight - desktopSize.xLeft) -
             (dlgSize.xRight - dlgSize.xLeft)) / 2),
           desktopSize.yBottom + (
            ((desktopSize.yTop - desktopSize.yBottom) -
             (dlgSize.yTop - dlgSize.yBottom)) / 2),
           0, 0, SWP_MOVE );
          // Center the dialog window
          
          if ( WinProcessDlg( dlgWin ) == DID_OK )
          {
            char *text, *tmpString;
            ULONG txtLen;
            
            debugf( "Modifying a session record.\n" );
            
            txtLen = WinQueryDlgItemTextLength( dlgWin, MRM_EN_ScreenName ) + 1;
            CHECKED_MALLOC( txtLen, text );
            WinQueryDlgItemText( dlgWin, MRM_EN_ScreenName, txtLen, text );
            tmpString = (*info)[myEntry].userName;
            (*info)[myEntry].userName = text;
            record->pszIcon = text;
            CHECKED_FREE( tmpString );
            WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
             CM_INVALIDATERECORD, MPFROMP( &record ),
             MPFROM2SHORT( 1, CMA_TEXTCHANGED ) );
            txtLen = WinQueryDlgItemTextLength( dlgWin, MRM_EN_Password ) + 1;
            CHECKED_MALLOC( txtLen, text );
            WinQueryDlgItemText( dlgWin, MRM_EN_Password, txtLen, text );
            tmpString = (*info)[myEntry].password;
            (*info)[myEntry].password = text;
            CHECKED_FREE( tmpString );
          }
          
          WinDestroyWindow( dlgWin );
        }
        break;
      }
    break;
    case WM_GIMMEDROPCOPYDATA:
    {
      MINIRECORDCORE *theRecord;
      char *sharedMem;
      ULONG i;
      ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
      sessionThreadInfo **info =
       (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
      HMTX infoMux = WinQueryWindowULong( win, 12 );
      
      debugf( "Another window requested dragged data for copy operation.\n" );
      
      sharedMem = (char *) PVOIDFROMMP( mp1 );
      theRecord = (MINIRECORDCORE *) PVOIDFROMMP( mp2 );
      
      DosGetSharedMem( sharedMem, PAG_WRITE );
      DosRequestMutexSem( infoMux, SEM_INDEFINITE_WAIT );
      for ( i=0; i<(*numItems); ++i )
      {
        if ( (*info)[i].userName && (*info)[i].record == theRecord )
        {
          strcpy( sharedMem, (*info)[i].userName );
          strcpy( sharedMem + strlen( (*info)[i].userName ) + 1,
           (*info)[i].password );
          break;
        }
      }
      DosReleaseMutexSem( infoMux );
      DosFreeMem( sharedMem );
    }
    break;
    case WM_WINPPCHANGED:
    {
      HINI iniFile = WinQueryWindowULong( win, 20 );
      ULONG ppType = LONGFROMMP( mp2 );
      ULONG lenRet;
      char keyName[10];
      char buffer[1024];
      HWND theWin;
      // Presentation parameter changed on a subclassed child window
      
      if ( !WinQueryWindowULong( win, 52 ) ) break;
      // Already handling button presentation parameter changes.  Just fall
      //  through to the default window procedure.
      
      if ( !iniFile ) break;
      // INI file hasn't been opened yet - PresParams changed during WM_CREATE
      
      if ( SHORT1FROMMP( mp1 ) )
      {
        theWin = WinWindowFromID( win, SHORT1FROMMP( mp1 ) );
      } else {
        theWin = win;
      }
      
      if ( ppType == PP_FONTNAMESIZE && SHORT1FROMMP( mp1 ) > 100 &&
            SHORT1FROMMP( mp1 ) < 105 )
      {
        // May need to resize buttons
        RECTL winRect;
        WinQueryWindowRect( win, &winRect );
        WinPostMsg( win, WM_SIZE,
         MPFROM2SHORT( winRect.xRight - winRect.xLeft,
                       winRect.yTop - winRect.yBottom ),
         MPFROM2SHORT( winRect.xRight - winRect.xLeft,
                       winRect.yTop - winRect.yBottom ) );
      }
      
      lenRet = WinQueryPresParam( theWin, ppType, 0, NULL, 1024, buffer, 0 );
      // No way to query PP size that I know of.  Goofy.
      
      if ( (ppType == PP_BACKGROUNDCOLORINDEX ||
            ppType == PP_BACKGROUNDCOLOR) && SHORT1FROMMP( mp1 ) == 0 )
      {
        WinInvalidateRect( win, NULL, TRUE );
        // Make sure background is repainted
      }
      
      if ( lenRet )
      {
        sprintf( keyName, "%d", ppType );
        
        if ( theWin == win )
        {
          PrfWriteProfileData( iniFile, "Session Manager Presentation Parameters",
           keyName, buffer, lenRet );
        } else {
          if ( SHORT1FROMMP( mp1 ) > 100 && SHORT1FROMMP( mp1 ) < 105 )
          {
            // One of the buttons.  Clone the presentation parameters across
            //  all buttons.
            WinSetWindowULong( win, 52, 0 );
            // Prevent recursion.
            
            PrfWriteProfileData( iniFile, "Session Manager Presentation Parameters (Buttons)",
             keyName, buffer, lenRet );
            WinSetPresParam( WinWindowFromID( win, MRM_StartSession ),
             ppType, lenRet, buffer );
            WinSetPresParam( WinWindowFromID( win, MRM_CreateSession ),
             ppType, lenRet, buffer );
            WinSetPresParam( WinWindowFromID( win, MRM_EditSession ),
             ppType, lenRet, buffer );
            WinSetPresParam( WinWindowFromID( win, MRM_DeleteSession ),
             ppType, lenRet, buffer );
            
            WinSetWindowULong( win, 52, 1 );
          } else {
            PrfWriteProfileData( iniFile, "Session Manager Presentation Parameters (Container)",
             keyName, buffer, lenRet );
          }
        }
      }
    }
    break;
    case WM_PRESPARAMCHANGED:
      WinPostMsg( win, WM_WINPPCHANGED, 0, mp1 );
    break;
    case WM_CLOSE:
    {
      HWND theWin;
      MINIRECORDCORE *record;
      HPOINTER icon;
      ULONG dataLen, bigDataLen, i, txtLen, numRealSessions;
      unsigned char *prfData;
      HINI iniFile = WinQueryWindowULong( win, 20 );
      ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
      ULONG cursoredItem;
      sessionThreadInfo **info =
       (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
      HWND menuWin, frameWin;
      char keyName[80];
      SWP swp;
      
      for ( i=0; i<(*numItems); ++i )
      {
        if ( (*info)[i].userName && (*info)[i].theSession )
        {
          debugf( "Telling session %d to close due to Session Manager exit.\n",
           i );
          (*info)[i].theSession->closeSession();
        }
      }
      
      debugf( "Saving changes to profile.\n" );
      
      frameWin = WinQueryWindow( win, QW_PARENT );
      theWin = WinWindowFromID( win, WID_SESSION_CONTAINER );
      WinQueryWindowPos( frameWin, &swp );
      if ( (swp.fl & SWP_MINIMIZE) || (swp.fl & SWP_MAXIMIZE) )
      {
        swp.x = WinQueryWindowUShort( frameWin, QWS_XRESTORE );
        swp.y = WinQueryWindowUShort( frameWin, QWS_YRESTORE );
        swp.cx = WinQueryWindowUShort( frameWin, QWS_CXRESTORE );
        swp.cy = WinQueryWindowUShort( frameWin, QWS_CYRESTORE );
      }
      PrfWriteProfileData( iniFile, "Session Manager", "Window position data",
       &swp, sizeof( SWP ) );
      
      bigDataLen = 0;
      for ( i=0; i<*numItems; ++i )
      {
        if ( (*info)[i].userName )
        {
          dataLen = 4 + strlen( (*info)[i].userName ) +
           strlen( (*info)[i].password );
          if ( dataLen > bigDataLen ) bigDataLen = dataLen;
        }
      }
      
      CHECKED_MALLOC( bigDataLen, prfData );
      numRealSessions = 0;
      
      record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD,
       MPFROMSHORT( CMA_FIRST ), MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER ) );
      while ( record && (LONG) record != -1 )
      {
        for ( i=0; i<(*numItems); ++i )
        {
          if ( (*info)[i].record == record )
          {
            if ( (*info)[i].userName )
            {
              dataLen = 0;
              txtLen = strlen( (*info)[i].userName );
              *((USHORT *)(prfData + dataLen)) = (USHORT) txtLen;
              dataLen += 2;
              strncpy( (char *)(prfData + dataLen), (*info)[i].userName,
               txtLen );
              dataLen += txtLen;
              txtLen = strlen( (*info)[i].password );
              *((USHORT *)(prfData + dataLen)) = (USHORT) txtLen;
              dataLen += 2;
              strncpy( (char *)(prfData + dataLen),
               (char *)aim_encode_password( (*info)[i].password ), txtLen );
              dataLen += txtLen;
              numRealSessions++;
              debugf( "Saving record: %s.\n", (*info)[i].userName );
              sprintf( keyName, "Session %d data", numRealSessions );
              PrfWriteProfileData( iniFile, "Session Manager", keyName,
               prfData, dataLen );
            }
            break;
          }
        }
        record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD,
         MPFROMP( record ), MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER ) );
      }

      PrfWriteProfileData( iniFile, "Session Manager", "Number of Sessions",
       &numRealSessions, 4 );
      
      cursoredItem = 0;
      i = 0;
      
      record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD, 
       NULL, MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER ) );
      
      while ( record && (LONG)record != -1 )
      {
        ++i;
        if ( record->flRecordAttr & CRA_CURSORED )
        {
          cursoredItem = i;
          // 1-based
        }
        
        record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD, 
         MPFROMP( record ), MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER ) );
      }
      
      debugf( "Deleting all container objects.\n" );
      do {
        record = (MINIRECORDCORE *) WinSendMsg( theWin, CM_QUERYRECORD, NULL,
         MPFROM2SHORT( CMA_LAST, CMA_ITEMORDER ) );
        if ( !record || (LONG)record == -1 ) break;
        
        WinSendMsg( theWin, CM_REMOVERECORD, MPFROMP( &record ),
         MPFROM2SHORT( 1, CMA_FREE ) );
      } while ( 1 );
      // Need a separate loop here because if we remove the cursored item,
      //  then the next item *becomes* cursored.  So do the remove in a
      //  separate step.
      
      PrfWriteProfileData( iniFile, "Session Manager", "Cursored Record",
       &cursoredItem, 4 );
      
      WinDestroyWindow( theWin );
      WinDestroyWindow( WinWindowFromID( win, MRM_StartSession ) );
      WinDestroyWindow( WinWindowFromID( win, MRM_CreateSession ) );
      WinDestroyWindow( WinWindowFromID( win, MRM_EditSession ) );
      WinDestroyWindow( WinWindowFromID( win, MRM_DeleteSession ) );
      
      menuWin = WinQueryWindowULong( win, 24 );
      WinDestroyWindow( menuWin );
      menuWin = WinQueryWindowULong( win, 28 );
      WinDestroyWindow( menuWin );
      
      icon = WinQueryWindowULong( win, 16 );
      if ( icon )
      {
        WinDestroyPointer( icon );
      }
      
      debugf( "Session manager window is closing.\n" );
    }
    break;
    case WM_MENUSELECT:
    {
      USHORT command = SHORT1FROMMP( mp1 );
      
      if ( command == MRM_EditSession || command == MRM_DeleteSession ||
            command == MRM_StartSession || command == MRM_SessionSettings )
      {
        if ( SHORT2FROMMP( mp1 ) )
        {
          // Menu item was selected and WM_COMMAND was posted as a result
          WinSetWindowULong( win, 60, 1 );
        } else {
          WinSetWindowULong( win, 60, 0 );
        }
      }
    }
    break;
    case WM_MENUEND:
    {
      if ( !WinQueryWindowULong( win, 60 ) )
      {
        // If a context menu did not invoke a WM_COMMAND, the remove the source
        //  emphasis on the item for which it was invoked.  Otherwise, leave the
        //  source emphasis in place, and it will be removed when the WM_COMMAND
        //  gets processed.
        
        MINIRECORDCORE *record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
         WID_SESSION_CONTAINER, CM_QUERYRECORDEMPHASIS, MPFROMP( CMA_FIRST ),
         MPFROMSHORT( CRA_SOURCE ) );
        
        if ( record && (LONG)record != -1 )
        {
          WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
           CM_SETRECORDEMPHASIS, MPFROMP( record ),
           MPFROM2SHORT( FALSE, CRA_SOURCE ) );
        } else {
          WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
           CM_SETRECORDEMPHASIS, MPFROMP( NULL ),
           MPFROM2SHORT( FALSE, CRA_SOURCE ) );
        }
      }
    }
    break;
    case WM_CONTROL:
    {
      if ( SHORT1FROMMP( mp1 ) == WID_SESSION_CONTAINER )
      {
        switch ( SHORT2FROMMP( mp1 ) )
        {
          case CN_CONTEXTMENU:
          {
            MINIRECORDCORE *record = (MINIRECORDCORE *) PVOIDFROMMP( mp2 );
            HWND menuWin;
            POINTL pt;
            ULONG numSelected = WinQueryWindowULong( win, 0 );
            
            if ( !record )
            {
              // Pop-up menu for background window
              menuWin = WinQueryWindowULong( win, 28 );
            } else {
              // Pop-up on an object
              menuWin = WinQueryWindowULong( win, 24 );
              if ( numSelected > 1 && (record->flRecordAttr & CRA_SELECTED) )
              {
                // Attempted to select more than one.  Disable edit.
                WinEnableMenuItem( menuWin, MRM_EditSession, FALSE );
              } else {
                WinEnableMenuItem( menuWin, MRM_EditSession, TRUE );
              }
            }
            
            WinQueryPointerPos( HWND_DESKTOP, &pt );
            
            WinSendMsg( WinWindowFromID( win, WID_SESSION_CONTAINER ),
             CM_SETRECORDEMPHASIS, MPFROMP( record ),
             MPFROM2SHORT( TRUE, CRA_SOURCE ) );
            
            WinSetWindowULong( win, 60, 0 );
            
            WinPopupMenu( HWND_DESKTOP, win, menuWin, pt.x, pt.y, 0,
             PU_KEYBOARD | PU_MOUSEBUTTON1 | PU_MOUSEBUTTON2 );
          }
          break;
          case CN_REALLOCPSZ:
          {
            CNREDITDATA *editData = (CNREDITDATA *) PVOIDFROMMP( mp2 );
            MINIRECORDCORE *record = (MINIRECORDCORE *) editData->pRecord;
            ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
            sessionThreadInfo **info =
             (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
            sessionThreadSettings *mySettings = NULL;
            ULONG i;
            
            for ( i=0; i<(*numItems); ++i )
            {
              if ( (*info)[i].userName && (*info)[i].record == record )
              {
                mySettings = &((*info)[i].settings);
              }
            }
            
            if ( record->flRecordAttr & CRA_INUSE )
            {
              // Disallow editting the title if it's in use
              char *titleString;
              CHECKED_MALLOC( 24 + strlen( *(editData->ppszText) ), titleString );
              sprintf( titleString, "Cannot modify session: %s",
               *(editData->ppszText) );
              if ( mySettings && (mySettings->settingFlags[ EVENT_ERRORBOX ] &
                    SET_SOUND_ENABLED) )
              {
                mciPlayFile( HWND_DESKTOP, mySettings->sounds[ EVENT_ERRORBOX ],
                 MCI_ASYNC, NULL, NULLHANDLE );
              }
              WinMessageBox( HWND_DESKTOP, win,
               "Session is currently opened and running.  You cannot modify it.  Please close the session and then proceed with your edits.",
               titleString, 503,
               MB_CANCEL | MB_APPLMODAL | MB_MOVEABLE );
              CHECKED_FREE( titleString );
              return MRFROMLONG( FALSE );
            }
            
            for ( i=0; i<(*numItems); ++i )
            {
              if ( (*info)[i].userName && (*info)[i].record == record )
              {
                char *tmpString, *newString;
                tmpString = (*info)[i].userName;
                CHECKED_MALLOC( editData->cbText, newString );
                (*info)[i].userName = newString;
                *(editData->ppszText) = newString;
                CHECKED_FREE( tmpString );
                return MRFROMLONG( TRUE );
              }
            }
            return MRFROMLONG( FALSE );
          }
          case CN_EMPHASIS:
          {
            NOTIFYRECORDEMPHASIS *emphasis;
            emphasis = (NOTIFYRECORDEMPHASIS *) PVOIDFROMMP( mp2 );
            if ( emphasis->fEmphasisMask & CRA_SELECTED )
            {
              // Something was selected or de-selected
              MINIRECORDCORE *record = (MINIRECORDCORE *) emphasis->pRecord;
              ULONG selectedCount = WinQueryWindowULong( win, 0 );
              HWND theWin;
              
              if ( record->flRecordAttr & CRA_SELECTED )
              {
                if ( !selectedCount )
                {
                  // Enable options that require a selection
                  WinEnableWindow( WinWindowFromID( win, MRM_StartSession ),
                   TRUE );
                  WinEnableWindow( WinWindowFromID( win, MRM_EditSession ),
                   TRUE );
                  WinEnableWindow( WinWindowFromID( win, MRM_DeleteSession ),
                   TRUE );
                  theWin = WinQueryWindow( win, QW_PARENT );
                  theWin = WinWindowFromID( theWin, FID_MENU );
                  WinEnableMenuItem( theWin, MRM_StartSession, TRUE );
                  WinEnableMenuItem( theWin, MRM_EditSession, TRUE );
                  WinEnableMenuItem( theWin, MRM_DeleteSession, TRUE );
                  WinEnableMenuItem( theWin, MRM_SessionSettings, TRUE );
                } else {
                  theWin = WinQueryWindow( win, QW_PARENT );
                  theWin = WinWindowFromID( theWin, FID_MENU );
                  WinEnableWindow( WinWindowFromID( win, MRM_EditSession ),
                   FALSE );
                  WinEnableMenuItem( theWin, MRM_EditSession, FALSE );
                  // Can only edit one entry
                }
                selectedCount++;
              } else {
                selectedCount--;
                if ( !selectedCount )
                {
                  // Disable options that require a selection
                  WinEnableWindow( WinWindowFromID( win, MRM_StartSession ),
                   FALSE );
                  WinEnableWindow( WinWindowFromID( win, MRM_EditSession ),
                   FALSE );
                  WinEnableWindow( WinWindowFromID( win, MRM_DeleteSession ),
                   FALSE );
                  theWin = WinQueryWindow( win, QW_PARENT );
                  theWin = WinWindowFromID( theWin, FID_MENU );
                  WinEnableMenuItem( theWin, MRM_StartSession, FALSE );
                  WinEnableMenuItem( theWin, MRM_EditSession, FALSE );
                  WinEnableMenuItem( theWin, MRM_DeleteSession, FALSE );
                  WinEnableMenuItem( theWin, MRM_SessionSettings, FALSE );
                } else if ( selectedCount == 1 )
                {
                  WinEnableWindow( WinWindowFromID( win, MRM_EditSession ),
                   TRUE );
                  theWin = WinQueryWindow( win, QW_PARENT );
                  theWin = WinWindowFromID( theWin, FID_MENU );
                  WinEnableMenuItem( theWin, MRM_EditSession, TRUE );
                  // Can only edit one entry
                }
              }
              WinSetWindowULong( win, 0, selectedCount );
            }
          }
          break;
          case CN_ENTER:
            if ( WinQueryWindowULong( win, 0 ) )
            {
              // If something is selected, pretend we hit the Start button
              WinPostMsg( win, WM_COMMAND, MPFROMSHORT( MRM_StartSession ),
               NULL );
            }
          break;
          case CN_INITDRAG:
          {
            DRAGINFO *dragInfo;
            DRAGIMAGE dragImage;
            DRAGITEM dragItem;
            CNRDRAGINIT *dragInitInfo = (CNRDRAGINIT *) PVOIDFROMMP( mp2 );
            
            if ( !dragInitInfo->pRecord ) break;
            debugf( "Dragging entry %s.\n", dragInitInfo->pRecord->pszIcon );
             
            dragItem.hwndItem = win;
            dragItem.hstrType = DrgAddStrHandle( DRT_UNKNOWN );
            dragItem.hstrRMF = DrgAddStrHandle( "<DRM_DDE,DRF_TEXT>" );
            dragItem.hstrContainerName =
             DrgAddStrHandle( "Mr. Message Session Manager" );
            dragItem.cxOffset = 0;
            dragItem.cyOffset = 0;
            dragItem.fsControl = 0;
            dragItem.fsSupportedOps = DO_MOVEABLE | DO_COPYABLE;
            dragImage.cb = sizeof(DRAGIMAGE);
            dragImage.cptl = 0;
            dragImage.hImage = WinQueryWindowULong( win, 16 );
            dragImage.sizlStretch.cx = 0;
            dragImage.sizlStretch.cy = 0;
            dragImage.fl = DRG_ICON;
            dragImage.cxOffset = 0;
            dragImage.cyOffset = 0;
            dragItem.ulItemID = (ULONG) dragInitInfo->pRecord;
            dragItem.hstrSourceName =
             DrgAddStrHandle( dragInitInfo->pRecord->pszIcon );
            dragItem.hstrTargetName =
             DrgAddStrHandle( dragInitInfo->pRecord->pszIcon );
            
            dragInfo = DrgAllocDraginfo( 1 );
            dragInfo->usOperation = DO_DEFAULT;
            
            DrgSetDragitem( dragInfo, &dragItem, sizeof( DRAGITEM ), 0 );
            
            debugf( "Initiating direct manipulation (drag-n-drop).\n" );
            DrgDrag( win, dragInfo, &dragImage, 1, VK_ENDDRAG, NULL );
            
            DrgDeleteDraginfoStrHandles( dragInfo );
            DrgFreeDraginfo( dragInfo );
            debugf( "Freed direct manipulation resources.\n" );
          }
          break;
          case CN_DRAGOVER:
          {
            CNRDRAGINFO *cnrDragInfo = (CNRDRAGINFO *) PVOIDFROMMP( mp2 );
            DRAGINFO *dragInfo = cnrDragInfo->pDragInfo;
            DRAGITEM *dragItem;
            USHORT defaultAction = DO_MOVE;
            ULONG containerNameLen;
            const char *acceptContainerString = "Mr. Message Session Manager";
            char buffer[28];
            
            if ( !DrgAccessDraginfo( dragInfo ) )
            {
              return MRFROM2SHORT( DOR_NEVERDROP, defaultAction );
            }
            
            dragItem = DrgQueryDragitemPtr( dragInfo, 0 );
            containerNameLen = DrgQueryStrName(
             dragItem->hstrContainerName, 28, buffer );
            
            if ( containerNameLen < strlen( acceptContainerString ) )
            {
              return MRFROM2SHORT( DOR_NODROPOP, 0 );
            } else {
              if ( strcmp( buffer, acceptContainerString ) != 0 )
              {
                return MRFROM2SHORT( DOR_NODROPOP, 0 );
              }
            }
            
            if ( dragInfo->hwndSource != win )
            {
              defaultAction = DO_COPY;
            }
            
            if ( cnrDragInfo->pRecord == NULL )
            {
              DrgFreeDraginfo( dragInfo );
              return MRFROM2SHORT( DOR_DROP, defaultAction );
            }
            
            DrgFreeDraginfo( dragInfo );
            return MRFROM2SHORT( DOR_NODROP, defaultAction );
          }
          case CN_DROP:
          {
            CNRDRAGINFO *cnrDragInfo = (CNRDRAGINFO *) PVOIDFROMMP( mp2 );
            DRAGINFO *dragInfo = cnrDragInfo->pDragInfo;
            DRAGITEM *dragItem;
            MINIRECORDCORE *theRecord, *record;
            RECORDINSERT insert;
            
            if ( !DrgAccessDraginfo( dragInfo ) )
            {
              return NULL;
            }
            
            if ( cnrDragInfo->pRecord != NULL )
            {
              // Can't drop on an existing entry.  No meaning to that here.
              return NULL;
            }
            
            if ( dragInfo->usOperation == DO_COPY )
            {
              // Wait for a subsequent sent message to create the copy
              char *sharedMem;
              HPOINTER icon;
              char *text;
              ULONG len, len2, myEntry;
              HMTX infoMux = WinQueryWindowULong( win, 12 );
              ULONG *numItems = (ULONG *) WinQueryWindowPtr( win, 4 );
              sessionThreadInfo **info =
               (sessionThreadInfo **) WinQueryWindowPtr( win, 8 );
              
              dragItem = DrgQueryDragitemPtr( dragInfo, 0 );
              theRecord = (MINIRECORDCORE *) dragItem->ulItemID;
              
              DosAllocSharedMem( (PVOID *)&sharedMem, NULL, 4096, OBJ_GETTABLE |
               PAG_COMMIT | PAG_READ | PAG_WRITE );
              
              debugf( "Drag source was a different window.  Copying instead of moving.\n" );
              DrgSendTransferMsg( dragInfo->hwndSource, WM_GIMMEDROPCOPYDATA,
               MPFROMP( sharedMem ), MPFROMP( theRecord ) );
              
              record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
               WID_SESSION_CONTAINER, CM_ALLOCRECORD, MPFROMLONG( 0 ),
               MPFROMSHORT( 1 ) );
              if ( !record )
              {
                debugf( "Error allocating record in container.\n" );
                DosFreeMem( sharedMem );
                return 0;
              }
              
              icon = WinQueryWindowULong( win, 16 );
              
              len = strlen( sharedMem );
              if ( !len )
              {
                DosFreeMem( sharedMem );
                break;
              }
              
              CHECKED_MALLOC( len + 1, text );
              strcpy( text, sharedMem );
              
              debugf( "Entry %s is being copied.\n", text );
              
              record->flRecordAttr = CRA_CURSORED;
              record->ptlIcon.x = 0;
              record->ptlIcon.y = 0;
              record->pszIcon = text;
              record->hptrIcon = icon;
              insert.cb = sizeof( RECORDINSERT );
              insert.pRecordOrder = (RECORDCORE *)CMA_END;
              insert.pRecordParent = NULL;
              insert.fInvalidateRecord = TRUE;
              insert.zOrder = CMA_TOP;
              insert.cRecordsInsert = 1;
              
              WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INSERTRECORD,
               MPFROMP( record ), MPFROMP( &insert ) );
              
              DosRequestMutexSem( infoMux, SEM_INDEFINITE_WAIT );
              if ( !(*numItems) )
              {
                CHECKED_MALLOC( sizeof( sessionThreadInfo ), *info );
                myEntry = 0;
                (*numItems)++;
              } else {
                ULONG i;
                myEntry = *numItems;
                for ( i=0; i<*numItems; ++i )
                {
                  if ( !(*info)[i].userName ) myEntry = i;
                }
                if ( *numItems == myEntry )
                {
                  CHECKED_REALLOC( *info, ((*numItems) + 1) * sizeof( sessionThreadInfo ), *info );
                  (*numItems)++;
                }
              }
              
              (*info)[ myEntry ].userName = text;
              DosReleaseMutexSem( infoMux );
            
              len2 = strlen( sharedMem + len + 1 );
              CHECKED_MALLOC( len2 + 1, text );
              strcpy( text, sharedMem + len + 1 );
              
              DosFreeMem( sharedMem );
              
              (*info)[ myEntry ].password = text;
              (*info)[ myEntry ].managerWin = win;
              (*info)[ myEntry ].theSession = NULL;
              (*info)[ myEntry ].record = record;
              (*info)[ myEntry ].threadID = 0;
            } else if ( dragInfo->usOperation == DO_MOVE )
            {
              // Move means that it originated in this very window and process.
              //  So our dragItem should give us the MINIRECORDCORE pointer
              //  which we can reposition as requested by the user.
              RECTL itemRect, theRect;
              QUERYRECORDRECT qrr;
              LONG tmpL;
              POINTL dropPt;
              MINIRECORDCORE *prevRecord = (MINIRECORDCORE *)CMA_FIRST;
              
              dragItem = DrgQueryDragitemPtr( dragInfo, 0 );
              theRecord = (MINIRECORDCORE *) dragItem->ulItemID;
              dropPt.x = dragInfo->xDrop;
              dropPt.y = dragInfo->yDrop;
              WinMapWindowPoints( HWND_DESKTOP,
               WinWindowFromID( win, WID_SESSION_CONTAINER ), &dropPt, 1 );
              WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
               CM_QUERYVIEWPORTRECT, MPFROMP( &theRect ),
               MPFROM2SHORT( CMA_WORKSPACE, FALSE ) );
              dropPt.x += theRect.xLeft;
              dropPt.y += theRect.yBottom;
              
              debugf( "Dropped at %d, %d (container workspace coordinates).\n",
               dropPt.x, dropPt.y );
              
              record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
               WID_SESSION_CONTAINER, CM_QUERYRECORD, MPFROMSHORT( CMA_FIRST ),
               MPFROM2SHORT( CMA_FIRST, CMA_ITEMORDER ) );
              qrr.cb = sizeof( QUERYRECORDRECT );
              qrr.fRightSplitWindow = FALSE;
              qrr.fsExtent = CMA_ICON | CMA_TEXT;
              while ( record && (LONG) record != -1 )
              {
                qrr.pRecord = (RECORDCORE *)record;
                WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
                 CM_QUERYRECORDRECT, MPFROMP( &itemRect ), MPFROMP( &qrr ) );
                 
                if ( dropPt.y > itemRect.yTop ) break;
                // Insert the record before this one here
                
                if ( dropPt.y <= itemRect.yTop &&
                      dropPt.y >= itemRect.yBottom )
                {
                  tmpL = itemRect.xRight - itemRect.xLeft;
                  if ( dropPt.x < itemRect.xLeft + (tmpL / 2) ) break;
                  // Insert the record before this one here
                }
                
                prevRecord = record;
                record = (MINIRECORDCORE *) WinSendDlgItemMsg( win,
                 WID_SESSION_CONTAINER, CM_QUERYRECORD, MPFROMP( record ),
                 MPFROM2SHORT( CMA_NEXT, CMA_ITEMORDER ) );
              }
              
              if ( record && (LONG) record != -1 )
              {
                if ( prevRecord == theRecord ) break;
                
                WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_REMOVERECORD,
                 MPFROMP( &theRecord ), MPFROM2SHORT( 1, 0 ) );
                 
                insert.cb = sizeof( RECORDINSERT );
                insert.pRecordOrder = (RECORDCORE *) prevRecord;
                insert.pRecordParent = NULL;
                insert.fInvalidateRecord = TRUE;
                insert.zOrder = CMA_TOP;
                insert.cRecordsInsert = 1;
                WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INSERTRECORD,
                 MPFROMP( theRecord ), MPFROMP( &insert ) );
              } else {
                // Move to end
                
                WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_REMOVERECORD,
                 MPFROMP( &theRecord ), MPFROM2SHORT( 1, 0 ) );
                 
                insert.cb = sizeof( RECORDINSERT );
                insert.pRecordOrder = (RECORDCORE *) CMA_END;
                insert.pRecordParent = NULL;
                insert.fInvalidateRecord = TRUE;
                insert.zOrder = CMA_TOP;
                insert.cRecordsInsert = 1;
                WinSendDlgItemMsg( win, WID_SESSION_CONTAINER, CM_INSERTRECORD,
                 MPFROMP( theRecord ), MPFROMP( &insert ) );
              }
              
              WinSendDlgItemMsg( win, WID_SESSION_CONTAINER,
               CM_INVALIDATERECORD, MPFROMP( &theRecord ),
               MPFROM2SHORT( 1, CMA_REPOSITION ) );
              DrgFreeDraginfo( dragInfo );
            }
          }
          break;
        }
      }
    }
    break;
  }
  return WinDefWindowProc( win, msg, mp1, mp2 );
}

SessionManager :: SessionManager()
{
  QMSG qmsg;
  ULONG frameflgs = FCF_TITLEBAR | FCF_SYSMENU | FCF_SIZEBORDER | 
   FCF_MINMAX | FCF_TASKLIST | FCF_MENU;
  ULONG numItems, i, j;
  sessionThreadInfo *sessions;
  RECTL desktopSize;
  HMODULE dllModule = getModuleHandle();
  HWND mgrWin, clientWin;
  HAB ab;
  HMQ messq;
  SWP swp;
  sessionManagerULONGs initData;
  sessionThreadSettings globalSettings;
  
  debugf( "Session manager started.\n" );
  
  infoMutex = NULLHANDLE;
  if ( DosCreateMutexSem( NULL, &infoMutex, 0, 0 ) )
  {
    debugf( "Could not create mutex semaphore for session manager.\n" );
    return;
  }
  
  debugf( "Session manager mutex semaphore was created.\n" );
  
  ab = WinInitialize( 0 );
  messq = WinCreateMsgQueue( ab, 0 );

  initData.iniFile = PrfOpenProfile( ab, "MRMESSAGE.INI" );
  if ( !initData.iniFile )
  {
    debugf( "Error opening profile!\n" );
    return;
  }
  
  WinRegisterClass( ab, "MrMessage Splash Screen", pmSplashScreen,
   0, 4 );
  // 4 bytes extra storage for HBITMAP
  
  globalSettings.settingFlags[EVENT_SPLASHSCREEN] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_SPLASHSCREEN] = strdup( "coffee.wav" );
  globalSettings.settingFlags[EVENT_STARTSESSION] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_STARTSESSION] = strdup( "vacusuck.wav" );
  globalSettings.settingFlags[EVENT_ERRORBOX] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_ERRORBOX] = strdup( "lookinat.wav" );
  globalSettings.settingFlags[EVENT_FIRSTCOMM] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_FIRSTCOMM] = strdup( "firsttime.wav" );
  globalSettings.settingFlags[EVENT_RECEIVED] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_RECEIVED] = strdup( "boing.wav" );
  globalSettings.settingFlags[EVENT_SENT] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_SENT] = strdup( "twip.wav" );
  globalSettings.settingFlags[EVENT_ARRIVE] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_ARRIVE] = strdup( "drwropen.wav" );
  globalSettings.settingFlags[EVENT_LEAVE] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_LEAVE] = strdup( "drwclose.wav" );
  globalSettings.settingFlags[EVENT_ENDSESSION] = SET_SOUND_ENABLED;
  globalSettings.sounds[EVENT_ENDSESSION] = strdup( "stopped.wav" );
  globalSettings.inheritSettings = 0;
  for ( i=0; i<EVENT_MAXEVENTS; ++i )
  {
    globalSettings.rexxScripts[i] = NULL;
    globalSettings.shellCmds[i] = NULL;
  }
  
  querySessionSettings( initData.iniFile, "Global Settings", &globalSettings );
  
  mgrWin = WinCreateWindow( HWND_DESKTOP, "MrMessage Splash Screen",
   "Mr. Message for OS/2", 0, 0, 0, 0, 0, HWND_DESKTOP, HWND_TOP,
   100, &globalSettings, NULL );
  
  while ( WinGetMsg( hab, &qmsg, NULLHANDLE, 0, 0 ) )
    WinDispatchMsg( hab, &qmsg );
  
  WinDestroyWindow( mgrWin );
  
  WinRegisterClass( ab, "MrMessage Session Manager", pmSessionManager,
   0, 68 );
  // 68 bytes extra storage
  // Offset 0  = number of selected sessions in GUI
  //        4  = pointer to number of sessions
  //        8  = pointer to session thread information
  //        12 = mutex semaphore handle for session thread information
  //        16 = AIM icon pointer handle
  //        20 = INI file handle
  //        24 = session manager popup menu HWND
  //        28 = session manager background popup menu HWND
  //        32 = System-defined window procedure for window ID 100
  //        36 = System-defined window procedure for window ID 101
  //     ...48 = System-defined window procedure for window ID 104
  //        52 = Allow distribution of button pres params (prevent recursion)
  //        56 = Allow edit session when session running
  //        60 = 1 if context menu invoked a WM_COMMAND
  //        64 = Pointer to global settings
  
  WinRegisterClass( ab, "MrMessage Buddy List", pmBuddyList, 0, 64 );
  // 64 bytes extra storage
  // Offset 0  = Pointer to my MINIRECORDCORE entry in the session manager
  // Offset 4  = Semaphore to wake up the session thread (when closing)
  // Offset 8  = Icon handle for away buddy icon
  // Offset 12 = Icon handle for folder for group icon
  // Offset 16 = Window handle of session manager
  // Offset 20 = Icon handle for online buddy icon
  // Offset 24 = Pointer to OscarSession object
  // Offset 28 = Icon handle for unplugged icon
  // Offset 32 = Container window original window proc
  // Offset 36 = Bubble window handle
  // Offset 40 = Icon handle for "hot" online buddy icon
  // Offset 44 = Record used to create bubble help window
  // Offset 48 = Filtration in effect on this list
  // Offset 52 = Icon handle for the big red X
  // Offset 56 = Switch list entry handle
  // Offset 60 = Idle-ness was reported already
  
  WinRegisterClass( ab, "MrMessage Bubble Window", pmBubbleWindow,
   CS_MOVENOTIFY, 24 );
  // 24 bytes extra storage
  // Offset  0 = Pointer to the text contained in the bubble
  // Offset  4 = Non-cached presentation space handle
  // Offset  8 = Memory device context
  // Offset 12 = Memory presentation space
  // Offset 16 = Background bitmap handle
  // Offset 20 = Character line height
  
  WinRegisterClass( ab, "MrMessage Instant Message Window", pmIMwindow, 0, 20 );
  // 20 bytes extra storage
  // Offset  0 = Pointer to Buddy data
  // Offset  4 = Buddy list window handle
  // Offset  8 = Allow send (yes/no depending on recipient online status)
  // Offset 12 = Window procedure for MLE (needed for subclassing)
  // Offset 16 = Switch list entry handle

  WinQueryWindowRect( HWND_DESKTOP, &desktopSize );
  
  mgrWin = WinCreateStdWindow( HWND_DESKTOP, WS_ANIMATE, &frameflgs,
   "MrMessage Session Manager",
   "Mr. Message for OS/2 - Session Manager",
   0, dllModule, MRM_SessionManagerMenu, &clientWin );

  numItems = 0;
  sessions = NULL;
  
  initData.numItemsPtr = &numItems;
  initData.threadInfoPtr = &sessions;
  initData.mutexSem = infoMutex;
  initData.globalSettings = &globalSettings;
  
  swp.x = desktopSize.xLeft + (((desktopSize.xRight - desktopSize.xLeft) -
   SESSION_MGR_WIDTH) / 2);
  swp.y = desktopSize.yBottom + (((desktopSize.yTop - desktopSize.yBottom) -
   SESSION_MGR_HEIGHT) / 2);
  swp.cx = SESSION_MGR_WIDTH;
  swp.cy = SESSION_MGR_HEIGHT;
  // Set up the default information if it's not in the profile

  i = sizeof( SWP );
  PrfQueryProfileData( initData.iniFile, "Session Manager",
   "Window position data", &swp, &i );
  
  WinSetWindowPos( mgrWin, NULLHANDLE, swp.x, swp.y, swp.cx, swp.cy,
   SWP_SIZE | SWP_MOVE );
  
  WinPostMsg( mgrWin, WM_INITULONGS, MPFROMP( &initData ), NULL );

  WinSetFocus( HWND_DESKTOP, mgrWin );

  while ( WinGetMsg( hab, &qmsg, NULLHANDLE, 0, 0 ) )
    WinDispatchMsg( hab, &qmsg );
  
  WinDestroyWindow( mgrWin );
  WinDestroyMsgQueue( messq );
  
  debugf( "Cleaning up and waiting for active sessions to end.\n" );
  
  DosRequestMutexSem( infoMutex, SEM_INDEFINITE_WAIT );
  
  if ( sessions && numItems )
  {
    int activeEntries = 0, entriesCleaned = 0;
    
    for ( i=0; i<numItems; ++i )
    {
      if ( sessions[i].userName )
      {
        entriesCleaned++;
        if ( sessions[i].theSession )
        {
          ULONG theThread = sessions[i].threadID;
          
          // Shut down and clean up the session nicely.  We do this outside of
          //  the message queue loop so that we don't hang up the GUI.
          activeEntries++;
          if ( theThread )
          {
            int rc, timePolled = 0;

            debugf( "Session in thread %d is still active.  Waiting for it to end.\n",
             theThread );
            
            rc = DosWaitThread( &theThread, DCWW_NOWAIT );
            while ( rc == ERROR_THREAD_NOT_TERMINATED &&
                     sessions[i].threadID && timePolled < 5 )
            {
              DosSleep( 1000 );
              timePolled++;
              rc = DosWaitThread( &theThread, DCWW_NOWAIT );
            }
            
            if ( rc == ERROR_THREAD_NOT_TERMINATED && sessions[i].threadID )
            {
              debugf( "Timed out waiting for thread to end.  Killing thread %d.\n",
               theThread );
              DosKillThread( theThread );
              // The session might have been hung up in its destructor, so
              //  don't try to delete it again here.  Just let the memory
              //  potentially leak so we can get out of here safely.
            }
          }
          // This will delete the session for us.
          
          for ( j=0; j<EVENT_MAXEVENTS; ++j )
          {
            if ( sessions[i].settings.sounds[j] )
              CHECKED_FREE( sessions[i].settings.sounds[j] );
            if ( sessions[i].settings.rexxScripts[j] )
              CHECKED_FREE( sessions[i].settings.rexxScripts[j] );
            if ( sessions[i].settings.rexxScripts[j] )
              CHECKED_FREE( sessions[i].settings.shellCmds[j] );
          }
        }
        CHECKED_FREE( sessions[i].userName );
        if ( sessions[i].password ) CHECKED_FREE( sessions[i].password );
      }
    }
    
    DosReleaseMutexSem( infoMutex );
    
    CHECKED_FREE( sessions );
    
    debugf( "Cleaned up %d session entries (%d of which were active).\n",
     entriesCleaned, activeEntries );
  }
  
  PrfCloseProfile( initData.iniFile );
  WinTerminate( ab );
  
  for ( i=0; i<EVENT_MAXEVENTS; ++i )
  {
    if ( globalSettings.sounds[i] )
      CHECKED_FREE( globalSettings.sounds[i] );
    if ( globalSettings.rexxScripts[i] )
      CHECKED_FREE( globalSettings.rexxScripts[i] );
    if ( globalSettings.shellCmds[i] )
      CHECKED_FREE( globalSettings.shellCmds[i] );
  }
}

SessionManager :: ~SessionManager()
{
  debugf( "Session manager exiting.\n" );
  if ( infoMutex )
  {
    debugf( "Closing session manager mutex semaphore.\n" );
    DosCloseMutexSem( infoMutex );
  }
}

