#define INCL_DOS
#define INCL_WIN
#define INCL_DOSERRORS

#include <os2.h>
#include <types.h>
#include <sys\socket.h>
#include <arpa\inet.h>
#include <netinet\in.h>
#include <netdb.h>
#include <unistd.h>
#include <string.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include "oscarsession.h"
#include "sessionmanager.h"

#define RESPONSE_TIMEOUT 5000

#define SET_NEXT_STATE( nextState )                        \
  DosRequestMutexSem( waitStateMux, SEM_INDEFINITE_WAIT ); \
  DosResetEventSem( stateWakeup, (ULONG *)&rc );           \
  waitForState = nextState;                                \
  DosPostEventSem( continuousReceive );                    \
  DosReleaseMutexSem( waitStateMux );

#define ADVANCE_STATE_MACHINE \
  status = waitForState;      \
  waitForState = fatalError;

#define WAIT_CHECK_FOR_ERRORS( timeout, waitFailureCode ) \
  if ( DosWaitEventSem( stateWakeup, timeout ) )          \
  {                                                       \
    errorCode = waitFailureCode;                          \
    status = fatalError;                                  \
    debugf( "%s\n", errorMessageStrings[ errorCode ] );   \
    WinPostMsg( mgrWin, WM_ERRORMESSAGE,                  \
     MPFROMSHORT( errorCode ), MPFROMP( theRecord ) );    \
    return;                                               \
  }                                                       \
  if ( status == fatalError )                             \
  {                                                       \
    debugf( "%s\n", errorMessageStrings[ errorCode ] );   \
    WinPostMsg( mgrWin, WM_ERRORMESSAGE,                  \
     MPFROMSHORT( errorCode ), MPFROMP( theRecord ) );    \
    receiveData.printData();                              \
    return;                                               \
  }


void dataReceiverThread( OscarSession *sessionData )
{
  OscarData *dataPacket = &(sessionData->receiveData);
  UserInformation tmpUserInfo;
  unsigned short AIMerror;
  
  int theSocket, firstIn = 1;
  
  debugf( "Starting data receiver thread.\n" );
  
  while ( 1 )
  {
    ULONG rc;
    
    if ( !firstIn )
    {
      debugf( "Receiver thread waiting.\n" );
      DosWaitEventSem( sessionData->continuousReceive, SEM_INDEFINITE_WAIT );
      debugf( "Receiver thread was told to wake up.\n" );
    } else firstIn = 0;
    
    DosResetEventSem( sessionData->continuousReceive, &rc );
    
    if ( sessionData->status == fatalError ||
          sessionData->status == shutDown ) break;
    
    if ( sessionData->waitForState < BOSconnectionAck )
    {
      theSocket = sessionData->oscarSocket;
      debugf( "Listening to Oscar socket (wait for state %d)\n",
       sessionData->waitForState );
    } else {
      theSocket = sessionData->BOSsocket;
      debugf( "Listening to BOS socket (wait for state %d)\n",
       sessionData->waitForState );
    }
    
    debugf( "Waiting to receive on socket.\n" );
    dataPacket->receiveData( theSocket );
    
    if ( dataPacket->getStatus() == 10054 || dataPacket->getStatus() == -1 )
    {
      debugf( "We're getting booted by AOL.  Bummer.\n" );
      sessionData->status = fatalError;
      WinPostMsg( sessionData->sessionManagerWin, WM_USERBOOTED,
       MPFROMP( sessionData ), NULL );
      break;
      // Close out the receiver thread.  No point in hanging around looping.
    }

    debugf( "Data received.  Previous state: %d, Wait for state: %d.\n",
     sessionData->status, sessionData->waitForState );
    
    if ( dataPacket->isMOTD() )
    {
      debugf( "AOL message of the day was received and promptly ignored.\n" );
      firstIn = 1;
      continue;
    }
    
    AIMerror = dataPacket->getAIMerror();
    
    if ( AIMerror != 0 )
    {
      // May want to report certain AIM errors back to the user eventually.
      // For now they'll just crop up in the debug on the console.
    }
    
    switch ( sessionData->status )
    {
      case noSession:
        // Looking for a connection acknowledgement
        
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( dataPacket->isConnectionAck() &&
              sessionData->waitForState == oscarConnectionAck )
        {
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_OSC_BAD_ACK;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case oscarConnectionAck:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == oscarLoginAuthorization &&
              !dataPacket->getAuthResponseData( &sessionData->authData ) )
        {
          ULONG junk;
          
          DosResetEventSem( sessionData->continuousReceive, &junk );
          // Stops us from receiving after this packet
          
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_OSC_BAD_AUTH;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case nextSocket:
        // This receive call is the result of closing the Oscar socket in
        //  order to change over to the BOS socket.  Eat it without any
        //  nasty-grams.
        debugf( "Switching sockets\n" );
        sessionData->status = oscarLoginAuthorization;
      break;
      case oscarLoginAuthorization:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( dataPacket->isConnectionAck() &&
              sessionData->waitForState == BOSconnectionAck )
        {
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_BAD_ACK;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSconnectionAck:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( dataPacket->isHostReadyMessage(
              &sessionData->supportedFamilies ) &&
              sessionData->waitForState == BOShostReady )
        {
          sessionData->reportAvailableServices();
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_BAD_LOGIN;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOShostReady:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( dataPacket->isFamilyVersions() &&
              sessionData->waitForState == BOSservicesVersion )
        {
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_BAD_SERV_VER;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSservicesVersion:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( !dataPacket->getUserInformation( 0xe, &sessionData->userInfo ) &&
              sessionData->waitForState == BOScurrentUserInfo )
        {
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_SELF_INFO;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOScurrentUserInfo:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSrateInformation &&
              dataPacket->isRateInformation( &sessionData->rateInformation ) )
        {
          // Don't do anything with the rate information for now.
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_RATE_INFO;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSrateInformation:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSserverStoredLimits &&
              dataPacket->isSSIlimits() )
        {
          // Not doing anything with this information for now
          DosPostEventSem( sessionData->stateWakeup );
          debugf( "Received SSI limit information (ignored for now).\n" );
        } else {
          firstIn = 1;
          // Received something else (user info packet probably).  Cycle again.
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSserverStoredLimits:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSserverStoredInfo &&
              !dataPacket->getSSIData( &sessionData->ssiData ) )
        {
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_SSI;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSserverStoredInfo:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSlocationLimits )
        {
          // Not doing anything with this information for now
          DosPostEventSem( sessionData->stateWakeup );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_LOCATION_LIM;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSlocationLimits:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSbuddyManagementLimits )
        {
          // Not doing anything with this information for now
          DosPostEventSem( sessionData->stateWakeup );
          debugf( "Received location service limitations (ignored for now).\n" );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_BLM_LIM;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSbuddyManagementLimits:
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSprivacyLimits )
        {
          // Not doing anything with this information for now
          DosPostEventSem( sessionData->stateWakeup );
          debugf( "Received buddy list management limitations (ignored for now).\n" );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_PRIVACY_LIM;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      break;
      case BOSprivacyLimits:
      {
        DosRequestMutexSem( sessionData->waitStateMux, SEM_INDEFINITE_WAIT );
        // Make sure we post the wakeup for the appropriate waitFor state
        if ( sessionData->waitForState == BOSICBMparams )
        {
          // Not doing anything with this information for now
          DosPostEventSem( sessionData->stateWakeup );
          debugf( "Received ICBM parameters (ignored for now).\n" );
        } else {
          sessionData->status = fatalError;
          sessionData->errorCode = MRM_BOS_NO_PRIVACY_LIM;
          dataPacket->printData();
          DosPostEventSem( sessionData->stateWakeup );
        }
        DosReleaseMutexSem( sessionData->waitStateMux );
      }
      break;
      case BOSserviceLoop:
      {
        char *userName, *message;
        
        // Any number of things can fly across in this state.
        firstIn = 1; // Don't wait for any states before receiving
        
        if ( !dataPacket->getUserInformation( 0, &tmpUserInfo ) )
        {
          UserInformation *userInfo = new UserInformation( tmpUserInfo );
          
          userInfo->printData();
          
          WinPostMsg( sessionData->sessionWindow, WM_BUDDYSTATUS,
           MPFROMP( userInfo ), NULL );
        } else if ( dataPacket->instantMessageReceived( &userName, &message ) )
        {
          // Got an instant message from someone!
          
          WinPostMsg( sessionData->sessionWindow, WM_IMRECEIVED,
           MPFROMP( userName ), MPFROMP( message ) );
        } else if ( dataPacket->isServerPause() )
        {
          ULONG junk;
          DosResetEventSem( sessionData->pauseResume, &junk );
          debugf( "Server requested a pause.  Sending acknowledgement.\n" );
          sessionData->sendData.prepareServerPauseAck();
          sessionData->sendData.sendData( sessionData->BOSsocket );
        } else if ( dataPacket->isServerResume() ) {
          debugf( "Server resume requested.  Resuming communications.\n" );
          DosPostEventSem( sessionData->pauseResume );
        } else {
          debugf( "Unhandled data packet received from AOL server:\n" );
          dataPacket->printData();
        }
      }
      break;
    }
  }
  
  debugf( "Ending data receiver thread.\n" );
  sessionData->dataReceiverThreadID = 0;
}

void rateThread( userRateInformation *rateInfo )
{
  REQUESTDATA rd;
  ULONG dataLen;
  PVOID theData;
  BYTE priority;
  ULONG cTime, lastSend = 0;
  OscarData *queuedData;
  
  debugf( "Rate regulation thread is starting up.\n" );
  
  while ( DosReadQueue( rateInfo->outgoingData, &rd, &dataLen, &theData,
           0, DCWW_WAIT, &priority, 0 ) == 0 )
  {
    // First check if the server needs a breather.  This will stay posted
    //  until it is cleared due to a SNAC 1 sub B.
    DosWaitEventSem( rateInfo->pauseResume, SEM_INDEFINITE_WAIT );
    
    // Data is ready to be sent.  Check if we've waited long enough.
    
    if ( rd.ulData == 0 ) break;
    // That's our signal to shut down
    
    DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &cTime, 4 );
    if ( lastSend > cTime )
    {
      // Millisecond counter wrapped around at 49.7 days
      lastSend = 0;
    }

    if ( cTime - lastSend < rateInfo->clearLevel )
    {
      DosSleep( rateInfo->clearLevel - (cTime - lastSend) );
      debugf( "Slept %d milliseconds\n",
       rateInfo->clearLevel - (cTime - lastSend) );
    }
    
    debugf( "Sending queued data.  Last send: %ld, Queued: %ld, Rate: %ld\n",
     lastSend, cTime, rateInfo->clearLevel );
    
    queuedData = (OscarData *)rd.ulData;
    
    if ( rateInfo->sendSocket )
    {
      queuedData->sendData( rateInfo->sendSocket );
      DosQuerySysInfo( QSV_MS_COUNT, QSV_MS_COUNT, &lastSend, 4 );
    }
    // If the socket was zeroed out, just fast-forward through the rest of the
    //  queue and delete everything
    
    delete queuedData;
  }
  
  debugf( "Rate regulation thread is shutting down.\n" );
}

static void addAllGroupMembers( int entry, Buddy **buddyList,
 SSIData *ssiData, HWND sessionWindow, HEV wakeupSem );
// Defined below

OscarSession :: OscarSession( char *userName, char *password, HWND mgrWin,
 MINIRECORDCORE *theRecord, OscarSession **storeHere,
 sessionThreadSettings *mySettings )
{
  struct hostent *hostEnt;
  struct in_addr oscarAddr;
  struct sockaddr_in serverDesc;
  char *tmpStr;
  int rc, i, j, k, found;
  buddyListCreateData buddyCreateData;
  
  settings = mySettings;
  sendData.setSeqNumPointer( &currentSequence );
  
  sessionManagerWin = mgrWin;
  supportedFamilies = NULL;
  rateInformation = NULL;
  
  currentSequence = 0;
  myRecordInManager = theRecord;
  status = noSession;
  waitForState = fatalError;
  
  sessionWindow = 0;
  dataReceiverThreadID = 0;
  stateWakeup = 0;
  pauseResume = 0;
  
  errorCode = MRM_NO_ERROR;
  BOSsocket = 0;
  buddyList = NULL;
  numBuddies = 0;
  
  sessionShutdown = 0;
  numFlashing = 0;
  
  *storeHere = this;
  
  debugf( "Creating socket.\n" );
  
  oscarSocket = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
  if ( oscarSocket < 0 )
  {
    errorCode = MRM_OSC_SOCKET_OPEN;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  debugf( "Looking up Oscar server host IP address.\n" );
  
  hostEnt = gethostbyname( "login.oscar.aol.com" );
  if ( !hostEnt )
  {
    errorCode = MRM_OSC_GET_OSC_IP;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  memcpy( &oscarAddr.s_addr, hostEnt->h_addr, hostEnt->h_length );
  memset( &serverDesc, 0, sizeof( struct sockaddr_in ) );
  
  serverDesc.sin_len = sizeof( struct sockaddr_in );
  serverDesc.sin_family = AF_INET;
  serverDesc.sin_addr.s_addr = oscarAddr.s_addr;
  serverDesc.sin_port = htons( 5190 ); // Standard OSCAR port
  
  debugf( "Connecting socket to Oscar server.\n" );
  
  rc = connect( oscarSocket, (struct sockaddr *)&serverDesc,
   sizeof( struct sockaddr_in ) );
  
  if ( rc )
  {
    errorCode = MRM_OSC_SOCKET_CONNECT;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  debugf( "Connected to Oscar server.  Waiting for ACK.\n" );
  
  if ( DosCreateEventSem( NULL, &stateWakeup, 0, 0 ) )
  {
    errorCode = MRM_GEN_CREATE_SEM;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  if ( DosCreateEventSem( NULL, &continuousReceive, 0, FALSE ) )
  {
    errorCode = MRM_GEN_CREATE_SEM;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  if ( DosCreateEventSem( NULL, &pauseResume, 0, TRUE ) )
  {
    // Initially posted... clear when we need to pause data sending
    errorCode = MRM_GEN_CREATE_SEM;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  if ( DosCreateMutexSem( NULL, &waitStateMux, 0, 0 ) )
  {
    errorCode = MRM_GEN_CREATE_MTX;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  SET_NEXT_STATE( oscarConnectionAck );
  
  dataReceiverThreadID = _beginthread( (void (*) (void *))dataReceiverThread,
   NULL, 65536, this );
   
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_OSC_ACK_TIMEOUT );
  
  // Got our ACK within the timeout period

  ADVANCE_STATE_MACHINE;
  
  sendData.prepareLoginData( userName, password );
  debugf( "Sending login packet to Oscar server.\n" );
  sendData.sendData( oscarSocket );
  
  debugf( "Waiting for authorization response.\n" );
  
  SET_NEXT_STATE( oscarLoginAuthorization );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_OSC_AUTH_TIMEOUT );
  ADVANCE_STATE_MACHINE;
  
  authData.printData();

  if ( authData.errorCode )
  {
    USHORT theWindowError;
    
    errorCode = MRM_OSC_AUTH_ERROR;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    
    switch ( authData.errorCode )
    {
      case 1:
      case 4:
        theWindowError = MRM_AIM_BAD_IDPASS;
      break;
      case 5:
        theWindowError = MRM_AIM_BAD_PASSWORD;
      break;
      case 6:
        theWindowError = MRM_AIM_AUTH_FAILED;
      break;
      case 8:
        theWindowError = MRM_AIM_ACCOUNT_DELETED;
      break;
      case 12:
      case 13:
      case 18:
      case 19:
      case 20:
      case 21:
      case 26:
      case 31:
        theWindowError = MRM_AIM_SERVICE_TEMP_UNAVAILABLE;
      break;
      case 17:
        theWindowError = MRM_AIM_ACCOUNT_SUSPENDED;
      break;
      case 24:
      case 29:
        theWindowError = MRM_AIM_RATE_LIMIT_HIT;
      break;
      case 27:
      case 28:
        theWindowError = MRM_AIM_FORCE_UPGRADE;
      break;
      case 32:
        theWindowError = MRM_AIM_INVALID_SECUREID;
      break;
      case 34:
        theWindowError = MRM_AIM_ACCOUNT_SUSPENDED_MINOR;
      break;
      default:
        theWindowError = MRM_AIM_UNKNOWN_LOGIN_FAILURE;
    }
    
    debugf( "The specific problem was:\n" );
    debugf( "%s\n", errorMessageStrings[ theWindowError ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( theWindowError ),
     MPFROMP( theRecord ) );
    
    return;
  }
  
  if ( !authData.BOSserver )
  {
    errorCode = MRM_OSC_NO_BOS;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  BOSsocket = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
  
  if ( BOSsocket < 0 )
  {
    errorCode = MRM_BOS_SOCKET_OPEN;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  tmpStr = strrchr( authData.BOSserver, ':' );
  if ( tmpStr ) *tmpStr = 0;
  // The port number is usually returned, but not for old servers
  
  hostEnt = gethostbyname( authData.BOSserver );
  if ( !hostEnt )
  {
    errorCode = MRM_BOS_GET_BOS_IP;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  memcpy( &oscarAddr.s_addr, hostEnt->h_addr, hostEnt->h_length );
  
  debugf( "Connecting to BOS server: %s\n", inet_ntoa( oscarAddr ) );
  
  memset( &serverDesc, 0, sizeof( struct sockaddr_in ) );
  serverDesc.sin_len = sizeof( struct sockaddr_in );
  serverDesc.sin_family = AF_INET;
  serverDesc.sin_addr.s_addr = oscarAddr.s_addr;
  if ( tmpStr )
  {
    serverDesc.sin_port = htons( atol( tmpStr+1 ) );
  } else {
    serverDesc.sin_port = htons( 5190 ); // Standard OSCAR port
  }
  
  SET_NEXT_STATE( BOSconnectionAck );
  
  rc = connect( BOSsocket, (struct sockaddr *)&serverDesc,
   sizeof( struct sockaddr_in ) );
   
  if ( rc )
  {
    errorCode = MRM_BOS_SOCKET_CONNECT;
    status = fatalError;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    WinPostMsg( mgrWin, WM_ERRORMESSAGE, MPFROMSHORT( errorCode ),
     MPFROMP( theRecord ) );
    return;
  }
  
  sendData.prepareSignoffData();
  sendData.sendData( oscarSocket );
  
  rc = soclose( oscarSocket );
  oscarSocket = 0;
  if ( rc )
  {
    errorCode = MRM_OSC_SOCKET_CLOSE;
    debugf( "%s\n", errorMessageStrings[ errorCode ] );
    // Not really a fatal error, so keep going.
  }
  
  DosPostEventSem( continuousReceive );
  // Allow us to receive data again
  
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_BOS_ACK_TIMEOUT );
  ADVANCE_STATE_MACHINE;
  
  sendData.prepareAuthorizationData( &authData );
  debugf( "Sending authorization information.\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOShostReady );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_BOS_LOGIN_TIMEOUT );
  ADVANCE_STATE_MACHINE;

  sendData.prepareVersionsRequest();
  debugf( "Requesting versions of available Oscar services.\n" );
  sendData.sendData( BOSsocket );

  SET_NEXT_STATE( BOSservicesVersion );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_BOS_SERV_VER_TIMEOUT );
  ADVANCE_STATE_MACHINE;
  
  sendData.prepareReqForSelfInformation();
  debugf( "Requesting user information.\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOScurrentUserInfo );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
  userInfo.printData();
  
  sendData.prepareRateInfoRequest();
  debugf( "Requesting rate information.\n" );
  sendData.sendData( BOSsocket );

  SET_NEXT_STATE( BOSrateInformation );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
  if ( rateInformation )
  {
    sendData.prepareRateInformationAck( rateInformation );
    debugf( "Acknowledging rate information.\n" );
    sendData.sendData( BOSsocket );
  }
  
  sendData.prepareCapabilitiesInfo();
  debugf( "Sending capabilties information.\n" );
  sendData.sendData( BOSsocket );
  
  sendData.prepareRequestSSIlimits();
  debugf( "Requesting server-stored information limitations.\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOSserverStoredLimits );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;

  sendData.prepareReqForSSI();
  debugf( "Requesting server-stored information (SSI).\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOSserverStoredInfo );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
  ssiData.printData();
  
  buddyCreateData.theSession = this;
  buddyCreateData.userInfo = &userInfo;
  DosCreateEventSem( NULL, &buddyCreateData.wakeupSem, 0, 0 );
  
  buddyCreateData.buddyListWin = NULLHANDLE;
  // Will get populated after event sem is posted
  
  WinPostMsg( mgrWin, WM_CREATEBUDDYLIST, MPFROMP( theRecord ),
   MPFROMP( &buddyCreateData ) );
   
  DosWaitEventSem( buddyCreateData.wakeupSem, SEM_INDEFINITE_WAIT );
  // Buddy list window handle is now valid
  
  sessionWindow = buddyCreateData.buddyListWin;
  
  numBuddies = ssiData.numRootBuddies;
  
  CHECKED_CALLOC( numBuddies * sizeof( Buddy * ), buddyList );
  
  for ( i=0; i<numBuddies; ++i )
  {
    ssiData.rootBuddies[i].beenHere = 0;
  }
  
  for ( i=0; i<numBuddies; ++i )
  {
    if ( ssiData.rootBuddies[i].numMembers &&
          !ssiData.rootBuddies[i].beenHere )
    {
      found = 0;
      
      for ( j=0; j<numBuddies; ++j )
      {
        if ( !ssiData.rootBuddies[j].beenHere && j!=i &&
              ssiData.rootBuddies[j].numMembers )
        {
          for ( k=0; k<ssiData.rootBuddies[j].numMembers; ++k )
          {
            if ( ssiData.rootBuddies[j].memberIDs[k] ==
                  ssiData.rootBuddies[i].id &&
                  ssiData.rootBuddies[i].gid == ssiData.rootBuddies[j].gid )
            {
              found = 1;
            }
          }
        }
      }
      
      if ( !found )
      {
        // Current highest level group
        ssiData.rootBuddies[i].parentRecord = NULL;
        ssiData.rootBuddies[i].beenHere = 1;
        
        buddyList[i] = new Buddy( sessionWindow, buddyCreateData.wakeupSem,
         ssiData.rootBuddies + i );
        
        addAllGroupMembers( i, buddyList, &ssiData, sessionWindow,
         buddyCreateData.wakeupSem );
      }
    }
  }
  
  for ( i=0; i<numBuddies; ++i )
  {
    if ( !ssiData.rootBuddies[i].beenHere )
    {
      debugf( "Orphan Member: %s\n", ssiData.rootBuddies[i].entryName );
      ssiData.rootBuddies[i].parentRecord = NULL;
      buddyList[i] = new Buddy( sessionWindow, buddyCreateData.wakeupSem,
       ssiData.rootBuddies + i );
      // Semaphore is reset inside the constructor as needed.  It will be
      //  left in a posted state, though, so reset again if needed outside.

      addAllGroupMembers( i, buddyList, &ssiData, sessionWindow,
       buddyCreateData.wakeupSem );
    } else {
      ssiData.rootBuddies[i].beenHere = 0;
      // Reset the status to be neat about it
    }
  }
  
  DosCloseEventSem( buddyCreateData.wakeupSem );
  
  sendData.prepareReqLocationLimits();
  debugf( "Requesting location service limitations.\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOSlocationLimits );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
  sendData.prepareReqBuddyManagementLimits();
  debugf( "Requesting buddy list management limitations.\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOSbuddyManagementLimits );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
  sendData.prepareRequestPrivacyParams();
  debugf( "Requesting privacy management limitations.\n" );
  sendData.sendData( BOSsocket );

  SET_NEXT_STATE( BOSprivacyLimits );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
  sendData.prepareRequestICBMParams();
  debugf( "Requesting ICBM parameter information.\n" );
  sendData.sendData( BOSsocket );
  
  SET_NEXT_STATE( BOSICBMparams );
  WAIT_CHECK_FOR_ERRORS( RESPONSE_TIMEOUT, MRM_GEN_WAIT_SEM );
  ADVANCE_STATE_MACHINE;
  
/*
  sendData.prepareSetICBMParams( 1 );
  debugf( "Setting ICBM parameters for IMs.\n" );
  sendData.sendData( BOSsocket );

  sendData.prepareSetICBMParams( 2 );
  debugf( "Setting ICBM parameters for chat invitation/file xfer.\n" );
  sendData.sendData( BOSsocket );

  sendData.prepareSetICBMParams( 3 );
  debugf( "Setting ICBM parameters for chat messages.\n" );
  sendData.sendData( BOSsocket );
*/

  sendData.prepareSSIActivation();
  debugf( "Activating server-stored profile.\n" );
  sendData.sendData( BOSsocket );
  
  sendData.prepareSetStatus();
  debugf( "Setting status to on-line.\n" );
  sendData.sendData( BOSsocket );

/*
  sendData.prepareSetUserInformation(
   "This IM user is running a 100% native OS/2 AIM client!", NULL );
  debugf( "Sending user profile and capabilities information.\n" );
  sendData.sendData( BOSsocket );
*/

  status = BOSserviceLoop;
  DosPostEventSem( continuousReceive );
  // All received data should be asynchronous from now on...
  
  sendData.prepareClientReady( supportedFamilies );
  debugf( "Sending the client ready signal.\n" );
  sendData.sendData( BOSsocket );
  
  CHECKED_MALLOC( 17 + strlen( userInfo.screenName ), tmpStr );
  sprintf( tmpStr, "\\QUEUES\\%s\\Rate", userInfo.screenName );
  j = strlen( tmpStr );
  
  for ( i=0; rateInformation[i].rateClass; ++i )
  {
    // Create rate regulation threads and queues
    if ( rateInformation[i].numAppliesTo )
    {
      rateInformation[i].sendSocket = BOSsocket;
      rateInformation[i].pauseResume = pauseResume;
      ltoa( i, tmpStr + j, 10 );
      if ( DosCreateQueue( &rateInformation[i].outgoingData, QUE_FIFO, tmpStr ) )
      {
        debugf( "Failed to create rate regulation queue.  Cannot service session.\n" );
        status = fatalError;
        CHECKED_FREE( tmpStr );
        return;
      }
      rateInformation[i].rateThread =
       _beginthread( (void (*) (void *))rateThread, NULL, 16384,
        rateInformation + i );
    }
  }
  
  CHECKED_FREE( tmpStr );
}

static void addAllGroupMembers( int entry, Buddy **buddyList,
 SSIData *ssiData, HWND sessionWindow, HEV wakeupSem )
{
  int i, j;
  
  for ( i=0; i < ssiData->rootBuddies[entry].numMembers; ++i )
  {
    for ( j=0; j < ssiData->numRootBuddies; ++j )
    {
      if ( ssiData->rootBuddies[j].id ==
            ssiData->rootBuddies[entry].memberIDs[i] &&
           ssiData->rootBuddies[j].gid == ssiData->rootBuddies[entry].gid )
      {
        ssiData->rootBuddies[j].parentRecord =
         ssiData->rootBuddies[entry].myRecord;
         
        ssiData->rootBuddies[j].beenHere = 1;
        
        buddyList[j] = new Buddy( sessionWindow, wakeupSem,
         ssiData->rootBuddies + j );
        // Semaphore is reset inside the constructor as needed.  It will be
        // left in a posted state, though, so reset again if needed outside.
        
        if ( ssiData->rootBuddies[j].numMembers )
        {
          // Recurse until we hit the leaf nodes
          addAllGroupMembers( j, buddyList, ssiData, sessionWindow, wakeupSem );
        }
      }
    }
  }
}

void OscarSession :: serviceSessionUntilClosed( void )
{
  if ( status == fatalError ) return;
  // Nothing to service.  We couldn't start up.
  
  debugf( "Session was established successfully.\n" );
  
  DosCreateEventSem( NULL, &sessionShutdown, 0, 0 );
  WinPostMsg( sessionWindow, WM_REGISTERWAKEUPSEM,
   MPFROMLONG( sessionShutdown ), NULL );
  DosWaitEventSem( sessionShutdown, SEM_INDEFINITE_WAIT );
  debugf( "Session thread %d awoke from its slumber.\n", *_threadid );
}

void OscarSession :: closeSession( void )
{
  HWND tmpWin = sessionWindow;
  sessionWindow = 0;
  debugf( "A forceful close of this session was initiated.  Closing window.\n" );
  WinSendMsg( tmpWin, WM_CLOSE, NULL, NULL );
}

void OscarSession :: windowClosed( void )
{
  sessionWindow = 0;
  // Make sure we don't try to close the window again when we destroy this
  //  object as it has already been closed by the user.
}

OscarSession :: ~OscarSession()
{
  status = shutDown;
  
  if ( sessionWindow )
  {
    debugf( "Closing up session's buddy list window.\n" );
    WinPostMsg( sessionWindow, WM_CLOSE, NULL, NULL );
  }
  
  receiveData.shutdown();
  
  if ( continuousReceive )
  {
    DosPostEventSem( continuousReceive );
  }
  
  if ( pauseResume )
  {
    DosPostEventSem( pauseResume );
  }
  
  if ( BOSsocket )
  {
    debugf( "Cleanup shutting down BOS socket.\n" );
    soclose( BOSsocket );
  }
  
  if ( oscarSocket )
  {
    debugf( "Cleanup shutting down Oscar socket.\n" );
    soclose( oscarSocket );
  }
  
  // Hopefully closing the sockets will wake up the thread waiting to receive
  //  from them.
  
  ULONG receiverThread = dataReceiverThreadID;
  
  if ( receiverThread )
  {
    int rc, timePolled = 0;
    debugf( "Receiver thread %d is still active.  Waiting for it to end.\n",
     receiverThread );
    
    rc = DosWaitThread( &receiverThread, DCWW_NOWAIT );
    
    while ( dataReceiverThreadID && timePolled < 5 &&
             rc == ERROR_THREAD_NOT_TERMINATED )
    {
      DosSleep( 1000 );
      timePolled++;
      rc = DosWaitThread( &receiverThread, DCWW_NOWAIT );
    }
    
    if ( rc == ERROR_THREAD_NOT_TERMINATED && dataReceiverThreadID )
    {
      debugf( "Timed out waiting for thread to end.  Killing thread %d.\n",
       receiverThread );
      DosKillThread( receiverThread );
      // Somehow we stayed blocked somewhere in the receiver thread even
      //  though the sockets were closed.  Should never really happen, but
      //  just in case, this will get us out of a tight spot.
    } else {
      debugf( "Receiver thread %d ended nicely.\n", receiverThread );
    }
  }

  if ( stateWakeup )
  {
    DosCloseEventSem( stateWakeup );
  }
  
  if ( continuousReceive )
  {
    DosCloseEventSem( continuousReceive );
  }
  
  if ( pauseResume )
  {
    DosCloseEventSem( pauseResume );
  }
  
  if ( waitStateMux )
  {
    DosCloseMutexSem( waitStateMux );
  }

  if ( supportedFamilies ) CHECKED_FREE( supportedFamilies );
  
  if ( rateInformation )
  {
    REQUESTDATA rd;
    PVOID crap;
    BYTE crap2;
    ULONG len;
    int i = 0;
    OscarData *queuedData;
    
    while ( rateInformation[i].rateClass )
    {
      if ( rateInformation[i].numAppliesTo )
      {
        rateInformation[i].sendSocket = 0;
        // Fast-forward through any pending data by not sending it
        
        DosWriteQueue( rateInformation[i].outgoingData, 0, 0, 0, 0 );
        // Unblock if we're waiting for data
        
        DosWaitThread( &rateInformation[i].rateThread, DCWW_WAIT );
        // Thread should be shutting down or done shutting down after
        //  writing a 0 to the queue.
        
        while ( DosReadQueue( rateInformation[i].outgoingData, &rd, &len, &crap,
                 0, DCWW_NOWAIT, &crap2, NULLHANDLE ) == 0 )
        {
          queuedData = (OscarData *)rd.ulData;
          delete queuedData;
          // Free any remaining data in the queue
        }
        
        DosCloseQueue( rateInformation[i].outgoingData );
        
        CHECKED_FREE( rateInformation[i].appliesToFam );
        CHECKED_FREE( rateInformation[i].appliesToSub );
      }
      ++i;
    }
    CHECKED_FREE( rateInformation );
  }
  
  if ( sessionShutdown ) DosCloseEventSem( sessionShutdown );
  
  if ( numBuddies )
  {
    int i;
    for ( i=0; i<numBuddies; ++i )
    {
      delete buddyList[i];
      buddyList[i] = NULL;
    }
    CHECKED_FREE( buddyList );
  }

  debugf( "Session has ended.\n" );
}

void OscarSession :: reportAvailableServices( void )
{
  unsigned short i=0;
  
  debugf( "\n" );
  debugf( "Available IM services include:\n" );
  debugf( "--------------------------------------\n" );  
  
  while ( supportedFamilies[i] )
  {
    switch ( supportedFamilies[i] )
    {
      case 1:  debugf( "Generic service controls   [Family 1]\n" );  break;
      case 2:  debugf( "Location services          [Family 2]\n" );  break;
      case 3:  debugf( "Buddy List management      [Family 3]\n" );  break;
      case 4:  debugf( "Messaging (ICBM) service   [Family 4]\n" );  break;
      case 5:  debugf( "Advertisements 'service'   [Family 5]\n" );  break;
      case 6:  debugf( "Invitation service         [Family 6]\n" );  break;
      case 7:  debugf( "Administration service     [Family 7]\n" );  break;
      case 8:  debugf( "Popup notices service      [Family 8]\n" );  break;
      case 9:  debugf( "Privacy management         [Family 9]\n" );  break;
      case 10: debugf( "User lookup service        [Family 10]\n" ); break;
      case 11: debugf( "Usage stats service        [Family 11]\n" ); break;
      case 12: debugf( "Translation service        [Family 12]\n" ); break;
      case 13: debugf( "Chat navigation service    [Family 13]\n" ); break;
      case 14: debugf( "Chat service               [Family 14]\n" ); break;
      case 15: debugf( "Directory user search      [Family 15]\n" ); break;
      case 16: debugf( "Buddy icon service         [Family 16]\n" ); break;
      case 19: debugf( "Server-stored information  [Family 19]\n" ); break;
      case 21: debugf( "ICQ extensions             [Family 21]\n" ); break;
      case 23: debugf( "Authorization/registration [Family 23]\n" ); break;
      default:
        debugf( "Unknown service            [Family %d]\n",
         supportedFamilies[i] );
    }
    ++i;
  }
  debugf( "--------------------------------------\n" );
  debugf( "\n" );
}

void OscarSession :: queueForSend( OscarData *queueData )
{
  unsigned short family, subType;
  int i, j;
  
  queueData->setSeqNumPointer( &currentSequence );
  // Ensure that this is set correctly.
  
  queueData->getSNACtype( &family, &subType );
  if ( !family )
  {
    // No SNAC.  Just send it directly.
    queueData->sendData( BOSsocket );
    return;
  }
  
  debugf( "Enqueuing data for family %d, subtype %d.\n", family, subType );
  
  for ( i=0; rateInformation[i].rateClass; ++i )
  {
    for ( j=0; j<rateInformation[i].numAppliesTo; ++j )
    {
      if ( rateInformation[i].appliesToFam[j] == family &&
            rateInformation[i].appliesToSub[j] == subType )
      {
        DosWriteQueue( rateInformation[i].outgoingData, (ULONG)queueData,
         0, NULL, 0 );
        return;
      }
    }
  }
  
  // If we got here, then we couldn't match the family/subtype with a rate
  //  class.  So just send the data in an unrestricted way.
  
  debugf( "Family/subtype is not regulated." );
  queueData->sendData( BOSsocket );
}

